/***************************************************************************
 FILE: bool.c
 LVU : 0.8.0

 DESC:
 This contains the code to evaluate boolean expressions used in the IF,
 WHILE and WHILST commands.

 Copyright (C) Neil Robertson 1999
 ***************************************************************************/

#include "globals.h"


/*** Evaluate complete boolean which involves dealing with compound
     expressions inside brackets ***/
int evaluate_complete_boolean(int pc, int *res)
{
int pc2,start,end,result;
int cur_brackets,ret;

*res=0;
result=0;
end=pword[pc].end;
for(pc2=pc;pc2 <= end;++pc2) {
	pword[pc2].processed=NO;
	pword[pc2].result=-1;
	}
START:
start=pc;
cur_brackets=0;

/* Find stuff inside round brackets. Ignore any inside curly
   brackets. */
for(pc2=pc;pc2 <= end;++pc2) {

	switch(pword[pc2].word[0]) {
		case '{':
		++cur_brackets;  continue;

		case '}':
		--cur_brackets;  continue;
		}
	if (cur_brackets) continue;
	if (pword[pc2].processed) continue;

	switch(pword[pc2].word[0]) {
		case '(':
		start=pc2;  
		continue;

		case ')':
		if (pc2 - start < 2) return ERR_EMPTY_EXPRESSION;
		pword[start++].processed=YES;
		while(pword[start].processed && start <= end) ++start;
		if ((ret=evaluate_compound_boolean(pc,start,&result))!=OK)
			return ret;
		pword[pc2].processed=YES;
		pword[start].result=(double)result;
		pword[start].processed=NO;
		goto START;
		}
	}
return evaluate_compound_boolean(pc,start,res);
}



/*** Evaluate compound boolean expression. eg: x < y and y > 2 and ... ***/
int evaluate_compound_boolean(int exp_start, int pc, int *res)
{
int pc2,pc3,ret;
int type,single_result;

pc2=pc;
type=0;
*res=0;
while(NOT_PAST_END) {
	if (pword[pc2].processed) {  ++pc2;  continue;  }
	if (pword[pc2].opcode && pword[pc2].word[0]!='{') return ERR_SYNTAX;

	pc3=pc2;
	if (pword[pc2].result!=-1) {
		single_result=(int)pword[pc2].result;  ++pc2;
		}
	else {
		if ((ret=evaluate_single_boolean(exp_start,&pc2,&single_result))!=OK) 
			return ret;
		pword[pc3].result=(double)single_result;
		}
	
	/* Set to processed */
	for(;pc3 < pc2;++pc3) pword[pc3].processed=YES;

	if (type) {
		switch(type) {
			case AND_COM: *res&=single_result;  break;
			case OR_COM : *res|=single_result;  break;
			case XOR_COM: *res^=single_result;  break;
			default : return ERR_INTERNAL;
			}
		}
	else *res=single_result;
	while(pword[pc2].processed) ++pc2;

	if (IS_PAST_END) break;
	if (pword[pc2].word[0]==')') break;

	/* Get boolean command */
	type=pword[pc2].com;
	pword[pc2].processed=YES;
	if (type!=AND_COM && type!=OR_COM && type!=XOR_COM) return ERR_SYNTAX; 

	if (pword[++pc2].processed) continue;

	/* We might be at a "3<2 and (" type situation */
	if (pword[pc2].word[0]=='(') break;
	pword[pc2-1].processed=YES;
	}
return OK;
}



/*** Evaluate single boolean expression Eg: x < y , x != y etc ***/
int evaluate_single_boolean(int exp_start, int *pc, int *result)
{
int pc2,ret,stored;
int opcode,compres,both_numbers;
char *val1,*val2;

pc2=*pc;

/* If we're going to run a CALL command which may be a resursive 
   then store current results for this proc stack context */
if (pword[pc2].word[0]=='{' && nested_call_command(pc2)) {
	if ((ret=store_in_pstack(exp_start))!=OK) return ret;
	stored=1;
	}
else stored=0;

if ((ret=push_rstack_result(&pc2))!=OK) return ret; 
val1=rstack_ptr->value;

if (stored) copy_from_pstack(exp_start);

/* If no operation given then result is TRUE if value is not NULL and not 
   zero */
if (pword[*pc].end < pc2 || 
    pword[pc2].word[0]==')' ||
    pword[pc2].com==AND_COM ||
    pword[pc2].com==OR_COM ||
    pword[pc2].com==XOR_COM) {
	if (val1) *result=(strcmp(val1,"0") ? 1 : 0);
	else *result=0;
	*pc=pc2;
	return OK;
	}
if (pword[pc2].end == pc2) return ERR_SYNTAX;
opcode=pword[pc2].opcode;
++pc2;

/* Get value of 2nd argument */
if (pword[pc2].word[0]=='{' && nested_call_command(pc2)) {
	if ((ret=store_in_pstack(exp_start))!=OK) return ret;
	stored=1;
	}
else stored=0;

if ((ret=push_rstack_result(&pc2))!=OK) return ret;
val2=rstack_ptr->value;

if (stored) copy_from_pstack(exp_start);

if (number_type(val1) != NOT_NUMBER && number_type(val2) != NOT_NUMBER) 
	both_numbers=1;
else {
	compres=compare_strings(val1,val2);
	both_numbers=0;
	}

switch(opcode) {
	case EQUALS:
	if (both_numbers) *result=(atof(val1) == atof(val2));
	else *result=!compres;
	break;

	case NOT_EQUAL:
	if (both_numbers) *result=(atof(val1) != atof(val2));
	else *result=(compres != 0);
	break;

	case LESS_THAN:
	if (both_numbers) *result=(atof(val1) < atof(val2));
	else *result=(compres < 0);
	break;

	case GREATER_THAN:
	if (both_numbers) *result=(atof(val1) > atof(val2));
	else *result=(compres > 0);
	break;

	case LESS_THAN_EQUALS:
	if (both_numbers) *result=(atof(val1) <= atof(val2));
	else *result=!(compres > 0);
	break;

	case GREATER_THAN_EQUALS:
	if (both_numbers) *result=(atof(val1) >= atof(val2));
	else *result=!(compres < 0);
	break;

	default: return ERR_SYNTAX;
	}
*pc=pc2;
return OK;
}



/*** Compare to strings for equality or not ***/
int compare_strings(char *s1, char *s2)
{
int res;

if (s1==NULL) {
	if (s2==NULL || !*s2) return 0;
	return -1;
	}
if (s2==NULL) {
	if (s1==NULL || !*s1) return 0;
	return 1;
	}
res=strcoll(s1,s2);
return (res < 0 ? -1 : (res > 0 ? 1 : 0));
}
