/***************************************************************************
 FILE: com_excep.c
 LVU : 1.0.0

 DESC:
 This contains some of the exception handling commands.

 Copyright (C) Neil Robertson 1999, 2000
 ***************************************************************************/

#include "globals.h"

/*** Trap command ***/
int com_trap(int com, int pc)
{
int ret,i;

if (pword[pc].end == pc) return ERR_SYNTAX;
if (pword[pc+1].word[0]=='{' || pword[pc+1].word[0]=='"') 
	return ERR_INVALID_ARGUMENT;

/* Can only trap errors stored in err_trapped */
++pc;
ret=run_command(&pc);
if (ret!=OK) {
	for(i=0;i < EXCP_TRAP_COUNT;++i)
		if (ret==err_trapped[i].err) break;
	if (i==EXCP_TRAP_COUNT) return ERR_CANT_TRAP_ERROR;
	}

/* Set error values */
sprintf(text,"%d",errno);
set_var_value("$last_os_error",NULL,text,YES);
sprintf(text,"%d",ret);
set_var_value("$last_error",NULL,text,YES);

return push_rstack(text);
}



/*** Manually throw an exception. This basically just returns whatever 
     error you want and the normal error handling code sorts out the rest ***/
int com_throw(int com, int *pc)
{
int ret,end,i;

if ((end=pword[*pc].end) == *pc) return ERR_SYNTAX;

/* Get exception */
(*pc)++;
if ((ret=push_rstack_result(pc)) != OK) return ret;
if (*pc <= end) return ERR_SYNTAX;
if (isnull(rstack_ptr->value)) return ERR_INVALID_ARGUMENT;

/* Check valid exception */
for(i=0;i < EXCP_TRAP_COUNT;++i) 
	if (!strcasecmp(rstack_ptr->value,err_trapped[i].name)) break;
if (i==EXCP_TRAP_COUNT) return ERR_INVALID_EXCEPTION;

manual_throw=YES;
return err_trapped[i].err;
}



/*** Return from a catch block to the next statement after the one that
     failed in the try block. ***/
int com_catchreturn(int com, int *pc)
{
int index;

if (pword[*pc].end != *pc) return ERR_SYNTAX;

/* Different index if catchreturn or catchreturnend */
index=com - CATCHRETURN;

/* This generally won't happen unless we're in a catchrun - caught block and
   didn't get there via an exception */
if (!pstack_ptr->catchreturn_pc[index]) return ERR_CANT_CATCHRETURN;

/* Jump to line after exception line or after try-tried block if catchreturn
   or catchreturnend respectively */
*pc=pstack_ptr->catchreturn_pc[index];
cleardown_lw_stacks(*pc);

return OK;
}



/*** Break out of a catch-caught block and jump to the statement immediately
     following it ***/
int com_catchbreak(int com, int *pc)
{
if (pword[*pc].end != *pc) return ERR_SYNTAX;
if (!pstack_ptr->catchbreak_pc) return ERR_CANT_CATCHBREAK;

*pc=pstack_ptr->catchbreak_pc;
cleardown_lw_stacks(*pc);
return OK;
}
