/***************************************************************************
 FILE: com_io.c
 LVU : 1.0.0

 DESC:
 This is the code for the input, output and filesystem commands.

 Copyright (C) Neil Robertson 1999, 2000
 ***************************************************************************/

#include "globals.h"


/*** Select on a number of streams ***/
int com_select(int com, int pc)
{
STREAM_T st;
VAR_T var,ve;
struct timeval tv;
char *sub;
int ret,pc2,secs,cnt,val;
fd_set rmask,wmask;

if (pword[pc].end == pc) return ERR_SYNTAX;

/* Get array */
if ((ret=get_var_main(pword[pc+1].word,pstack_ptr,&var,&sub)) != OK)
	return ret;
if (!var->dim) return ERR_VAR_IS_NOT_ARRAY;

/* Get optional timeout in seconds */
if (pword[pc].end > pc+1) {
	pc2=pc+2;
	if ((ret=push_rstack_result(&pc2)) != OK) return ret;
	if (NOT_PAST_END) return ERR_SYNTAX;

	if (number_type(rstack_ptr->value) == NOT_NUMBER ||
	    (secs=atoi(rstack_ptr->value)) < 0)
		return ERR_INVALID_ARGUMENT;

	tv.tv_sec=secs;
	tv.tv_usec=get_usecs(rstack_ptr->value);
	}
else secs=-1;

/* Setup select readmask & writemask. The stream names are in the SUBSCRIPT, 
   NOT the value which holds the 0 or 1 result. If the value is "R" before
   setting that means is for a read select, "W" means a write select, "RW"
   means read-write. Anything else is an error. */
FD_ZERO(&rmask);
FD_ZERO(&wmask);
FOR_VAR_ELEMENTS(var,ve) {
	if ((st=get_stream(ve->name)) == NULL) return ERR_UNDEFINED_STREAM;

	if (ve->value==NULL || 
	    (ve->value[1] && ve->value[1] != 'W') ||
	    strlen(ve->value) > 2) return ERR_INVALID_ARGUMENT;

	switch(ve->value[0]) {
		case 'R': 
		if (st->open_mode==STM_WRITE || st->open_mode==STM_APPEND)
			return ERR_STREAM_OUTPUT_ONLY;
		FD_SET(st->fd,&rmask);  
		if (ve->value[1] != 'W') break;

		case 'W': 
		if (st->open_mode==STM_READ) return ERR_STREAM_INPUT_ONLY;
		FD_SET(st->fd,&wmask);  
		break;

		default : return ERR_INVALID_ARGUMENT;
		}
	set_string(&ve->value,"0");
	}

/* Do the select */
if (secs==-1) ret=select(FD_SETSIZE,&rmask,&wmask,0,0); /* wait forever */
else ret=select(FD_SETSIZE,&rmask,&wmask,0,&tv); /* wait for timeout */

/* If we've been interrupted by a signal just return */
if (ret==-1) return push_rstack("-1");

/* Set results in array and count how many streams have data */
cnt=0;
FOR_VAR_ELEMENTS(var,ve) {
	if (st=get_stream(ve->name)) {
		val=((FD_ISSET(st->fd,&rmask) != 0) + 2*(FD_ISSET(st->fd,&wmask) != 0));
		if (val) {
			sprintf(text,"%d",val);
			set_string(&ve->value,text);  
			++cnt;
			}
		}
	}	
sprintf(text,"%d",cnt);
return push_rstack(text);
}



/*** Print output to stdout or given stream if stprint(nl) ***/
int com_print(int com, int pc)
{
STREAM_T st;
int ret,pc2,len;

pc2=pc+1;
if (com >= PRINT && com <= PRINTNL_SHORTCUT) {
	if (out_stream==NULL) return ERR_OUTPUT_STREAM_CLOSED;
	st=out_stream;
	}
else {
	if (pword[pc].end == pc) return ERR_SYNTAX;

	/* Get the stream to output to */
	if ((ret=push_rstack_result(&pc2))!=OK) return ret;
	if (isnull(rstack_ptr->value)) return ERR_INVALID_ARGUMENT;

	if ((st=get_stream(rstack_ptr->value)) == NULL)
		return ERR_UNDEFINED_STREAM;

	if (st->open_mode==STM_READ) return ERR_STREAM_INPUT_ONLY;
	if (IS_PAST_END) goto END;
	}

/* Output the data */
while(NOT_PAST_END) {
	if ((ret=push_rstack_result(&pc2))!=OK) return ret;
	if (rstack_ptr->value) {
		len=strlen(rstack_ptr->value);
		write_stream(st,rstack_ptr->value,len);
		st->pos+=len;
		st->write_cnt+=len;
		}
	}

/* Print newline if ..nl command */
END:
if (com==PRINTNL || 
    com==PRINTNL_SHORTCUT || 
    com==STPRINTNL ||
    com==STPRINTNL_SHORTCUT) {
	write_stream(st,"\n",1);  
	st->pos++;
	st->write_cnt++;
	}
create_stream_array_entry(st);
return OK;
}



/*** Get input from stdin ***/
int com_input(int com, int pc)
{
VAR_T var;
STREAM_T st;
struct timeval tv;
char *sub,c;
int ret,pc2,dummy,selret;
int bytes,cnt,not_binput;
fd_set mask;

if (pword[pc].end == pc) return ERR_SYNTAX;

pc2=pc+1;
if (com==INPUT || com==BINPUT) {
	if (in_stream==NULL) return ERR_INPUT_STREAM_CLOSED;
	st=in_stream;
	}
else {
	/* Get the stream to input from */
	if ((ret=push_rstack_result(&pc2))!=OK) return ret;
	if (isnull(rstack_ptr->value)) return ERR_INVALID_ARGUMENT;

	if ((st=get_stream(rstack_ptr->value)) == NULL)
		return ERR_UNDEFINED_STREAM;

	if (st->open_mode==STM_APPEND || st->open_mode==STM_WRITE) 
		return ERR_STREAM_OUTPUT_ONLY;
	if (IS_PAST_END) return ERR_SYNTAX;
	}
	
if (pword[pc2].word[0]=='{' || pword[pc2].word[0]=='"')
	return ERR_INVALID_ARGUMENT;

/* Get the variable */
if ((ret=get_var_main(pword[pc2].word,pstack_ptr,&var,&sub))!=OK)
	return ret;
if (var->read_only) return ERR_READONLY_VARIABLE;
if (sub &&
    (ret=getset_var_element(var,pstack_ptr,&var,NULL,sub,NO,&dummy))!=OK)
	return ret;

/* Get optional number of bytes to input */
if (pc2 < pword[pc].end) {
	pc2++;
	if ((ret=push_rstack_result(&pc2))!=OK) return ret;

	if (NOT_PAST_END) return ERR_SYNTAX;

	if (number_type(rstack_ptr->value) != INTEGER ||
	    (bytes=atoi(rstack_ptr->value)) < 0 ||
	    (!bytes && (com==BINPUT || com==STBINPUT))) 
		return ERR_INVALID_ARGUMENT;
	}
else {
	if (com==BINPUT || com==STBINPUT) 
		return ERR_SYNTAX; /* need length arg for binput */
	bytes=0;
	}

/* Set up some stuff */
cnt=0;
FREE(var->value);
var->value=NULL;
set_var_value("$EOF",NULL,"0",YES);
set_var_value("$telopt",NULL,"0",YES);
not_binput=(com==INPUT || com==STINPUT);

/* Get the input. We either count up to a certain number of characters
   or return on reaching a newline */
while(!bytes || cnt != bytes) {
	/* NONBLOCK  - exit loop whether we've got a character or not.
	   SEMIBLOCK - exit loop after given number of characters or
	               newline are entered
	   BLOCK - loop till newline pressed
	 */
	FD_ZERO(&mask);
	FD_SET(st->fd,&mask);

	/* If non blocking do a non blocking select */
	if (st->block != YES) {
		tv.tv_sec=0;
		tv.tv_usec=0;
		selret=select(FD_SETSIZE,&mask,0,0,&tv);

		if (st->block==NO && !FD_ISSET(st->fd,&mask)) break;
		}
	else selret=select(FD_SETSIZE,&mask,0,0,0);

	/* Can't do much about an error, could be an alarm signal for 
	   example */
	if (selret==-1) break; 

	/* See if we've got something to read and if so read it */
	if (FD_ISSET(st->fd,&mask)) {
		/* End of file? */
		if (read(st->fd,&c,1) < 1) {
			set_var_value("$EOF",NULL,"1",YES);  break;
			}
		st->pos++;
		st->read_cnt++;
		create_stream_array_entry(st);

		/* If not binary input then check for various termination
		   conditions */
		if (not_binput) {
			if (st->type==ST_SOCKET) {
				if (st->telopt_cnt==2) st->telopt_cnt=1;
				else
				if (st->telopt_cnt==1) {
					st->telopt_cnt=0;  
					append_char(&var->value,c,cnt);
					++cnt;
					goto END;
					}
				else
				/* If we've got potential telopt then set 
				   $telopt and set countdown */
				if ((u_char)c==255) {
					set_var_value("$telopt",NULL,"1",YES);
					st->telopt_cnt=2;
					}
				}

			if (!c) break;
			if (c=='\r') continue;

			/* If we're stripping newlines then break here */
			if (st->stripnl && !bytes && c=='\n')
				break;
			}
		append_char(&var->value,c,cnt);
		++cnt;

		/* Break if we've just past a newline */
		if (not_binput && !bytes && c=='\n') break;
		}
	}
END:
append_char(&var->value,'\0',cnt);

/* Return count of number of characters we've read or -1 if select was
   interrupted. */
if (selret==-1) return push_rstack("-1");
sprintf(text,"%d",cnt);
return push_rstack(text);
}



/*** Change the input or output stream ***/
int com_in_out(int com, int pc)
{
STREAM_T st;
int ret,pc2;

if (pword[pc].end == pc) return ERR_SYNTAX;

/* Get stream name */
pc2=pc+1;
if ((ret=push_rstack_result(&pc2))!=OK) return ret;
if (NOT_PAST_END) return ERR_SYNTAX;
if (isnull(rstack_ptr->value)) return ERR_INVALID_ARGUMENT;

/* Look for stream */
if ((st=get_stream(rstack_ptr->value)) == NULL)
	return ERR_UNDEFINED_STREAM;

/* Make sure stream is correct type */
if (com==OUT) {
	/* output stream */
	if (st->open_mode==STM_READ) return ERR_STREAM_INPUT_ONLY;
	if ((ret=set_var_value("$out_stream",NULL,st->name,YES)) != OK)
		return ret;

	out_stream=st;
	return OK;  /* Stream name is on top of the stack still */
	}

/* input stream */
if (st->open_mode==STM_APPEND || st->open_mode==STM_WRITE) 
	return ERR_STREAM_OUTPUT_ONLY;

if ((ret=set_var_value("$in_stream",NULL,st->name,YES)) != OK) return ret;

in_stream=st;
return OK;
}



/*** Open a file/stream to read/write etc ***/
int com_open(int com, int pc)
{
STREAM_T new;
char *filename;
int end,ret,pc2,type;
mode_t perm;

end=pword[pc].end;
if (end - pc < 3) return ERR_SYNTAX;

/* Get the filename */
pc2=pc+1;
if ((ret=push_rstack_result(&pc2))!=OK) return ret;
if (pc2 >= end) return ERR_SYNTAX;
if (isnull(rstack_ptr->value)) return ERR_INVALID_ARGUMENT;
filename=rstack_ptr->value;

/* Noise word "to" */
if (strcasecmp(pword[pc2].word,"to")) return ERR_SYNTAX;

/* Open method */
pc2++;
if (!strcasecmp(pword[pc2].word,"read")) type=STM_READ;
else
if (!strcasecmp(pword[pc2].word,"write")) type=STM_WRITE;
else
if (!strcasecmp(pword[pc2].word,"append")) type=STM_APPEND;
else
if (!strcasecmp(pword[pc2].word,"readwrite")) type=STM_READWRITE;
else
return ERR_SYNTAX;

/* See if we have some file permissions following */
if (pc2 < end) {
	++pc2;
	if (strcasecmp(pword[pc2].word,"perm")) return ERR_SYNTAX;
	if (type!=STM_WRITE && type!=STM_READWRITE) 
		return ERR_PERM_ONLY_WITH_WRITE;
	++pc2;
	if (pc2 > end) return ERR_SYNTAX;
	if ((ret=push_rstack_result(&pc2))!=OK) return ret;
	if (pc2 <= end) return ERR_SYNTAX;
	if (!get_perm_value(&perm)) return ERR_INVALID_ARGUMENT;
	}
else perm=DEFAULT_FILE_PERMS; /* defined in globals.h */

if ((ret=create_stream(filename,ST_FILE,type,perm,-1,0,&new))!=OK) return ret;
return push_rstack(new->name);
}



/*** Close a list of files/streams or just close the read/write descriptors ***/
int com_close(int com, int pc)
{
STREAM_T st;
int ret,pc2;

if (pword[pc].end == pc) return ERR_SYNTAX;

for(pc2=pc+1;NOT_PAST_END;) {
	if ((ret=push_rstack_result(&pc2))!=OK) return ret;

	/* "close" just destroys the stream */
	if (com==CLOSE) {
		if ((ret=destroy_stream(rstack_ptr->value)) != OK) return ret;
		continue;
		}

	/* Find the stream for rclose & wclose */
	if ((st=get_stream(rstack_ptr->value)) == NULL)
		return ERR_UNDEFINED_STREAM;

	/* Can only do this on pipe streams for now. The main point of this 
	   is to force an EOF for readwrite streams. */
	if (st->type != ST_PIPE) return ERR_INVALID_STREAM_TYPE;

	switch(st->open_mode) {
		case STM_READ:
		if (com==WCLOSE) return ERR_STREAM_INPUT_ONLY;
	
		/* Destroy stream since now both descriptors will be closed */
		if ((ret=destroy_stream(st->name)) != OK) return ret;
		continue;

		case STM_WRITE:
		if (com==RCLOSE) return ERR_STREAM_INPUT_ONLY;
		if ((ret=destroy_stream(st->name)) != OK) return ret;
		continue;

		case STM_READWRITE:
		/* Close the appropriate descriptor and change the stream
		   type including the entry in the $streams array */
		if (com==RCLOSE) {
			close(st->fd);  st->open_mode=STM_WRITE;
			}
		else {
			close(st->fd2);  st->open_mode=STM_READ;
			}
		if ((ret=delete_array_entry("$streams",st->name)) != OK)
			return ret;

		if ((ret=create_stream_array_entry(st)) != OK) return ret;
		continue;


		/* Pipes shouldn't be of any other type */
		default: return ERR_INTERNAL;
		}	
	}
return OK;
}



/*** Seek on an input stream. Wind (as in winding up a clock) sounds 
     nicer tho ;) ***/
int com_wind(int com, int pc)
{
STREAM_T st;
int ret,pc2,end;
int whence,bytes,pos;

if ((end=pword[pc].end) - pc < 3) return ERR_SYNTAX;

/* Get stream */
pc2=pc+1;
if ((ret=push_rstack_result(&pc2))!=OK) return ret;
if (pc2 > end) return ERR_SYNTAX;
if (isnull(rstack_ptr->value)) return ERR_INVALID_ARGUMENT;

if ((st=get_stream(rstack_ptr->value)) == NULL)
	return ERR_UNDEFINED_STREAM;

/* "to" noise word */
if (strcasecmp(pword[pc2].word,"to")) return ERR_SYNTAX;

/* Get seek option */
++pc2;
if (!strcasecmp(pword[pc2].word,"start")) whence=SEEK_SET;
else
if (!strcasecmp(pword[pc2].word,"current")) whence=SEEK_CUR;
else
if (!strcasecmp(pword[pc2].word,"end")) whence=SEEK_END;
else
return ERR_SYNTAX;

/* Get optional seek byte count */
if (++pc2 <= end) {
	if ((ret=push_rstack_result(&pc2))!=OK) return ret;
	if (pc2 <= end) return ERR_SYNTAX;
	if (number_type(rstack_ptr->value) != INTEGER) 
		return ERR_INVALID_ARGUMENT;
	bytes=atoi(rstack_ptr->value);
	}
else bytes=0;

/* Do the seek. If bytes is negative and whence = SEET_SET I could throw
   an invalid argument error but who knows , maybe some machines allow seeking
   before the start of the file. */
if ((pos=lseek(st->fd,bytes,whence)) == -1)
	return ERR_CANT_WIND;
st->pos=pos;
create_stream_array_entry(st);
return OK;
}



/*** Return the current position in the given stream ***/
int com_spos(int com, int pc)
{
STREAM_T st;
int ret,pc2;

if (pword[pc].end == pc) return ERR_SYNTAX;

/* Get stream */
pc2=pc+1;
if ((ret=push_rstack_result(&pc2))!=OK) return ret;
if (NOT_PAST_END) return ERR_SYNTAX;
if (isnull(rstack_ptr->value)) return ERR_INVALID_ARGUMENT;

if ((st=get_stream(rstack_ptr->value)) == NULL)
	return ERR_UNDEFINED_STREAM;

sprintf(text,"%u",st->pos);
return push_rstack(text);
}



/*** Stat a file. File info is returned as a list. ***/
int com_stat(int com, int pc)
{
struct stat fs;
struct passwd *pwd;
struct group *grp;
char linked_file[256];
int ret,pc2,len;

if (pword[pc].end == pc) return ERR_SYNTAX;
pc2=pc+1;

/* Get the filename */
if ((ret=push_rstack_result(&pc2))!=OK) return ret;
if (NOT_PAST_END) return ERR_SYNTAX;
if (isnull(rstack_ptr->value)) return ERR_INVALID_ARGUMENT;

/* Stat it */
if (com==STAT) {
	if (stat(rstack_ptr->value,&fs) == -1) return ERR_CANT_STAT_FILE;
	}
else if (lstat(rstack_ptr->value,&fs) == -1) return ERR_CANT_STAT_FILE;

/* Create list. Get type first. */
switch(fs.st_mode & S_IFMT) {
	case S_IFDIR : strcpy(text,"DIR");   break;
	case S_IFREG : strcpy(text,"FILE");  break;
	case S_IFLNK : strcpy(text,"LINK");  break;
	case S_IFCHR : strcpy(text,"CDEV");  break;
	case S_IFBLK : strcpy(text,"BDEV");  break;
	case S_IFIFO : strcpy(text,"FIFO");  break;
	case S_IFSOCK: strcpy(text,"SOCK");  break;

	default: strcpy(text,"?"); 
	}

/* Get owner */
if ((pwd=getpwuid(fs.st_uid))==NULL) sprintf(text,"%s %d",text,fs.st_uid);
else sprintf(text,"%s %s",text,pwd->pw_name);

/* Get group */
if ((grp=getgrgid(fs.st_gid))==NULL) sprintf(text,"%s %d",text,fs.st_gid);
else sprintf(text,"%s %s",text,grp->gr_name);

/* Get numeric file data: 
   permissions, size, last_modified, last_changed, last_access */
sprintf(text,"%s %04o %u %u %u %u",text,(u_int)fs.st_mode & 0xFFF,(u_int)fs.st_size,(u_int)fs.st_mtime,(u_int)fs.st_ctime,(u_int)fs.st_atime);

/* If file is a symlink then add file link is pointing to */
if ((fs.st_mode & S_IFMT) == S_IFLNK) {
	if ((len=readlink(rstack_ptr->value,linked_file,256)) == -1)
		return ERR_CANT_READ_LINK;
	linked_file[len]='\0';
	sprintf(text,"%s %s",text,linked_file);
	}

return push_rstack(text);
}



/*** Set the option on the given stream ***/
int com_ioctl(int com, int pc)
{
STREAM_T st;
struct termios tio;
int ret,pc2,opt,flags;
char seq[4];
char *option[]={
	"echo","noecho","block","semiblock","nonblock","nostripnl","stripnl",
	"*"
	};

if (pword[pc].end == pc) return ERR_SYNTAX;

/* Get stream */
pc2=pc+1;
if ((ret=push_rstack_result(&pc2))!=OK) return ret;
if (isnull(rstack_ptr->value)) return ERR_UNDEFINED_STREAM;

if ((st=get_stream(rstack_ptr->value)) == NULL)
	return ERR_UNDEFINED_STREAM;

/* Get option(s) */
for(;NOT_PAST_END;++pc2) {
	if (pword[pc2].word[0]=='"' || pword[pc2].word[0]=='{')
		return ERR_SYNTAX;

	for(opt=0;option[opt][0] != '*';++opt) 
		if (!strcasecmp(pword[pc2].word,option[opt])) break;
	if (option[opt][0]=='*') return ERR_INVALID_OPTION;

	/* Do the business. */
	switch(opt) {
		case 0: 
		/* ECHO */
		if (st->open_mode==STM_WRITE || st->open_mode==STM_APPEND) 
			return ERR_INVALID_OPTION_FOR_STREAM;

		switch(st->type) {
			case ST_FILE:
			case ST_LSOCKET:
			case ST_PIPE:
			return ERR_INVALID_OPTION_FOR_STREAM;

			case ST_CONSOLE:
			tcgetattr(0,&tio);
			tio.c_lflag |= ECHO;
			tcsetattr(0,TCSANOW,&tio);
			break;

			case ST_SOCKET:
			/* Send telopt codes */
			sprintf(seq,"%c%c%c",255,252,1);
			write_stream(st,seq,3);
			break;

			default: return ERR_INTERNAL;
			}
		break;


		case 1:
		/* NOECHO */
		if (st->open_mode==STM_WRITE || st->open_mode==STM_APPEND) 
			return ERR_INVALID_OPTION_FOR_STREAM;

		switch(st->type) {
			case ST_FILE:
			case ST_LSOCKET:
			case ST_PIPE:
			return ERR_INVALID_OPTION_FOR_STREAM;

			case ST_CONSOLE:
			tcgetattr(0,&tio);
			tio.c_lflag &= ~ECHO;
			tcsetattr(0,TCSANOW,&tio);
			break;

			case ST_SOCKET:	
			/* Send telopt codes */
			sprintf(seq,"%c%c%c",255,251,1);
			write_stream(st,seq,3);
			break;

			default: return ERR_INTERNAL;
			}
		break;


		case 2:
		/* BLOCK */
		switch(st->type) {
			case ST_CONSOLE:
			if (st->fd) return ERR_INVALID_OPTION_FOR_STREAM;
			tcgetattr(0,&tio);
			tio.c_lflag |= ICANON;
			tcsetattr(0,TCSANOW,&tio);
			st->block=YES;
			break;

			/* Who needs blocking/non blocking control on a file 
			   for gods sake?? Oh well , have it anyway cos someone
			   will whinge about it if I don't put it in */
			case ST_FILE:
			case ST_SOCKET:
			case ST_LSOCKET:
			/* Read the flags and unset the non-block flag as there 
			   is no O_BLOCK option. */
			flags=fcntl(st->fd,F_GETFL);
			flags &= ~O_NONBLOCK;
			fcntl(st->fd,F_SETFL,flags);
			st->block=YES;
			break;

			default: return ERR_INTERNAL;
			}
		break;
	

		case 3:
		case 4:
		/* SEMI & NON BLOCK */
		switch(st->type) {
			case ST_CONSOLE:
			if (st->fd) return ERR_INVALID_OPTION_FOR_STREAM;
			tcgetattr(0,&tio);
			tio.c_lflag &= ~ICANON;
			tio.c_cc[VTIME] = 0;
			tio.c_cc[VMIN] = 1;  /* Buffer size to 1 */
			tcsetattr(0,TCSANOW,&tio);
			if (opt==3) st->block=SEMI; else st->block=NO;
			break;

			case ST_FILE: 
			case ST_SOCKET:
			case ST_LSOCKET:
			fcntl(st->fd,F_SETFL,O_NONBLOCK);
			if (opt==3) st->block=SEMI; else st->block=NO;
			break;

			default: return ERR_INTERNAL;
			}
		break;


		case 5:
		case 6:
		/* STRIPNL & NONSTRIPNL. Can only strip input streams */
		if (st->type==ST_LSOCKET || 
		    st->open_mode==STM_WRITE || 
		    st->open_mode==STM_APPEND) 
			return ERR_INVALID_OPTION_FOR_STREAM;

		st->stripnl=opt-5;
		break;


		default: return ERR_INTERNAL;
		}
	} /* end for */
return OK;
}



/*** Return a directory listing as ... errr... well a list. Funky huh? ***/
int com_dir(int com, int pc)
{
DIR *dir;
loor_dirent *ds;
int ret,pc2;
char *dirname,*reg,*result;
char c,*slashpos,*regpos;

reg=NULL;
regpos=NULL;

/* Get optional dir, else default to "."  */
if (pword[pc].end > pc) {
	pc2=pc+1;
	if ((ret=push_rstack_result(&pc2))!=OK) return ret;
	if (NOT_PAST_END) return ERR_SYNTAX;
	if (isnull(rstack_ptr->value)) return ERR_INVALID_ARGUMENT;

	/* See if we've got a regexp */
	get_sw_pos(rstack_ptr->value,&slashpos,&regpos);

	/* If we've got a regexp get the dirname */
	if (regpos) {
		if (slashpos==NULL || slashpos > regpos) 
			return ERR_CANT_USE_REGEXP_IN_DIRNAME;

		/* Get dirname */
		if (slashpos) {
			c=*slashpos;
			*slashpos='\0';
			dirname=strdup(rstack_ptr->value);
			*slashpos=c;
			}
		else {
			dirname=strdup(".");
			slashpos=rstack_ptr->value-1;
			}
		reg=slashpos+1;
		}
	else dirname=rstack_ptr->value;

	/* Open dir */
	strip_escapes(dirname);
	if (!(dir=opendir(dirname))) return ERR_CANT_OPEN_DIR;
	}
else if (!(dir=opendir("."))) return ERR_CANT_OPEN_DIR;

/* Loop through it */
result=NULL;
while(ds=(loor_dirent *)readdir(dir)) {
	if (reg && !regexp_match(ds->d_name,reg)) continue;
	append_string(&result,ds->d_name);
	append_string(&result," ");
	}
closedir(dir);

/* Finish off */
push_rstack(result);
FREE(result);
if (regpos) free(dirname);

return OK;
}



/*** Change to or create a directory or create a fifo ***/
int com_chmkdf(int com, int pc)
{
int ret,pc2;
mode_t perm;
char *pathname;

if (pword[pc].end == pc) return ERR_SYNTAX;

/* Get pathname */
pc2=pc+1;
if ((ret=push_rstack_result(&pc2))!=OK) return ret;
if (isnull(rstack_ptr->value)) return ERR_INVALID_ARGUMENT;
pathname=rstack_ptr->value;

/* Get optional permissions */
if (NOT_PAST_END) {
	if (com==CHDIR || strcasecmp(pword[pc2].word,"perm")) 
		return ERR_SYNTAX;
	++pc2;
	if (IS_PAST_END) return ERR_SYNTAX;

	if ((ret=push_rstack_result(&pc2))!=OK) return ret;
	if (NOT_PAST_END) return ERR_SYNTAX;

	if (!get_perm_value(&perm)) return ERR_INVALID_ARGUMENT;
	}
else perm=DEFAULT_DIR_PERMS; /* defined in globals.h */
	
switch(com) {
	case CHDIR:
	if (chdir(pathname) == -1) return ERR_CANT_CHANGE_DIR;
	return OK;

	case MKDIR:
	if (mkdir(pathname,perm) == -1) return ERR_CANT_CREATE_DIR;
	return OK;

	case MKFIFO:
	if (mkfifo(pathname,perm) == -1) return ERR_CANT_CREATE_FIFO;
	return OK;
	}
return ERR_INTERNAL;
}



/*** Create a hard or soft link, move or copy a file. ***/
int com_lcm(int com, int pc)
{
DIR *dir;
loor_dirent *ds;
struct stat fs;
mode_t perm;
int perm_arg,ret,pc2,is_dir,cnt,len;
char *arg1,*arg2,*path,c;
char *slashpos,*regpos,text2[SCRATCH_SIZE];

if (pword[pc].end - pc < 3) return ERR_SYNTAX;

/* Get name of link */
pc2=pc+1;
if ((ret=push_rstack_result(&pc2)) != OK) return ret;
if (IS_PAST_END) return ERR_SYNTAX;
if (isnull(rstack_ptr->value)) return ERR_INVALID_ARGUMENT;
arg1=rstack_ptr->value;

/* "to" noise word */
if (strcasecmp(pword[pc2].word,"to")) return ERR_SYNTAX;

/* Get file */
pc2++;
if (IS_PAST_END) return ERR_SYNTAX;
if ((ret=push_rstack_result(&pc2)) != OK) return ret;
if (isnull(rstack_ptr->value)) return ERR_INVALID_ARGUMENT;
arg2=rstack_ptr->value;

/* Get optional permissions arg (for CP & MV) only */
if (NOT_PAST_END) {
	if (com!=CP && com!=MV) return ERR_SYNTAX;

	if (strcasecmp(pword[pc2].word,"perm") || ++pc2 > pword[pc].end) 
		return ERR_SYNTAX;

	if ((ret=push_rstack_result(&pc2)) != OK) return ret;
	if (NOT_PAST_END) return ERR_SYNTAX;
	if (!get_perm_value(&perm)) return ERR_INVALID_ARGUMENT;
	perm_arg=YES;
	}
else perm_arg=NO;

/* Do the business */
if (is_regexp(arg2)) return ERR_UNEXPECTED_REGEXP;
strip_escapes(arg2);

switch(com) {
	case MKSLINK:
	if (is_regexp(arg1)) return ERR_UNEXPECTED_REGEXP;
	strip_escapes(arg1);
	return (symlink(arg2,arg1) ? ERR_CANT_CREATE_LINK : OK);

	case MKHLINK:
	if (is_regexp(arg1)) return ERR_UNEXPECTED_REGEXP;
	strip_escapes(arg1);
	return (link(arg2,arg1) ? ERR_CANT_CREATE_LINK : OK);
	}

/* MV & CP. */

/* Stat 2nd argument to see if its a filename or dirname */
if (lstat(arg2,&fs) == -1) is_dir=0;
else is_dir = ((fs.st_mode & S_IFMT)==S_IFDIR); 

errno=0;  /* Might have been set by lstat */

/* Get slash and regular expression positions */
len=strlen(arg1);
get_sw_pos(arg1,&slashpos,&regpos);
strip_escapes(arg1);
slashpos-=(len - strlen(arg1));
regpos-=(len - strlen(arg1));

/* If no regexp then its fairly simple */
if (regpos==NULL) {
	/* If 2nd argument is a directory then append name of file onto it
	   since rename() won't do it for us */
	if (is_dir) {
		if (slashpos) append_string(&arg2,slashpos);
		else {
			append_string(&arg2,"/");
			append_string(&arg2,arg1);
			}
		}
	if (com==CP) {
		/* Can't copy a dir */
		if (lstat(arg1,&fs) == -1) return ERR_CANT_STAT_FILE;
		if ((fs.st_mode & S_IFMT)==S_IFDIR) return ERR_CANT_COPY_DIR;

		/* Copy file */
		if (perm_arg) ret=copy_file(arg1,arg2,perm);
		else ret=copy_file(arg1,arg2,fs.st_mode);
		if (ret!=OK) return ret;

		return push_rstack("1");
		}
	if (rename(arg1,arg2)==-1) return ERR_CANT_MOVE_FILE;
	if (perm_arg) chmod(arg2,perm);
	return push_rstack("1");
	}

/* Don't want to overwrite one file with multiple ones */
if (!is_dir) return ERR_CANT_COPY_MOVE_MULTIPLE_ONTO_ONE;

/* Can't handle regexps in any directory names in the path, just 
   too complex. */
if (slashpos && regpos && slashpos > regpos) 
	return ERR_CANT_USE_REGEXP_IN_DIRNAME;

/* Go through everything in the directory that matches the regexp 
   pattern. Get path first/ */
if (slashpos) {
	c=*slashpos;
	*slashpos='\0';
	path=strdup(arg1);
	*slashpos=c;
	}
else {
	path=strdup(".");
	slashpos=arg1-1;
	}

/* Open the path (arg1) and scan through its contents */
cnt=0;
if (!(dir=opendir(path))) {
	free(path);  return ERR_CANT_OPEN_DIR;
	}
while(ds=(loor_dirent *)readdir(dir)) {
	/* Dont try and move or copy "." or ".." */
	if (!strcmp(ds->d_name,".") ||
	    !strcmp(ds->d_name,"..") ||
	    !regexp_match(ds->d_name,slashpos+1)) continue;

	/* Got a match */
	strcpy(text2,arg2);
	if (is_dir) {
		strcat(text2,"/");
		strcat(text2,ds->d_name);
		}
	sprintf(text,"%s/%s",path,ds->d_name);

	/* MV */
	if (com==MV) {
		if (rename(text,text2)==-1) {
			free(path);  return ERR_CANT_MOVE_FILE;
			}
		if (perm_arg) chmod(text2,perm);
		++cnt;
		continue;
		}

	/* CP: Can't copy dir */
	if (lstat(text,&fs) == -1) {
		free(path);  return ERR_CANT_STAT_FILE;
		}
	if ((fs.st_mode & S_IFMT)==S_IFDIR) continue;

	/* Copy file */
	if (perm_arg) ret=copy_file(text,text2,perm);
	else ret=copy_file(text,text2,fs.st_mode);
	if (ret!=OK) {
		free(path);  return ret;
		}
	++cnt;
	}
closedir(dir);
free(path);
sprintf(text,"%d",cnt);
return push_rstack(text);
}



/*** Copy a file from one place to another and set permissions. 
     Used by com_lcm() ***/
int copy_file(char *from, char *to, mode_t perm)
{
FILE *fpf,*fpt;
int c;

if (!(fpf=fopen(from,"r"))) return ERR_CANT_OPEN_FILE;
if (!(fpt=fopen(to,"w"))) return ERR_CANT_OPEN_FILE;
while((c=getc(fpf)) != EOF) putc(c,fpt);
fclose(fpf);
fclose(fpt);

/* Don't error if permissions won't set */
chmod(to,perm);

return OK;
}



/*** Delete a file, directory or link or change an attribute ***/
int com_rm_attrib(int com, int pc)
{
DIR *dir;
loor_dirent *ds;
struct passwd *pwd;
struct group *grs;
mode_t perm;
uid_t user_id;
gid_t group_id;
int ret,pc2,cnt,file_cnt,len,os_err;
char *s,*regpos,*slashpos;
char c,*path;
char *user,*group;

pc2=pc+1;
switch(com) {
	case RM: 
	if (pword[pc].end == pc) return ERR_SYNTAX;
	break;

	case CHMOD:
	case CHOWN: 
	if (pword[pc].end - pc < 2) return ERR_SYNTAX;

	/* Get permissions */
	if ((ret=push_rstack_result(&pc2)) != OK) return ret;
	if (IS_PAST_END) return ERR_SYNTAX;
	if (com==CHMOD) {
		if (!get_perm_value(&perm)) return ERR_INVALID_ARGUMENT;
		break;
		}

	/* CHOWN: Parse user:group parameter. Should either be "<user>",
	   "<user>:", "<user>:<group>" or ":<group>" */
	if (isnull(rstack_ptr->value)) return ERR_INVALID_ARGUMENT;

	/* Find ':' */
	for(s=rstack_ptr->value;*s && *s!=':';++s);

	/* Set strings */
	user=NULL;
	group=NULL;
	if (*s) {
		*s='\0';
		set_string(&user,rstack_ptr->value);
		set_string(&group,s+1);
		*s=':';
		}
	else set_string(&user,rstack_ptr->value);

	/* Can't both be null */
	if (user==NULL && group==NULL) return ERR_INVALID_ARGUMENT;

	/* Get user and group id's */
	user_id=-1;
	group_id=-1;

	if (user) {
		if (number_type(user) == INTEGER) user_id=(uid_t)atoi(user);
		else {
			if ((pwd=getpwnam(user)) == NULL) {
				free(user);  return ERR_NO_SUCH_USER;
				}
			user_id=pwd->pw_uid;
			free(user);
			}
		}

	if (group) {
		if (number_type(group) == INTEGER) group_id=(gid_t)atoi(group);
		else {
			if ((grs=getgrnam(group)) == NULL) {
				free(group);  return ERR_NO_SUCH_GROUP;
				}
			group_id=grs->gr_gid;
			free(group);
			}
		}
	break;

	default: return ERR_INTERNAL;
	}

/* Get file or directory names */
while(NOT_PAST_END) {
	if ((ret=push_rstack_result(&pc2))!=OK) return ret;
	if (isnull(rstack_ptr->value)) return ERR_INVALID_ARGUMENT;

	/* If its not a regular expression then just pass name to remove() 
	   or chmod() */
	if (!is_regexp(rstack_ptr->value)) {
		strip_escapes(rstack_ptr->value);
		switch(com) {
			case RM:
			if (remove(rstack_ptr->value)) 
				return ERR_CANT_DEL_FILE_OR_DIR;
			break;

			case CHMOD:
			if (chmod(rstack_ptr->value,perm))
				return ERR_CANT_CHANGE_FILE_PERM;
			break;
			
			case CHOWN:
			if (chown(rstack_ptr->value,user_id,group_id))
				return ERR_CANT_CHANGE_FILE_OWNERSHIP;
			}
		return push_rstack("1");
		}

	/* See if we have a path rather than just a name. Can't handle
	   regexps in any directory names in the path, just too complex. */
	len=strlen(rstack_ptr->value);
	get_sw_pos(rstack_ptr->value,&slashpos,&regpos);
	strip_escapes(rstack_ptr->value);
	slashpos-=(len - strlen(rstack_ptr->value));
	regpos-=(len - strlen(rstack_ptr->value));

	if (slashpos && regpos && slashpos > regpos) 
		return ERR_CANT_USE_REGEXP_IN_DIRNAME;

	/* If slash is last char then whinge */
	if (slashpos==s-1) return ERR_INVALID_ARGUMENT;

	/* Get path */
	if (slashpos) {
		c=*slashpos;
		*slashpos='\0';
		path=strdup(rstack_ptr->value);
		*slashpos=c;
		}
	else {
		path=strdup(".");
		slashpos=rstack_ptr->value-1;
		}

	/* Open the path and scan through its contents */
	if (!(dir=opendir(path))) {
		free(path);  return ERR_CANT_OPEN_DIR;
		}
	cnt=0;
	file_cnt=0;
	while(ds=(loor_dirent *)readdir(dir)) {
		/* Don't try and remove or chmod "." and ".." */
		if (strcmp(ds->d_name,".") && 
		    strcmp(ds->d_name,"..") &&
		    regexp_match(ds->d_name,slashpos+1)) {
			sprintf(text,"%s/%s",path,ds->d_name);
			file_cnt++;

			/* If unable to complete task don't error cos it'll
			   piss user off, just do what we can */
			switch(com) {
				case RM:
				if (!(os_err=remove(text))) ++cnt;
				break;

				case CHMOD:
				if (!(os_err=chmod(text,perm))) ++cnt;
				break;

				case CHOWN:
				if (!(os_err=chown(text,user_id,group_id))) 
					++cnt;
				}
			}
		}

	/* If we've only found 1 file then return a unix error if there
	   was one */
	if (file_cnt==1 && os_err) {
		switch(com) {
			case RM:    return ERR_CANT_DEL_FILE_OR_DIR;
			case CHMOD: return ERR_CANT_CHANGE_FILE_PERM;
			case CHOWN: return ERR_CANT_CHANGE_FILE_OWNERSHIP;
			}
		}
	free(path);
	closedir(dir);
	}
sprintf(text,"%d",cnt);
return push_rstack(text);
}



/*** Set the umask ***/
int com_umask(int com, int pc)
{
int ret,pc2;
mode_t perm;

if (pc == pword[pc].end) return ERR_SYNTAX;
pc2=pc+1;
if ((ret=push_rstack_result(&pc2)) != OK) return ret;
if (NOT_PAST_END) return ERR_SYNTAX;
if (!get_perm_value(&perm)) return ERR_INVALID_ARGUMENT;

umask(perm);
return OK;
}




/*** Support functions for this module ***/


/*** Check and get valid file permissions. It assumes rstack contains the
     value in question. Used by com_open and com_mkdir ***/
int get_perm_value(mode_t *perm)
{
char *s;

s=rstack_ptr->value;

if (isnull(s) || 
    *s=='-' || 
    strlen(s)!=4 || 
    number_type(s)!=INTEGER ||
    s[0] > '7' || s[1] > '7' || s[2] > '7' || s[3] > '7') return 0;

*perm= ((s[0]-48) << 9) | ((s[1]-48) << 6) | ((s[2]-48) << 3) | (s[3]-48);
return 1;
}



/*** Get the last slash and last regexp positions in a path ***/
void get_sw_pos(char *str,char **sp,char **wp)
{
int escaped=0;

*sp=NULL;
*wp=NULL;
for(;*str;++str) {
	if (*str=='\\') escaped=1;
	else
	/* Get first posn of 1st regexp char and last slash */
	if (*wp==NULL && !escaped && strchr("*?[{",*str))
		*wp=str;
	else
	escaped=0;

	if (*str=='/') *sp=str;
	}
}



/*** Strip regexp escape characters '\' out of a pathname up to just before
     final slash. ***/
void strip_escapes(char *path)
{
char *s,*sp;
int escaped=0;

/* Find position of last dir slash */
for(sp=path+strlen(path)-1;sp >= path && *sp!='/';--sp);

for(s=path;*s && s!=sp;++s) {
	if (*s=='\\') escaped=1;
	else
	if (escaped && strchr(REGEXP_CHARS,*s)) {
		strcpy(s-1,s);  escaped=0;
		}
	else escaped=0;
	}
}
