/***************************************************************************
 FILE: com_lists.c
 LVU : 1.0.0

 DESC:
 This contains the code for all the list processing commands.

 Copyright (C) Neil Robertson 1999, 2000
 ***************************************************************************/


#include "globals.h"


/*** Get the head/rhead or tail/rtail of a string ***/
int com_head_tail(int com, int pc)
{
int ret,pc2;
char *val,*s,*s2,*s3,*es;

if (pword[pc].end - pc < 1) return ERR_SYNTAX;

/* Get string/list */
pc2=pc+1;
if (number_type(pword[pc2].word) != NOT_NUMBER) return ERR_INVALID_ARGUMENT;
if ((ret=push_rstack_result(&pc2)) != OK) return ret;
if ((val=rstack_ptr->value)==NULL) return OK;

/* Get any optional alternate element seperators */
if (NOT_PAST_END) {
	if ((ret=push_rstack_result(&pc2)) != OK) return ret;
	if (NOT_PAST_END) return ERR_SYNTAX;
	es=rstack_ptr->value;
	}
else es=element_seps;

s=val;

/* Head & tail */
if (com==HEAD || com==TAIL) {
	while(*s && is_element_sep(*s,es)) ++s; /* Find start of 1st element */
	val=s;
	while(*s && !is_element_sep(*s,es)) ++s; /* Find end of 1st element */
	if (com==HEAD) {
		*s='\0';  return push_rstack(val);  
		}
	/* Tail */
	if (!*s) return push_rstack(NULL);
	while(*s && is_element_sep(*s,es)) ++s;
	return push_rstack(s);
	}

/* Reverse head & tail */
s2=s+strlen(s)-1;
/* Find end of last element */
while(s2 >= s && is_element_sep(*s2,es)) --s2;
if (s2 < s) return push_rstack(NULL);
s3=s2+1;

/* Find start of last element */
while(s2 >= s && !is_element_sep(*s2,es)) --s2;
if (com==RHEAD) {
	*s3='\0';  return push_rstack(s2+1);
	}

/* Rtail - Find start of 2nd last element */
while(s2 >= s && is_element_sep(*s2,es)) --s2;
if (s2++ < s) return push_rstack(NULL);
*s2='\0';
return push_rstack(s);
}



/*** Get the given element from a list. If the element is not found a null
     string is returned. Element and character position numbers start at 1 ***/
int com_element(int com, int pc)
{
int ret,pc2;
int pos,cnt;
char *list,*es,*s,*s2;

if (pword[pc].end - pc < 2) return ERR_SYNTAX;

/* Get list */
pc2=pc+1;
if ((ret=push_rstack_result(&pc2)) != OK) return ret;
if (isnull(list=rstack_ptr->value)) return OK;
if (IS_PAST_END) return ERR_SYNTAX;

/* Get element number/ character pos and optional element sep. arg. */
if ((ret=push_rstack_result(&pc2)) != OK) return ret;
if (number_type(rstack_ptr->value)!=INTEGER ||
    (pos=atoi(rstack_ptr->value)) < 1) return ERR_INVALID_ARGUMENT;

if (com==CPELEMENT && pos > strlen(list)) return push_rstack(NULL);

if (NOT_PAST_END) {
	if ((ret=push_rstack_result(&pc2)) != OK) return ret;
	if (NOT_PAST_END) return ERR_SYNTAX;
	es=rstack_ptr->value;
	}
else es=element_seps;

if (com==CPELEMENT) {
	/* Find start of element */
	s=list+(pos-1);
	if (is_element_sep(*s,es)) return push_rstack(NULL);
	while(s >= list && !is_element_sep(*s,es)) --s;
	++s;

	/* Find end of element */
	s2=list+pos;
	while (*s2 && !is_element_sep(*s2,es)) ++s2;

	/* Return what we've found */
	*s2='\0';
	return push_rstack(s);
	}

/* ELEMENT command; search through list */
cnt=1;
s=list;
while(*s) {
	/* find start of first/next element */
	while(*s && is_element_sep(*s,es)) ++s;
	if (!*s) return push_rstack(NULL);

	/* Find end of element */
	s2=s;
	while(*s2 && !is_element_sep(*s2,es)) ++s2;
	if (cnt++ == pos) {
		*s2='\0';  return push_rstack(s);
		}	
	s=s2;
	}
return push_rstack(NULL);
}



/*** Return the elements between and including the positions given ***/
int com_midlist(int com, int pc)
{
int ret,pc2,elnum,elcnt,cnt,cnt2;
char *list,*es,*s,*start,*end;

if (pword[pc].end - pc < 3) return ERR_SYNTAX;

pc2=pc+1;

/* Get list */
if ((ret=push_rstack_result(&pc2)) != OK) return ret;
if (isnull(list=rstack_ptr->value)) return OK;
if (IS_PAST_END) return ERR_SYNTAX;

/* Get 1st element number */
if ((ret=push_rstack_result(&pc2)) != OK) return ret;
if (number_type(rstack_ptr->value)!=INTEGER ||
    (elnum=atoi(rstack_ptr->value)) < 1) return ERR_INVALID_ARGUMENT;
if (IS_PAST_END) return ERR_SYNTAX;

/* Get element count */
if ((ret=push_rstack_result(&pc2)) != OK) return ret;
if (number_type(rstack_ptr->value)!=INTEGER ||
    (elcnt=atoi(rstack_ptr->value)) < 1) return ERR_INVALID_ARGUMENT;

/* Get option element seperator */
if (NOT_PAST_END) {
	if ((ret=push_rstack_result(&pc2)) != OK) return ret;
	if (NOT_PAST_END) return ERR_SYNTAX;
	es=rstack_ptr->value;
	}
else es=element_seps;

cnt=0;
cnt2=0;
s=list;
start=NULL;
end=NULL;
while(*s) {
	/* Find start of an element */
	while(*s && is_element_sep(*s,es)) ++s;
	if (++cnt >= elnum && start==NULL) start=s;
	if (!*s) {
		if (end) *end='\0';
		return push_rstack(start);
		}

	/* Find end of element */
	while(*s && !is_element_sep(*s,es)) ++s;
	end=s;
	if (start && ++cnt2==elcnt) {
		*s='\0';  return push_rstack(start);
		}
	}
return push_rstack(start);
}



/*** Returns position of an element in a list if it exists in it else 
     returns 0 ***/
int com_member(int com, int pc)
{
int ret,pc2,cnt;
char *elem,*list,*es;
char *s,*s2,c;

if (pword[pc].end - pc < 2) return ERR_SYNTAX;

pc2=pc+1;

/* Get the list */
if ((ret=push_rstack_result(&pc2)) != OK) return ret;
if (isnull(list=rstack_ptr->value)) return push_rstack(NULL);
if (IS_PAST_END) return ERR_SYNTAX;

/* Get the element and option element sep. arg. */
if ((ret=push_rstack_result(&pc2)) != OK) return ret;
elem=rstack_ptr->value;
if (isnull(elem) || number_type(elem) != NOT_NUMBER)
	return ERR_INVALID_ARGUMENT;
if (NOT_PAST_END) {
	if ((ret=push_rstack_result(&pc2)) != OK) return ret;
	if (NOT_PAST_END) return ERR_SYNTAX;
	es=rstack_ptr->value;
	}
else es=element_seps;

/* Search through list */
cnt=1;
s=list;
while(*s) {
	/* find start of first/next element */
	while(*s && is_element_sep(*s,es)) ++s;
	if (!*s) return push_rstack("0");

	/* Find end of element */
	s2=s;
	while(*s2 && !is_element_sep(*s2,es)) ++s2;
	c=*s2;  *s2='\0';
	if (!strcmp(s,elem)) {	
		*s2=c;
		sprintf(text,"%d",cnt);
		return push_rstack(text);
		}
	*s2=c;
	s=s2;
	cnt++;
	}
return push_rstack("0");
}



/*** Tidy up a list so that only a single supplied seperator is placed
     between each of the elements ***/
int com_tidy(int com, int pc)
{
int ret,pc2,start;
char *list,*sep,*es;
char c,*s,*s2,*result;

if (pword[pc].end - pc < 2) return ERR_SYNTAX;

pc2=pc+1;

/* Get the list */
if ((ret=push_rstack_result(&pc2)) != OK) return ret;
if (isnull(list=rstack_ptr->value)) return push_rstack(NULL);
if (IS_PAST_END) return ERR_SYNTAX;

/* Get the output seperator */
if ((ret=push_rstack_result(&pc2)) != OK) return ret;
sep=rstack_ptr->value;

/* Get optional element seperator */
if (NOT_PAST_END) {
	if ((ret=push_rstack_result(&pc2)) != OK) return ret;
	if (NOT_PAST_END) return ERR_SYNTAX;
	es=rstack_ptr->value;
	}
else es=element_seps;

s=list;
start=1;
result=NULL;
while(1) {
	/* Find start of first/next element */
	while(*s && is_element_sep(*s,es)) ++s;
	if (!*s) break;
	if (!start) append_string(&result,sep);
	else start=0;

	/* Find end of element */
	s2=s;
	while(*s2 && !is_element_sep(*s2,es)) ++s2;
	if (!*s2) {
		append_string(&result,s);  break;
		}
	c=*s2;  *s2='\0';
	append_string(&result,s);
	*s2=c;
	s=s2;
	}
push_rstack(result);
FREE(result);
return OK;
}



/*** Overwrite or insert an element ***/
int com_oielem(int com, int pc)
{
int ret,pc2;
int pos,cnt;
char *list,*elem,*es;
char c,delim[2],*s,*s2,*result;

if (pword[pc].end - pc < 3) return ERR_SYNTAX;

/* Get the list */
pc2=pc+1;
if ((ret=push_rstack_result(&pc2)) != OK) return ret;
if (IS_PAST_END) return ERR_SYNTAX;
list=rstack_ptr->value;

/* Get new element */
if ((ret=push_rstack_result(&pc2)) != OK) return ret;
if (IS_PAST_END) return ERR_SYNTAX;
elem=rstack_ptr->value;

/* Get position */
if ((ret=push_rstack_result(&pc2)) != OK) return ret;
if (number_type(rstack_ptr->value) != INTEGER ||
    (pos=atoi(rstack_ptr->value)) < 1) return ERR_INVALID_ARGUMENT;

/* Get optional element sep. */
if (NOT_PAST_END) {
	if ((ret=push_rstack_result(&pc2)) != OK) return ret;
	if (NOT_PAST_END) return ERR_SYNTAX;
	es=rstack_ptr->value;
	}
else es=element_seps;

/* Get the result */
if (isnull(list)) return push_rstack(elem);
if (isnull(elem)) return push_rstack(list);

cnt=0;
s=list;
while(1) {
	/* Find start of first/next element */
	while(*s && is_element_sep(*s,es)) ++s;
	if (!*s) break;

	/* Find end of element */
	s2=s;
	while(*s2 && !is_element_sep(*s2,es)) ++s2;
	if (++cnt < pos) {  s=s2;  continue;  }

	/* Got required element, create result */
	c=*s;  *s='\0';	
	result=strdup(list);
	*s=c;
	append_string(&result,elem);	
	if (es==NULL) delim[0]=' '; else delim[0]=es[0];
	delim[1]='\0';

	if (com==OVERELEM) {
		if (*s2 && *s2!=delim[0]) append_string(&result,delim);
		append_string(&result,s2);
		}
	else {
		append_string(&result,delim);
		append_string(&result,s);
		}
	goto DONE;
	}

/* Fewer elements in list than given by position value */
if (es==NULL) delim[0]=' '; else delim[0]=es[0];
delim[1]='\0';
result=strdup(list);
append_string(&result,delim);
append_string(&result,elem);

DONE:
push_rstack(result);
free(result);
return OK;
}



/*** Return the number of elements in list (including the null element) 
     or return unique list. ***/
int com_sizeuniq(int com, int pc)
{
int ret,pc2,cnt,dupl;
char *list,*es,*result,delim[2];
char *s,*s2,*s3,*s4,c,c2;

if (pword[pc].end == pc) return ERR_SYNTAX;

/* Get the list */
pc2=pc+1;
if ((ret=push_rstack_result(&pc2)) != OK) return ret;
list=rstack_ptr->value;

/* Get optional element sep. */
if (NOT_PAST_END) {
	if ((ret=push_rstack_result(&pc2)) != OK) return ret;
	if (NOT_PAST_END) return ERR_SYNTAX;
	es=rstack_ptr->value;
	}
else es=element_seps;

if (com==LISTSIZE) {
	/* Do the count */
	if (isnull(list)) return push_rstack("0");

	cnt=0;
	while(*list) {
		/* Find start of an element */
		while(*list && is_element_sep(*list,es)) ++list;
		if (!*list) break;
		++cnt;	

		/* Find end of element */
		while(*list && !is_element_sep(*list,es)) ++list;
		}
	sprintf(text,"%d",cnt);
	return push_rstack(text);
	}

/* UNIQUE */
if (isnull(list)) return push_rstack(NULL);
if (es==NULL) delim[0]=' '; else delim[0]=es[0];
delim[1]='\0';

s=list;
result=NULL;

/* Loop through list */
while(1) {
	/* Find start of first/next element */
	while(*s && is_element_sep(*s,es)) ++s;
	if (!*s) break;

	/* Find end of element */
	s2=s;
	while(*s2 && !is_element_sep(*s2,es)) ++s2;
	c=*s2;  *s2='\0';

	/* Start from top of list again and see if we've already had this 
	   element */
	s3=list;
	dupl=0;
	while(1) {
		while(*s3 && is_element_sep(*s3,es)) ++s3;
		if (s3==s) break;
		s4=s3;
		while(*s4 && !is_element_sep(*s4,es)) ++s4;
		c2=*s4;  *s4='\0';
		if (!strcmp(s,s3)) {  *s4=c2;  dupl=1;  break;  }
		*s4=c2;
		s3=s4;
		}

	/* If not a duplicate add to the result string */
	if (!dupl) {
		append_string(&result,s);
		append_string(&result,delim);
		}
	*s2=c;
	s=s2;
	}
if (result==NULL) return push_rstack(NULL);

/* Get rid of trail delim */
result[strlen(result)-1]='\0';
push_rstack(result);
free(result);
return OK;
}



/*** Sort a list ***/
int com_sort(int com, int pc)
{
int ret,pc2,cnt;
char *es,*list,*elem,*elem_start,*elem_end,*result;
char c,*s,*s2,delim[2];
double elem_val;

if (pword[pc].end == pc) return ERR_SYNTAX;

/* Get the list */
pc2=pc+1;
if ((ret=push_rstack_result(&pc2)) != OK) return ret;
if (isnull(rstack_ptr->value)) return OK;

/* Copy list to temp working area */
list=strdup(rstack_ptr->value);

/* Get optional element sep. */
if (NOT_PAST_END) {
	if ((ret=push_rstack_result(&pc2)) != OK) return ret;
	if (NOT_PAST_END) return ERR_SYNTAX;
	es=rstack_ptr->value;
	}
else es=element_seps;

/* Set delimiter */
if (es==NULL) delim[0]=' '; else delim[0]=es[0];
delim[1]='\0';

/* Sort the list */
result=NULL;
while(1) {
	cnt=0;
	s=list;
	elem=NULL;
	elem_start=NULL;
	elem_end=NULL;

	while(1) {
		/* Find start of first/next element */
		while(*s && is_element_sep(*s,es)) ++s;
		if (!*s) break;

		/* Find end of element */
		s2=s;
		while(*s2 && !is_element_sep(*s2,es)) ++s2;
		c=*s2;  *s2='\0';

		/* Sort/nsort returns a list with minimums -> maximums,
		   rsort/nrsort does the reverse */
		switch(com) {
			case SORT:
			/* Compare element to last element. If its "less" than
			   elem or elem is null then put it as the current 
			   minimum */
			if (elem==NULL || strcoll(s,elem) < 0) break;
			goto CONT;

			case RSORT:
			if (elem==NULL || strcoll(s,elem) > 0) break;
			goto CONT;

			case NSORT:
			if (number_type(s) == NOT_NUMBER) 
				return ERR_INVALID_ARGUMENT;
			if (elem==NULL || atof(s) < elem_val) break;
			goto CONT;

			case RNSORT:
			if (number_type(s) == NOT_NUMBER) 
				return ERR_INVALID_ARGUMENT;
			if (elem==NULL || atof(s) > elem_val) break;
			goto CONT;
			}
		cnt++;
		elem_val=atof(s);
		elem=strdup(s);
		elem_start=s;
		elem_end=s2;

		CONT:
		*s2=c;
		s=s2;
		}

	/* If no new elements we found then we've finished */
	if (!cnt) break;

	/* Got the element, add to result and remove elem from the list */
	append_string(&result,elem);
	append_string(&result,delim);

	/* Remove element from temp list */
	strcpy(elem_start,elem_end);
	}
free(list);
if (result==NULL) return push_rstack(NULL);

/* Remove trailing delimiter */
result[strlen(result)-1]='\0';

/* Push result */
push_rstack(result);
free(result);
return OK;
}



/*** Function for SAMELIST and DIFFLIST commands ***/
int com_sdlist(int com, int pc)
{
int ret,pc2,same;
char *list1,*list2,*es,*result;
char *s1,*s2,*e1,*e2,c1,c2,delim[2];

if (pword[pc].end - pc < 2) return ERR_SYNTAX;

/* Get list 1 */
pc2=pc+1;
if ((ret=push_rstack_result(&pc2)) != OK) return ret;
if (IS_PAST_END) return ERR_SYNTAX;
list1=rstack_ptr->value;

/* Get list 2 */
if ((ret=push_rstack_result(&pc2)) != OK) return ret;
list2=rstack_ptr->value;

if (isnull(list1) || isnull(list2)) return OK;  /* Return empty string */

/* Get optional element seperator */
if (NOT_PAST_END) {
	if ((ret=push_rstack_result(&pc2)) != OK) return ret;
	if (NOT_PAST_END) return ERR_SYNTAX;
	es=rstack_ptr->value;
	}
else es=element_seps;

/* Set delimiter */
if (es==NULL) delim[0]=' '; else delim[0]=es[0];
delim[1]='\0';

/* Go through lists */
s1=list1;
result=NULL;
while(1) {
	/* find start of first/next element */
	while(*s1 && is_element_sep(*s1,es)) ++s1;
	if (!*s1) break;

	/* Find end of element */
	e1=s1;	
	while(*e1 && !is_element_sep(*e1,es)) ++e1;
	c1=*e1;  *e1='\0';

	/* Ok , got an element from list one , now cycle through list 2 */
	same=NO;
	s2=list2;
	while(1) {
		while(*s2 && is_element_sep(*s2,es)) ++s2;
		if (!*s2) break;
		e2=s2;	
		while(*e2 && !is_element_sep(*e2,es)) ++e2;
		c2=*e2;  *e2='\0';

		/* If element from list 1 and list 2 are the same add to the
		   result */
		if (!strcmp(s1,s2)) {
			if (com==SAMELIST) {
				append_string(&result,s1);
				append_string(&result,delim);
				}
			else same=YES;
			*e2=c2;
			break;
			}	
		*e2=c2;
		s2=e2;
		}
	if (com==DIFFLIST && !same) {
		append_string(&result,s1);
		append_string(&result,delim);
		}
	*e1=c1;
	s1=e1;
	}
if (result==NULL) return push_rstack(NULL);

/* Get rid of trail delim */
result[strlen(result)-1]='\0';

push_rstack(result);
free(result);
return OK;
}



/*** Replace list element command ***/
int com_replace(int com, int pc)
{
int ret,pc2,pos,cnt;
char *list,*elem1,*elem2,*es,*result;
char *s,*e,c,delim[2];

if (pword[pc].end - pc < 3) return ERR_SYNTAX;

/* Get list */
pc2=pc+1;
if ((ret=push_rstack_result(&pc2)) != OK) return ret;
if (IS_PAST_END) return ERR_SYNTAX;
list=rstack_ptr->value;

/* Get old element */
if ((ret=push_rstack_result(&pc2)) != OK) return ret;
if (IS_PAST_END) return ERR_SYNTAX;
elem1=rstack_ptr->value;

/* Get new element */
if ((ret=push_rstack_result(&pc2)) != OK) return ret;
elem2=rstack_ptr->value;

pos=1;
es=element_seps;

/* Get optional "pos" and "sep" arguments */
while(NOT_PAST_END) {
	if (!strcasecmp(pword[pc2].word,"pos")) {
		pc2++;
		if (IS_PAST_END) return ERR_SYNTAX;
		if ((ret=push_rstack_result(&pc2)) != OK) return ret;
		if (isnull(rstack_ptr->value) || 
		    (pos=atoi(rstack_ptr->value)) < 1) 
			return ERR_INVALID_ARGUMENT;
		}
	else
	if (!strcasecmp(pword[pc2].word,"sep")) {
		pc2++;
		if (IS_PAST_END) return ERR_SYNTAX;
		if ((ret=push_rstack_result(&pc2)) != OK) return ret;
		es=rstack_ptr->value;
		}
	else return ERR_SYNTAX;
	}

/* Check some values */
if (isnull(elem1)) return ERR_INVALID_ARGUMENT;
if (isnull(list)) return push_rstack(list);

/* Set delimiter */
if (es==NULL) delim[0]=' '; else delim[0]=es[0];
delim[1]='\0';

/* Go through list */
cnt=1;
s=list;
result=NULL;
while(1) {
	/* find start of first/next element */
	while(*s && is_element_sep(*s,es)) ++s;
	if (!*s) break;

	/* Find end of element */
	e=s;	
	while(*e && !is_element_sep(*e,es)) ++e;
	c=*e;  *e='\0';

	if (cnt >= pos && !strcmp(s,elem1)) append_string(&result,elem2);
	else append_string(&result,s);
	append_string(&result,delim);

	*e=c;
	s=e;
	++cnt;
	}
if (result==NULL) return push_rstack(NULL);

/* Get rid of trail delim */
result[strlen(result)-1]='\0';

push_rstack(result);
free(result);
return OK;
}



/*** Join the values or keys of an array together to form a list ***/
int com_join_keys(int com, int pc)
{
int ret,pc2,dummy,len;
VAR_T var,ve;
char *sub,*list,*es,c[2];

if (pword[pc].end == pc) return ERR_SYNTAX;

/* Get array variable */
if ((ret=get_var_main(pword[pc+1].word,pstack_ptr,&var,&sub)) != OK)
	return ret;

/* Get sub var if there is one */
if (sub && 
    (ret=getset_var_element(var,pstack_ptr,&var,NULL,sub,NO,&dummy)) != OK) 
	return ret;

if (!var->dim) return ERR_VAR_IS_NOT_ARRAY;

/* Get optional element seperator. */
if (pword[pc].end - pc > 1) {
	pc2=pc+2;
	if ((ret=push_rstack_result(&pc2)) != OK) return ret;
	if (NOT_PAST_END) return ERR_SYNTAX;
	es=rstack_ptr->value;
	}
else es=NULL;

/* Get element seperators if not given as an argument */
if (!es) {
	if (!element_seps) c[0]=' '; /* default to space */
	else c[0]=element_seps[0];
	c[1]='\0';
	len=1;
	}
else len=strlen(es);

/* Create list */
list=NULL;
FOR_VAR_ELEMENTS(var,ve) {
	append_string(&list,com==JOIN ? ve->value : ve->name);
	if (es) append_string(&list,es);
	else append_string(&list,c);
	}
if (list) list[strlen(list)-len]='\0';

push_rstack(list);
FREE(list);
return OK;
}



/*** Split up a list into its individual elements and store each on in the
     array given. ***/
int com_split(int com, int pc)
{
VAR_T var,arr_elem,new;
char *sub,c1,c2,*s1,*e1,*s2,*e2,*val;
char *es,elem_name[10];
int pc2,ret,dummy,cnt,lcnt;

if (pword[pc].end - pc < 2+(com==KEYSPLIT)) return ERR_SYNTAX;

/* Get array var */
if ((ret=get_var_main(pword[pc+1].word,pstack_ptr,&var,&sub)) != OK)
	return ret;

/* Get sub var if there is one */
if (sub && 
    (ret=getset_var_element(var,pstack_ptr,&var,NULL,sub,NO,&dummy))!= OK) 
	return ret;

if (!var->dim) return ERR_VAR_IS_NOT_ARRAY;

/* Get 1st list */
pc2=pc+2;
if ((ret=push_rstack_result(&pc2)) != OK) return ret;
if (isnull(rstack_ptr->value) && com==SPLIT) return push_rstack("0");
s1=rstack_ptr->value;

/* Get 2nd list if ksplit */
if (com==KEYSPLIT) {
	if ((ret=push_rstack_result(&pc2)) != OK) return ret;
	s2=rstack_ptr->value;
	}
else s2=NULL;

/* Get optional element seperators */
if (NOT_PAST_END) {
	if ((ret=push_rstack_result(&pc2)) != OK) return ret;
	if (NOT_PAST_END) return ERR_SYNTAX;
	es=rstack_ptr->value;
	}
else es=element_seps;

/* Find last element of array where we will start to add stuff in */
for(arr_elem=var->array,cnt=1;arr_elem;arr_elem=arr_elem->next,++cnt) 
	if (arr_elem->next==NULL) {  ++cnt;  break;  }

/* Go through list elements */
lcnt=0;
while((s1 && *s1) || com==KEYSPLIT) {
	/* Find start of element */
	if (s1) {
		while(*s1 && is_element_sep(*s1,es)) ++s1;
		if (!*s1 && com==SPLIT) break;

		/* Find end of element */
		e1=s1;
		while(*e1 && !is_element_sep(*e1,es)) ++e1;
		c1=*e1;  *e1='\0';
		}

	/* For KEYSPLIT find element in next list */
	if (com==KEYSPLIT) {
		if (s2) {
			while(*s2 && is_element_sep(*s2,es)) ++s2;
			if (!*s2) {
				s2=NULL; 
				if (s1==NULL || !*s1) break;
				}
			else {
				e2=s2;
				while(*e2 && !is_element_sep(*e2,es)) ++e2;
				c2=*e2;  *e2='\0';
				}
			}
		else
		if (s1==NULL || !*s1) break;
		}

	/* Add to array */
	sprintf(elem_name,"%d",cnt++);
	if (com==SPLIT) {
		val=s1;
		if ((new=create_variable(elem_name,NULL,NULL,NO,NO))==NULL)
			return ERR_MALLOC;
		}
	else {
		val=s2;
		if (s1 && *s1) new=create_variable(s1,NULL,NULL,NO,NO);
		else new=create_variable(elem_name,NULL,NULL,NO,NO);
		if (new==NULL) return ERR_MALLOC;
		}
	new->dim=var->dim-1;
	if (val) new->value=strdup(val);
	new->next=NULL;

	if (arr_elem) {
		arr_elem->next=new;
		new->prev=arr_elem;
		}
	arr_elem=new;
	if (var->array==NULL) var->array=new;

	/* Replace overwritten char */
	if (com==SPLIT) {
		*e1=c1;  s1=e1;
		}
	else {
		if (s1) {
			*e1=c1;  s1=e1;
			}
		if (s2) {
			*e2=c2;  s2=e2;
			}
		}

	++lcnt;
	}
sprintf(text,"%d",lcnt);
return push_rstack(text);
}
