/***************************************************************************
 FILE: com_loops.c
 LVU : 1.0.0

 DESC:
 This contains the code for the loop and the switch structures.

 Copyright (C) Neil Robertson 1999, 2000
 ***************************************************************************/

#include "globals.h"


/*** Jump to a label ***/
int com_goto(int com, int *pc)
{
LABEL_T lb;
int ret,pc2;

if (pword[*pc].end - *pc < 1) return ERR_SYNTAX;

pc2=*pc+1;
if ((ret=push_rstack_result(&pc2))!=OK) return ret;
if (pc2 <= pword[*pc].end) return ERR_SYNTAX;
if (isnull(rstack_ptr->value)) return ERR_INVALID_ARGUMENT;

/* Find label & jump to it */
FOR_ALL_LABELS(lb) if (!strcmp(lb->name,rstack_ptr->value)) break;
if (lb==NULL) return ERR_UNDEFINED_LABEL;

/* Set to new pos */
*pc=lb->pos;
cleardown_lw_stacks(*pc);
cleardown_rstack();

return OK;
}



/*** LOOP loop ***/
int com_loopcom(int com, int *pc)
{
LOOP_T lp;
int ret,pc2,end,to;

pc2=*pc;
end=pword[pc2].end;
if (end == *pc) return ERR_SYNTAX;
if ((lp=get_loop(pc2))==NULL) return ERR_INTERNAL;

if (!in_loop(lp)) {
	/* Set up count vars */
	pc2++;
	if ((ret=push_rstack_result(&pc2))!=OK) return ret;
	if (pc2 <= end) return ERR_SYNTAX;	
	if (number_type(rstack_ptr->value) != INTEGER ||
	    (to=atoi(rstack_ptr->value)) < 0) return ERR_INVALID_ARGUMENT;

	if (!to) {
		*pc=pword[lp->end].end+1;  return OK;
		}
	push_lstack(lp);
	lstack_ptr->loopcom_to=to;
	lstack_ptr->loopcom_curr=0;
	*pc=end+1;
	return OK;
	}

if (++lstack_ptr->loopcom_curr == lstack_ptr->loopcom_to) {
	pop_lstack();
	*pc=pword[lp->end].end+1;
	}
else *pc=end+1;
return OK;
}



/*** FOR loop ***/
int com_for(int com, int *pc)
{
LOOP_T lp;
int end,ret,pc2;
double var_val,to_val,step_val;
char *var,newval[20];

pc2=*pc;
end=pword[pc2].end;
if (end - pc2 < 5) return ERR_SYNTAX;
if ((lp=get_loop(pc2))==NULL) return ERR_INTERNAL;

/* Check '=' is in the right place */
if (pword[pc2+2].word[0] != '=') return ERR_SYNTAX;

var=pword[pc2+1].word;

/* If we're not in the loop yet set up the from, to and step vals */
if (!in_loop(lp)) {
	/* Get from value & set var */
	pc2+=3;
	if ((ret=push_rstack_result(&pc2))!=OK) return ret;
	if (pc2 >= end) return ERR_SYNTAX;
	if (strcasecmp(pword[pc2].word,"to")) return ERR_SYNTAX;

	if (number_type(rstack_ptr->value) == NOT_NUMBER) 
		return ERR_INVALID_ARGUMENT;

	/* Set var to initial value and return. Use the sprintf to get the
	   noughts after the decimal if its an integer. */
	var_val=atof(rstack_ptr->value);
	if ((ret=set_var_value(var,pstack_ptr,rstack_ptr->value,NO))!=OK)
		return ret;

	/* Get to value */
	++pc2;
	if ((ret=push_rstack_result(&pc2))!=OK) return ret;
	if (number_type(rstack_ptr->value) == NOT_NUMBER) 
		return ERR_INVALID_ARGUMENT;
	to_val=atof(rstack_ptr->value);

	/* Get stepping. Default is 1 or -1 */
	if (var_val < to_val) step_val=1; else step_val=-1;
	if (pc2 <= end) {
		if (pc2==end || 
		    strcasecmp(pword[pc2].word,"step")) return ERR_SYNTAX;
		pc2++;
		if ((ret=push_rstack_result(&pc2))!=OK) return ret;
		if (number_type(rstack_ptr->value) == NOT_NUMBER)
			return ERR_INVALID_ARGUMENT;
		step_val=atof(rstack_ptr->value);

		if (pc2 <= end) return ERR_SYNTAX;
		}

	/* If stepping is going the wrong way then don't bother with loop.
	   Have >= and <= instead of > and < so loop always runs at least
	   once even if to value = from value (but not if stepping is wrong) */
	if ((var_val <= to_val && step_val >= 0) ||
	    (var_val >= to_val && step_val < 0)) {
		push_lstack(lp);
		lstack_ptr->to_val=to_val;
		lstack_ptr->step_val=step_val;
		*pc=end+1;
		}
	else *pc=pword[lp->end].end+1;
	return OK;
	}

/* Get loop variable value */
if ((ret=get_var_value(var,pstack_ptr))!=OK)
	return ret;
if (number_type(rstack_ptr->value) == NOT_NUMBER) 
	return ERR_INVALID_ARGUMENT;

/* Set variable to new value. If its an integer result set variable as 
   such, ie: not .0000 etc on the end. */
var_val=atof(rstack_ptr->value)+lstack_ptr->step_val;
if ((int)var_val == var_val) sprintf(newval,"%d",(int)var_val);
else sprintf(newval,"%f",var_val);
if ((ret=set_var_value(var,pstack_ptr,newval,NO))!=OK) return ret;

/* See if we're done */
if ((lstack_ptr->step_val > 0 && var_val > lstack_ptr->to_val) ||
    (lstack_ptr->step_val < 0 && var_val < lstack_ptr->to_val)) {
	pop_lstack();
	*pc=pword[lp->end].end+1;
	return OK;
	}

*pc=end+1;
return OK;
}




/*** Forent goes through the entries of a 1D array and sets the 
     set loop variables to the array index string and the value 
     respectively ***/
int com_forent(int com, int *pc)
{
LOOP_T lp;
VAR_T v1,v2,v3;
char *sub;
int pc2,ret,dummy;

if (pword[*pc].end - *pc != 4) return ERR_SYNTAX;
if ((lp=get_loop(*pc))==NULL) return ERR_INTERNAL;

/* See if loop already set up */
if (in_loop(lp)) {
	/* If any of the variables have been set to null then exit */
	if (lstack_ptr->forent==NULL || 
	    lstack_ptr->arr_index==NULL ||
	    lstack_ptr->arr_value==NULL) {
		pop_lstack();
		*pc=pword[lp->end].end+1;
		return OK;
		}

	/* Yes , so set variables to next element in the array */
	set_string(&lstack_ptr->arr_index->value,lstack_ptr->forent->name);	
	set_string(&lstack_ptr->arr_value->value,lstack_ptr->forent->value);	
	lstack_ptr->forent=lstack_ptr->forent->next;
	*pc=pword[*pc].end+1;
	return OK;
	}	

pc2=*pc+1;
if (strcasecmp(pword[pc2+2].word,"in")) return ERR_SYNTAX;

/* Set up loop variables */
if ((ret=get_var_main(pword[pc2].word,pstack_ptr,&v1,&sub))!=OK)
	return ret;
if (v1->read_only) return ERR_READONLY_VARIABLE;
if (sub && (ret=getset_var_element(v1,pstack_ptr,&v1,NULL,sub,NO,&dummy))!=OK)
		return ret;
if (v1->read_only) return ERR_READONLY_VARIABLE;

++pc2;
if ((ret=get_var_main(pword[pc2].word,pstack_ptr,&v2,&sub))!=OK)
	return ret;
if (v2->read_only) return ERR_READONLY_VARIABLE;
if (sub && (ret=getset_var_element(v2,pstack_ptr,&v2,NULL,sub,NO,&dummy))!=OK)
		return ret;
if (v1->read_only) return ERR_READONLY_VARIABLE;

pc2+=2;
if ((ret=get_var_main(pword[pc2].word,pstack_ptr,&v3,&sub))!=OK)
	return ret;
if (sub && (ret=getset_var_element(v3,pstack_ptr,&v3,NULL,sub,NO,&dummy))!=OK)
		return ret;

if (!v3->dim) return ERR_VAR_IS_NOT_ARRAY;
if ((v3=v3->array)==NULL) {
	*pc=pword[lp->end].end+1;  return OK;
	}

push_lstack(lp);

/* Set values */
set_string(&v1->value,v3->name);
set_string(&v2->value,v3->value);
lstack_ptr->arr_index=v1;
lstack_ptr->arr_value=v2;
lstack_ptr->forent=v3->next;
*pc=pword[*pc].end+1;
return OK;
}



/*** This goes through a list and loops for each element ***/
int com_forelem(int com, int *pc)
{
LOOP_T lp;
int pc2,ret;
char *var,*s,c;
char *list,*es;

pc2=*pc;
if (pword[pc2].end - pc2 < 3 || strcasecmp(pword[pc2+2].word,"in")) 
	return ERR_SYNTAX;

if ((lp=get_loop(pc2))==NULL) return ERR_INTERNAL;
var=pword[pc2+1].word;

if (in_loop(lp)) {
	s=lstack_ptr->forelem_list;
	if (s==NULL || !*s) goto EXIT_LOOP;

	/* Get head of list */
	while(*s && !is_element_sep(*s,lstack_ptr->forelem_es)) ++s;
	c=*s;  *s='\0';
	if ((ret=set_var_value(var,pstack_ptr,lstack_ptr->forelem_list,NO))!=OK) 
		return ret;
	*s=c;

	/* Point list to next element */
	while(*s && is_element_sep(*s,lstack_ptr->forelem_es)) ++s;
	strcpy(lstack_ptr->forelem_list,s);

	*pc=pword[*pc].end+1;
	return OK;
	}

/* Get the list */
pc2+=3;
if ((ret=push_rstack_result(&pc2))!=OK) return ret;
list=rstack_ptr->value;
	
/* Get optional element seperators */
if (pc2 <= pword[*pc].end) {
	if ((ret=push_rstack_result(&pc2))!=OK) return ret;
	es=rstack_ptr->value;
	}
else es=element_seps;

/* If list is empty don't bother with loop */
if (list == NULL) {
	*pc=pword[lp->end].end+1;  return OK;
	}

/* Find 1st element in list. If nothing there then forget it */
while(*list && is_element_sep(*list,es)) ++list;
if (!*list) {
	*pc=pword[lp->end].end+1;  return OK;
	}

/* Get end of 1st element and set variable */
s=list;
while(*s && !is_element_sep(*s,es)) ++s;
c=*s;  *s='\0';
if ((ret=set_var_value(var,pstack_ptr,list,NO))!=OK) return ret;
*s=c;

/* Point list to next element */
while(*s && is_element_sep(*s,es)) ++s;

/* Create stack entry */
push_lstack(lp);
set_string(&lstack_ptr->forelem_list,s);
set_string(&lstack_ptr->forelem_es,es);

*pc=pword[*pc].end+1;
return OK;

	
EXIT_LOOP:
pop_lstack();
*pc=pword[lp->end].end+1;
return OK;
}



/*** This makes the decisions for the while loop ***/
int com_while(int com, int *pc)
{
LOOP_T lp;
int ret,res;

if (pword[*pc].end - *pc < 1) return ERR_SYNTAX;
if ((lp=get_loop(*pc))==NULL) return ERR_INTERNAL;

if (!in_loop(lp)) push_lstack(lp);

/* If expression is false jump to after WEND statement */
if ((ret=evaluate_complete_boolean(*pc+1,&res))!=OK) return ret;
if (!res) {
	*pc=pword[lp->end].end+1;
	if (lstack_ptr && lstack_ptr->loop==lp) pop_lstack();
	return OK;
	}
*pc=pword[*pc].end+1;
return OK;
}



/*** DO command ***/
int com_do(int com, int pc)
{
LOOP_T lp;

if (pword[pc].end != pc) return ERR_SYNTAX;
if ((lp=get_loop(pc))==NULL) return ERR_INTERNAL;

if (!in_loop(lp)) return push_lstack(lp);
return OK;
}



/*** WHILST command ***/
int com_whilst(int com, int *pc)
{
int res,ret;

if (lstack_ptr==NULL || 
    lstack_ptr->loop->type != DO ||
    lstack_ptr->context != pstack_ptr)
	return ERR_UNEXPECTED_COMMAND;

cleardown_rstack();

if ((ret=evaluate_complete_boolean(*pc+1,&res))!=OK) return ret;
if (res) {
	*pc=lstack_ptr->loop->start;  return OK;
	}	
*pc=pword[*pc].end+1;
pop_lstack();
return OK;
}



/*** SWITCH command ***/
int com_switch(int com, int *pc)
{
LOOP_T lp;
int ret,pc2,pc3;
int val_null,switch_cnt,default_pos;
char *val;

/* Get the switch value */
if (!(pword[*pc].end - *pc)) return ERR_SYNTAX;
pc2=*pc+1;
if ((ret=push_rstack_result(&pc2))!=OK) return ret;
if (pc2 <= pword[*pc].end) return ERR_SYNTAX;

/* Push the loop */
if ((lp=get_loop(*pc))==NULL) return ERR_INTERNAL;
if (!in_loop(lp)) push_lstack(lp);

/* Check val for null */
val=rstack_ptr->value;
if (isnull(val)) val_null=YES; else val_null=NO;
switch_cnt=0;
default_pos=0;

/* Find the value in the following CASE statements */
for(pc2=pword[*pc].end+1;pc2 < lp->end;++pc2) {
	switch(pword[pc2].com) {
		case SWITCH:   switch_cnt++;  continue;
		case SWITCHED: switch_cnt--;  continue;

		case DEFAULT:
		if (switch_cnt) continue;
		if (default_pos) return ERR_UNEXPECTED_COMMAND;
		default_pos=pc2;
		continue;

		case CASE:
		if (switch_cnt) continue;

		/* Loop through case line */
		for(pc3=pc2+1;pc3 <= pword[pc2].end;) {
			if ((ret=push_rstack_result(&pc3))!=OK) return ret;
			if (val_null) {
				if (isnull(rstack_ptr->value)) {
					*pc=pword[pc2].end+1;  return OK;
					}
				}
			else 
			if (rstack_ptr->value && 
			    !strcmp(val,rstack_ptr->value)) {
				*pc=pword[pc2].end+1;
				return OK;
				}
			}
		}
	}

/* If theres a default command jump to there else go to SWITCHED command */
if (default_pos) *pc=default_pos;
else *pc=lp->end;
return OK;
}



/*** Checks to see a CASE or DEFAULT command is not outside a SWITCH "loop" ***/
int com_case_def(int com, int pc)
{
if (lstack_ptr==NULL || 
    lstack_ptr->loop->type != SWITCH ||
    lstack_ptr->context != pstack_ptr)
	return ERR_UNEXPECTED_COMMAND;

if (com==DEFAULT && pword[pc].end != pc) return ERR_SYNTAX;
return OK;
}



/*** BREAK, CONTINUE & SWCONTINUE commands. ***/
int com_break_cont(int com, int *pc)
{
int ret,pc2,pc3,bcnt,cnt;

if (lstack_ptr==NULL) return ERR_UNEXPECTED_COMMAND;
pc2=*pc;

/* Get number of loops to break out from if argument is given */
if (pword[*pc].end > *pc) {
	pc3=*pc+1;
	if ((ret=push_rstack_result(&pc3))!=OK) return ret;
	if (pc3 <= pword[*pc].end) return ERR_SYNTAX;
	if (number_type(rstack_ptr->value) == NOT_NUMBER ||
	    (bcnt=atoi(rstack_ptr->value)) < 1)
		return ERR_INVALID_ARGUMENT;
	}
else bcnt=1;

/* Keep jumping out until we've hit correct amount. If amount given is
   greater than number of nested loops we're in flag an error. */
for(cnt=0;cnt < bcnt;++cnt) {
	if (lstack_ptr==NULL || lstack_ptr->context!=pstack_ptr) {
		if (pword[pc2].end == pc2) return ERR_UNEXPECTED_COMMAND;
		return ERR_VALUE_EXCEEDS_NESTING;
		}

	if (com==BREAK) {
		*pc=pword[lstack_ptr->loop->end].end+1;
		pop_lstack();
		}
	else {
		/* Clear rstack for this proc since we're looping */
		cleardown_rstack();

		/* Dont continue to top of switch statement if its the 
		   normal CONTINUE command */
		if (com==CONTINUE && 
		    lstack_ptr &&
		    lstack_ptr->loop->type==SWITCH) {
			pop_lstack();

			if (lstack_ptr==NULL || 
			    lstack_ptr->context != pstack_ptr) {
				if (pword[pc2].end == pc2) 
					return ERR_UNEXPECTED_COMMAND;
				return ERR_VALUE_EXCEEDS_NESTING;
				}
			}
		if (cnt != bcnt-1) pop_lstack();

		/* If we've emptied the stack already we'll catch it at the 
		   top but do check here or we'll crash if its NULL */
		if (lstack_ptr) *pc=lstack_ptr->loop->start;
		}
	}
return OK;
}



/*** Deals with all loop end commands except for WHILST ***/
int com_loopend(int com, int *pc)
{
WSTACK_T ws,prev;
int type;

if (pword[*pc].end != *pc) return ERR_SYNTAX;

switch(com) {
	case LOOPING : type=LOOP;    break;
	case WEND    : type=WHILE;    break;
	case NEXT    : type=FOR;      break;
	case NEXTENT : type=FORENT;   break;
	case NEXTELEM: type=FORELEM;  break;
	case SWITCHED: type=SWITCH;   break;

	default: return ERR_INTERNAL;
	}

/* Make sure we're in the right loop */
if (lstack_ptr==NULL || 
    lstack_ptr->loop->type != type ||
    lstack_ptr->context != pstack_ptr) 
	return ERR_UNEXPECTED_COMMAND;

/* If SWITCH just exit else jump back to start */
if (com==SWITCHED) {
	pop_lstack();
	*pc=pword[*pc].end+1;
	}
else *pc=lstack_ptr->loop->start;

/* See if we've left a WITH block */
for(ws=wstack_ptr;ws;ws=prev) {
	prev=ws->prev;
        if (ws->context==pstack_ptr && 
	    (*pc < ws->with->start || *pc > ws->with->end)) pop_wstack();
	ws=prev;
	}

/* Clear out result stack for current procedure or it'll just get 
   larger & larger as we loop */
cleardown_rstack();

return OK;
}



/*** Returns yes if we're in the given loop ***/
int in_loop(LOOP_T lp)
{
if (lstack_ptr && lstack_ptr->loop==lp && lstack_ptr->context==pstack_ptr)
	return YES;
return NO;
}
