/***************************************************************************
 FILE: com_maths.c
 LVU : 1.0.0

 DESC:
 This contains the code for the maths commands.

 Copyright (C) Neil Robertson 1999, 2000
 ***************************************************************************/

#include "globals.h"


/* Set up precedence. This is only used in this module so I define it here
   instead of globals.h */
static int prec[127];


/*** Function for EQN command ***/
int com_eqn(int com, int pc)
{
int pc2,ret;
int from;

prec[AND]=4; 
prec[OR]=4;
prec[XOR]=4;
prec[LEFT_SHIFT]=4;  /* << */
prec[RIGHT_SHIFT]=4; /* >> */

prec[POWER]=3;
prec[MOD]=3;

prec[MULTIPLY]=2;
prec[DIVIDE]=2;

prec[PLUS]=1;
prec[MINUS]=1;

prec[GREATER_THAN]=0;
prec[LESS_THAN]=0;
prec[EQUALS]=0;
prec[GREATER_THAN_EQUALS]=0;
prec[LESS_THAN_EQUALS]=0; 
prec[NOT_EQUAL]=0; 

if (pword[pc].end == pc) return ERR_SYNTAX;

for(pc2=pc+1;NOT_PAST_END;++pc2) {
	pword[pc2].processed=NO;
	pword[pc2].result_set=NO;
	pword[pc2].result=0;
	pword[pc2].next_pc=0;
	}

do {
	from=0;

	/* Find an expression surrounded by brackets */
	for(pc2=pc+1;NOT_PAST_END;++pc2) {
		if (pword[pc2].processed) continue;
		if (pword[pc2].word[0]=='(') from=pc2;
		else
		if (pword[pc2].word[0]==')') {
			if (!from) return ERR_SYNTAX;
			if ((ret=evaluate_equation(pc+1,from+1,pc2-1))!=OK)
				return ret;
			pword[from].processed=YES;
			pword[pc2].processed=YES;
			break;
			}
		}
	} while(NOT_PAST_END);

return evaluate_equation(pc+1,pc+1,pword[pc].end);
}



/*** Evaluate a simple mathematical equation (eg: 1+2*4) between round 
     bracket boundaries ***/
int evaluate_equation(int lstart, int from, int to)
{
int pr,pc,pc1,pc2,pc3;
int stage,ret,opcode;
int not1,not3;
int stored;
double val1,val3,result;
char res_str[20];

/* loop through in order of precedence */
for(pc=from;pc <= to;++pc)
	if (!pword[pc].processed) break;
from=pc;

for(pr=4;pr >= 0;--pr) {
	stage=1;
	for(pc=from;pc <= to;) {
		if (pword[pc].processed) {  
			if (pword[pc].word[0]=='{') pc=pword[pc].next_pc;
			else pc++;
			continue;  
			}

		switch(stage) {
			case 1:
			not1=0;

			/* Get 1st value */
			pc1=pc;
			if (pword[pc].result_set) {
				val1=pword[pc].result;
				pc=pword[pc].next_pc;
				}
			else {
				if (pword[pc].opcode) {
					if (pword[pc].opcode==NOT) not1=++pc;
					else
					if (pword[pc].word[0]!='{') 
						return ERR_SYNTAX;
					}

				/* If we're going to run a CALL command
				   which may be a resursive then store results
				   for this proc stack context */
				if (pword[pc].word[0]=='{' &&
				    nested_call_command(pc)) {
					if ((ret=store_in_pstack(lstart))!=OK)
						return ret;
					stored=1;
					}
				else stored=0;

				if ((ret=push_rstack_result(&pc))!=OK) 
					return ret;

				/* Copy results back */
				if (stored) copy_from_pstack(lstart);

				if (number_type(rstack_ptr->value) == NOT_NUMBER)
					return ERR_INVALID_ARGUMENT;

				val1=atof(rstack_ptr->value);
				if (not1) pword[pc1].result=(double)!val1;
				else pword[pc1].result=val1;
				pword[pc1].result_set=YES;
				pword[pc1].next_pc=pc;
				pword[not1].next_pc=pc;
				pop_rstack();
				}
			stage=2;
			continue;


			case 2:
			/* Get operator */
			if (!pword[pc].opcode) return ERR_SYNTAX; 
			opcode=pword[pc].opcode;
			pc2=pc++;
			stage=3;
			continue;


			case 3:
			not3=0;

			/* Get 2nd value */
			pc3=pc;
			if (pword[pc].result_set) {
				val3=pword[pc].result;
				pc=pword[pc].next_pc;
				}
			else {
				if (pword[pc].opcode) {
					if (pword[pc].opcode==NOT) not3=++pc;
					else
					if (pword[pc].word[0]!='{') 
						return ERR_SYNTAX;
					}

				if (pword[pc].word[0]=='{' &&
				    nested_call_command(pc)) {
					if ((ret=store_in_pstack(lstart))!=OK)
						return ret;
					stored=1;
					}
				else stored=0;

				if ((ret=push_rstack_result(&pc))!=OK) 
					return ret;

				if (stored) copy_from_pstack(lstart);

				if (number_type(rstack_ptr->value) == NOT_NUMBER)
					return ERR_INVALID_ARGUMENT;

				val3=atof(rstack_ptr->value);
				if (not3) pword[pc3].result=(double)!val3;
				else pword[pc3].result=val3;

				pword[pc3].result_set=YES;
				pword[pc3].next_pc=pc;
				pword[not3].next_pc=pc;
				pop_rstack();
				}

			if (not1) pword[not1].processed=YES;
			if (not3) pword[not3].processed=YES;

			/* If operator is wrong precedence continue */
			if (prec[opcode] < pr) {
				pc=pc3;  stage=1;  continue;
				}
			pword[pc2].processed=YES;
			pword[pc3].processed=YES;
	
			/* Do appropriate operation */
			switch(opcode) {
				case LESS_THAN:	
				result=(val1 < val3);  break;

				case GREATER_THAN:
				result=(val1 > val3);  break;

				case EQUALS:
				result=(val1 == val3);  break;

				case LESS_THAN_EQUALS:
				result=(val1 <= val3);  break;

				case GREATER_THAN_EQUALS:
				result=(val1 >= val3);  break;

				case NOT_EQUAL:
				result=(val1 != val3);  break;

				case PLUS    : result=val1+val3;  break;
				case MINUS   : result=val1-val3;  break;
				case MULTIPLY: result=val1*val3;  break;
				case DIVIDE  : 
				if (!val3) return ERR_DIVIDE_BY_ZERO;
				result=val1/val3;  break;

				case POWER:
				result=pow((double)val1,(double)val3);  break;

				case MOD: result=(int)val1 % (int)val3;  break;

				case AND: result=(int)val1 & (int)val3;  break;
				case OR : result=(int)val1 | (int)val3;  break;
				case XOR: result=(int)val1 ^ (int)val3;  break;

				case LEFT_SHIFT:
				result=(int)val1 << (int)val3;  break;

				case RIGHT_SHIFT:
				result=(int)val1 >> (int)val3;  break;

				default: return ERR_SYNTAX;
				}

			pword[pc1].result=result;
			pword[pc1].result_set=YES;
			pc=from;
			stage=1;
			}
		}
	if (stage!=2) return ERR_SYNTAX; 
	}

/* If result is an integer set it as such */
if (pword[from].result == (int)pword[from].result)
	sprintf(res_str,"%d",(int)pword[from].result);
else sprintf(res_str,"%f",pword[from].result);

return push_rstack(res_str);
}



/*** Return a not value for numbers and strings ***/
int com_not(int com, int pc)
{
int ret,pc2;

if (pword[pc].end == pc) return ERR_SYNTAX;

pc2=pc+1;
if ((ret=push_rstack_result(&pc2))!=OK) return ret;
if (NOT_PAST_END) return ERR_SYNTAX;
if (number_type(rstack_ptr->value) == NOT_NUMBER) {
	if (isnull(rstack_ptr->value)) return push_rstack("1");
	return push_rstack("0");
	}
if (atof(rstack_ptr->value)) return push_rstack("0");
return push_rstack("1");
}



/*** Increment or decrement a variable ***/
int com_incdec(int com, int pc)
{
int ret,pc2;
double varval,idval;
char *var;

if (pword[pc].end == pc) return ERR_SYNTAX;

/* Get the variables value */
var=pword[pc+1].word;
if (var[0]=='"' || var[0]=='{') return ERR_SYNTAX;
if ((ret=get_var_value(var,pstack_ptr))!=OK) return ret;
if (number_type(rstack_ptr->value) == NOT_NUMBER)
	return ERR_INVALID_ARGUMENT;
varval=atof(rstack_ptr->value);

/* Get optional inc/dec value */
if (pword[pc].end > pc+1) {
	pc2=pc+2;
	if ((ret=push_rstack_result(&pc2))!=OK) return ret;
	if (NOT_PAST_END) return ERR_SYNTAX;
	if (number_type(rstack_ptr->value) == NOT_NUMBER)
		return ERR_INVALID_ARGUMENT;
	idval=atof(rstack_ptr->value);
	}
else idval=1;

varval+=((com==INC || com==INC_SHORTCUT) ? idval : -idval);
if ((int)varval == varval) sprintf(text,"%d",(int)varval);
else sprintf(text,"%f",varval);
if ((ret=set_var_value(var,pstack_ptr,text,NO))!=OK) return ret;
return push_rstack(text);
}



/*** Return a random number between 0 and the number given. ***/
int com_random(int com, int pc)
{
int ret,pc2,max;

if (pword[pc].end == pc) return ERR_SYNTAX;

/* Get the max value */
pc2=pc+1;
if ((ret=push_rstack_result(&pc2))!=OK) return ret;
if (NOT_PAST_END) return ERR_SYNTAX;
if (number_type(rstack_ptr->value) != INTEGER ||
    (max=atoi(rstack_ptr->value)) < 1) return ERR_INVALID_ARGUMENT;

sprintf(text,"%d",random()%max);
return push_rstack(text);
}



/*** Return the maximum or minimum of a set of numbers ***/
int com_minmax(int com, int pc)
{
int ret,pc2,cnt;
double val,result;

if (pword[pc].end - pc < 2) return ERR_SYNTAX;

/* Loop through line arguments */
cnt=0;
for(pc2=pc+1;NOT_PAST_END;) {
	if ((ret=push_rstack_result(&pc2)) != OK) return ret;
	if (number_type(rstack_ptr->value) == NOT_NUMBER) 
		return ERR_INVALID_ARGUMENT;
	val=atof(rstack_ptr->value);

	if (!cnt) result=val;  
	else {
		switch(com) {
			case MINCOM:
			if (val < result) result=val;
			break;

			case MAXCOM:
			if (val > result) result=val;
			break;

			default: return ERR_INTERNAL;
			}
		}
	++cnt;
	}
if (cnt < 2) return ERR_SYNTAX;
if ((int)result == result) sprintf(text,"%d",(int)result);
else sprintf(text,"%f",result);
return push_rstack(text);
}



/*** Do some standard trigonometry stuff ***/
int com_trig(int com, int pc)
{
int ret,pc2;
double num,res,ratio;

if (pword[pc].end == pc) return ERR_SYNTAX;

/* Get value */
pc2=pc+1;
if ((ret=push_rstack_result(&pc2)) != OK) return ret;
if (number_type(rstack_ptr->value) == NOT_NUMBER) return ERR_INVALID_ARGUMENT;
num=atof(rstack_ptr->value);

/* Get optional type. Default is radians. */
if (NOT_PAST_END) {
	if (pc2+1 <= pword[pc].end) return ERR_SYNTAX;
	if (!strcasecmp(pword[pc2].word,"rad")) ratio=1;
	else
	if (!strcasecmp(pword[pc2].word,"deg")) ratio=DEGS_PER_RADIAN;
	else
	return ERR_SYNTAX;
	}
else ratio=1;

/* Asin & acos functions can only accept values from -1 to 1 inclusive */
if ((com==ASIN || com==ACOS) && (num < -1 || num > 1)) 
	return ERR_INVALID_ARGUMENT;

/* Select command */
switch(com) {
	case SIN: res=sin(num / ratio);  break;
	case COS: res=cos(num / ratio);  break;
	case TAN: res=tan(num / ratio);  break;
	case ASIN: res=asin(num) * ratio;  break;
	case ACOS: res=acos(num) * ratio;  break;
	case ATAN: res=atan(num) * ratio;  break;

	default: return ERR_INTERNAL;
	}

/* Return value as an int if its all noughts past the decimal point */
if ((int)res == res) sprintf(text,"%d",(int)res);
else sprintf(text,"%f",res);
return push_rstack(text);
}



/*** Maths operations that take one argument ***/
int com_maths1(int com, int pc)
{
int ret,pc2;
double num,res;

if (pword[pc].end == pc) return ERR_SYNTAX;

/* Get value */
pc2=pc+1;
if ((ret=push_rstack_result(&pc2)) != OK) return ret;
if (NOT_PAST_END) return ERR_SYNTAX;
if (number_type(rstack_ptr->value) == NOT_NUMBER) return ERR_INVALID_ARGUMENT;
num=atof(rstack_ptr->value);

switch(com) {
	case EXP:   res=exp(num);    break;
	case LOG:   res=log(num);    break;
	case LOG10: res=log10(num);  break;
	case SQRT:  res=sqrt(num);   break;
	case CBRT:  res=cbrt(num);   break;
	case ROUND: res=rint(num);   break;
	case CEIL:  res=ceil(num);   break;
	case FLOOR: res=floor(num);  break;

	default: return ERR_INTERNAL;
	}
if ((int)res == res) sprintf(text,"%d",(int)res);
else sprintf(text,"%f",res);
return push_rstack(text);
}



/*** Maths operations that take 2 number arguments ***/
int com_maths2(int com, int pc)
{
int ret,pc2;
int i,num,base,pos,val,last_one_pos;
double num1,num2,res,ratio;
char outstr[33];

if (pword[pc].end - pc < 2) return ERR_SYNTAX;

/* Get 1st value */
pc2=pc+1;
if ((ret=push_rstack_result(&pc2)) != OK) return ret;
if (IS_PAST_END) return ERR_SYNTAX;
switch(number_type(rstack_ptr->value)) {
	case NOT_NUMBER: return ERR_INVALID_ARGUMENT;

	case FLOAT:
	if (com==FROMBASE || com==TOBASE) return ERR_INVALID_ARGUMENT;
	}
num1=atof(rstack_ptr->value);

/* Get 2nd value */
if ((ret=push_rstack_result(&pc2)) != OK) return ret;
if (com != FROMBASE) {
	if (number_type(rstack_ptr->value) == NOT_NUMBER) 
		return ERR_INVALID_ARGUMENT;
	num2=atof(rstack_ptr->value);
	}

/* Get deg or rad if atan2 */
if (NOT_PAST_END) {
	if (com!=ATAN2) return ERR_SYNTAX;
	if (!strcasecmp(pword[pc2].word,"rad")) ratio=1;
	else
	if (!strcasecmp(pword[pc2].word,"deg")) ratio=DEGS_PER_RADIAN;
	else
	return ERR_SYNTAX;
	}

switch(com) {
	case ATAN2: res=atan2(num1,num2) * ratio;  break;
	case HYP  : res=hypot(num1,num2);   break;
	case FROMBASE:
	res=strtol(rstack_ptr->value,NULL,(int)num1);
	break;

	case TOBASE:
	/* Convert to base 2, 4, 8 or 16. Converting to other bases that
	   would involve more than simple bit shifting and and'ing and that
	   are not supported by sprintf is too hard for me to be arsed to do */
	switch(base=(int)num1) {
		case 2:
		case 4:
		outstr[32]='\0';
		last_one_pos=pos=31;
		num=(int)num2;
		for(i=0;i < sizeof(num)*8;i+=(base/2)) {
			val=(num >> i) & (base-1);
			if (val) last_one_pos=pos;
			outstr[pos--]=val + '0';
			}
		return push_rstack(outstr+last_one_pos);


		case 8:
		sprintf(text,"%o",(int)num2);  break;

		case 16:
		sprintf(text,"%X",(int)num2);  break;

		default: return ERR_CANT_CONVERT_TO_BASE;
		}
	return push_rstack(text);

	default: return ERR_INTERNAL;
	}
if ((int)res == res) sprintf(text,"%d",(int)res);
else sprintf(text,"%f",res);
return push_rstack(text);
}
