/***************************************************************************
 FILE: com_misc.c
 LVU : 1.0.0

 DESC:
 This contains stuff that doesn't really belong in any of the other com_*
 modules and don't really warrant a module of their own.

 Copyright (C) Neil Robertson 1999, 2000
 ***************************************************************************/

#include "globals.h"


/*** Print the ascii code of the given character ***/
int com_asc(int com, int pc)
{
int ret,pc2;

if (pword[pc].end == pc) return ERR_SYNTAX;
pc2=pc+1;
if ((ret=push_rstack_result(&pc2))!=OK) return ret;
if (NOT_PAST_END) return ERR_SYNTAX;
if (isnull(rstack_ptr->value) || strlen(rstack_ptr->value)!=1)
	return ERR_INVALID_ARGUMENT;

sprintf(text,"%d",rstack_ptr->value[0]);
return push_rstack(text);
}
 


/*** Force the interpreter to exit. Return the exit code to the OS so it
     can be accessed from the shell. If command is "stop" then call all
     object destructors first ***/
int com_exit(int com, int pc)
{
OBJECT_T obj,next;
PSTACK_T context;
char *code;
int ret,pc2;

if (pword[pc].end == pc) return ERR_SYNTAX;
pc2=pc+1;
if ((ret=push_rstack_result(&pc2))!=OK) return ret;
if (NOT_PAST_END) return ERR_SYNTAX;
code=rstack_ptr->value;
if (number_type(code) != INTEGER)
	return ERR_INVALID_ARGUMENT;

/* Die if EXIT */
if (com==EXIT) system_exit(atoi(code));

/* STOP. Can't use this command if inside a destructor proc or a proc called
   from one or we'll get an infinite loop */
for(context=pstack_ptr;context;context=context->prev) {
	if (!strcmp(context->proc->name,"destructor")) 
		return ERR_CANT_STOP_IN_DESTRUCTOR;
	}

/* Call all the objects destructors in turn and remove from array just in
   case any other code uses it. Unlike in com_delete() it doesn't matter if
   we destroy an object in context because we'll never proceed any further
   beyond the stop command so the problem of "$this->" being invalid won't
   arise. */
for(obj=first_obj;obj;obj=next) {
	next=obj->next;
	if ((ret=call_destructors(obj->class,obj)) != OK) 
		return ret;
	if (destroy_object(obj) != OK) return ERR_INTERNAL;
	}
system_exit(atoi(code));

/* Never get here, but this stops annoying IRIX compiler warning */
return OK;
}



/*** Return the value of an enviroment variable ***/
int com_getenv(int com, int pc)
{
int ret,pc2;

if (pword[pc].end == pc) return ERR_SYNTAX;
pc2=pc+1;
if ((ret=push_rstack_result(&pc2)) != OK) return ret;
if (NOT_PAST_END) return ERR_SYNTAX;
if (isnull(rstack_ptr->value)) return ERR_INVALID_ARGUMENT;
return push_rstack(getenv(rstack_ptr->value));
}



/*** Set an enviroment variable for any subshells that may be exec'd or
     system'ed ***/
int com_setenv(int com, int pc)
{
int ret,pc2;
char *var,*val;

if (pword[pc].end - pc < 2) return ERR_SYNTAX;

/* Get variable name */
pc2=pc+1;
if ((ret=push_rstack_result(&pc2)) != OK) return ret;
if (IS_PAST_END) return ERR_SYNTAX;
if (isnull(rstack_ptr->value)) return ERR_INVALID_ARGUMENT;
var=rstack_ptr->value;

/* Get value */
if ((ret=push_rstack_result(&pc2)) != OK) return ret;
if (NOT_PAST_END) return ERR_SYNTAX;
if (rstack_ptr->value==NULL) val="";
else val=rstack_ptr->value;

/* Set it */
sprintf(text,"%s=%s",var,val);
if (putenv(text)) return ERR_CANT_SET_ENV_VAR;
return OK;
}



/*** Encrypt a string ***/
int com_crypt(int com, int pc)
{
int ret,pc2;
char *salt;

if (pword[pc].end - pc < 2) return ERR_SYNTAX;

/* Get salt */
pc2=pc+1;
if ((ret=push_rstack_result(&pc2)) != OK) return ret;
if (IS_PAST_END) return ERR_SYNTAX;
if (isnull(rstack_ptr->value)) return ERR_INVALID_ARGUMENT;
salt=rstack_ptr->value;

/* Get string */
if ((ret=push_rstack_result(&pc2)) != OK) return ret;
if (NOT_PAST_END) return ERR_SYNTAX;
if (isnull(rstack_ptr->value)) return ERR_INVALID_ARGUMENT;

/* Return encryption */
#ifdef NO_CRYPT
return push_rstack(loor_crypt(rstack_ptr->value,salt));
#else
return push_rstack((char *)crypt(rstack_ptr->value,salt));
#endif
}



#ifdef NO_CRYPT
/*** A simple encrypting function to be used by com_crypt() if standard
     unix crypt() function not available (its often missing with some FreeBSD
     distributions). This is not as good as crypt but its better than naff all.
     It returns a 10 char encrypted word. Str and key can be reversed and the
     function will give the same result. ***/
char *loor_crypt(char *str,char *key)
{
u_char c,prev,*s,*k;
static char output[11];
int pos;

pos=0;
prev='\0';
s=(u_char *)str;
k=(u_char *)key;

do {
	c=(*s + *k + prev)%127;
	if (c<33) c+=33;
	prev=c;
	output[pos]=(char)c;

	if (!*(++s)) s=(u_char *)str;
	if (!*(++k)) k=(u_char *)key;
	} while(++pos<10);

output[pos]='\0';
return output;
}
#endif
