/***************************************************************************
 FILE: com_procs.c
 LVU : 0.8.0

 DESC:
 This contains the code for the CALL and RETURN commands.

 Copyright (C) Neil Robertson 1999
 ***************************************************************************/

#include "globals.h"


/*** Call a procedure ***/
int com_call(int com, int pc)
{
OBJECT_T obj;
PROC_T proc;
VAR_T var,newvar;
PSTACK_T context;
int have_call,pc2,ret,i,j,refvar;
int var_ret,cnt,cnt2,dim,dummy;
char *word,*s;
char *substart;
char vproc_arr_name[100],vproc_cnt[10];

/* The CALL command is optional so see if we have it */
if (com==CALL && strcasecmp(pword[pc].word,"CALL")) have_call=0;
else have_call=1;

if (have_call && pword[pc].end == pc) return ERR_SYNTAX;

/* Get procedure and object proc is in */
pc2=pc+have_call;
if (com==CALL) word=pword[pc2++].word;
else {
	if ((ret=push_rstack_result(&pc2))!=OK) return ret;
	word=rstack_ptr->value;
	if (isnull(word)) return ERR_INVALID_ARGUMENT;
	}
if ((ret=get_proc(word,&proc,&obj)) != OK) return ret;

/* If its a private procedure then we can only call it if 
   we're currently in a procedure belonging to the class or a 
   derived class of the object */
if (obj && 
    proc->private && 
    !is_derived_class(obj->class,pstack_ptr->proc->class))
			return ERR_PRIVATE_PROC;

/* Create proc entry in stack */
if ((ret=push_pstack(proc,pword[pc].end+1,obj))!=OK) return ret;

/* Count parameters of called proc */
for(j=proc->start+2, cnt=0;j <= pword[proc->start].end;++j) 
	if (pword[j].word[0]!='&') cnt++;

cnt2=0;
j=proc->start+2;

/* If vproc check we only have a single 1D array as a param and then set the 
   array up here in case no parameters are passed to the vproc in which case 
   the for loop below won't run */
if (proc->vproc) {
	if (cnt!=1) return ERR_SYNTAX;
	if (pword[j].word[0]=='&') return ERR_REFVAR_CANT_BE_ARRAY;
	for(dim=0, s=pword[j].word;*s;++s) if (*s=='@') ++dim;
	if (dim!=1) return ERR_ARRAY_NOT_1D;

	if ((newvar=create_variable(pword[j].word,NULL,pstack_ptr,NO,NO))==NULL)
		return ERR_MALLOC;

	add_variable(newvar);
	sprintf(vproc_arr_name,"%s[\"1\"]",pword[j].word+1);
	if ((ret=set_var_value(vproc_arr_name,pstack_ptr,"1",NO))!=OK)
		return ret;
	}

/* Set up procedures formal parameters by looping through call line */
for(i=pc2;i <= pword[pc].end;) {
	if (++cnt2 > cnt && !proc->vproc) return ERR_PARAM_COUNT_MISMATCH;

	/* Get possible variable in calling line */
	var_ret=get_var_main(pword[i].word, pstack_ptr->prev,&var,&substart);
	if (substart &&
	    (ret=getset_var_element(var,
	                            pstack_ptr->prev,
	                            &var,NULL,substart,NO,&dummy))!=OK)
		return ret;

	/* If call proc param is a reference variable check we're giving it 
	   a variable to reference! */
	if (pword[j].word[0]=='&') {  
		if (var_ret!=OK) {
			if (var_ret==ERR_UNDEFINED_VARIABLE) 
				return ERR_CANT_REF_NON_VAR;
			return var_ret;
			}
		if (pword[i].opcode) return ERR_INVALID_ARGUMENT;
		refvar=1;  
		++j;  
		++i;
		}
	else {
		refvar=0;
		context=pstack_ptr;	
		pstack_ptr=pstack_ptr->prev;
		if ((ret=push_rstack_result(&i))!=OK) return ret;
		pstack_ptr=context;
		}

	/* Set the array entry if vproc. Array will already have been created
	   just befor this loop */
	if (proc->vproc) {
		sprintf(vproc_arr_name,"%s[\"%d\"]",pword[j].word+1,cnt2+1);
		if ((ret=set_var_value(vproc_arr_name,
		                       pstack_ptr,
		                       rstack_ptr->value,NO))!=OK)
			return ret;
		continue;
		}

	/* Create command line parameter var. */
	for(dim=0, s=pword[j].word;*s;++s) if (*s=='@') ++dim;
	if (dim && refvar) return ERR_REFVAR_CANT_BE_ARRAY;
	if ((ret=check_variable_name(pword[j].word,obj,pstack_ptr))!=OK)
		return ret;

	if ((newvar=create_variable(pword[j].word,NULL,pstack_ptr,NO,NO))==NULL)
		return ERR_MALLOC;
	add_variable(newvar);

	++j;

	/* If param is reference variable set pointer */
	if (var) while(var->refvar) var=var->refvar;
	if (refvar) {
		newvar->refvar=var;  continue;
		}

	/* If passed var check dimensions are the same */
	if (var && dim != var->dim) return ERR_ARRAY_DIM_MISMATCH;

	/* Set new variables value if not array */
	if (!dim || var==NULL || (var && !var->dim)) {
		set_string(&newvar->value,rstack_ptr->value);
		continue;
		}

	/* If formal param is defined as an array copy over array contents. */
	if (dim && (ret=copy_variable(var,newvar,NO))!=OK) return ret;
	}

/* If vproc set 1st element of the array to a count of the number of
   elements in it (including 1st element) */
if (proc->vproc) {
	sprintf(vproc_arr_name,"%s[\"1\"]",pword[j].word+1);
	sprintf(vproc_cnt,"%d",cnt2+1);
	if ((ret=set_var_value(vproc_arr_name, pstack_ptr, vproc_cnt,NO))!=OK)
		return ret;
	}
else
if (cnt2 != cnt) return ERR_PARAM_COUNT_MISMATCH;

/* Jump to the procedure */
return mainloop(YES,proc->start);
}



/*** Return from a procedure ***/
int com_return(int com, int pc)
{
int ret,pc2;

if (pword[pc].end == pc) return push_rstack(NULL);
pc2=pc+1;
if ((ret=push_rstack_result(&pc2))!=OK) return ret; 
if (NOT_PAST_END) return ERR_SYNTAX;
return OK;
}
