/***************************************************************************
 FILE: com_signals.c
 LVU : 1.0.0

 DESC:
 This contains the code for the signal commands.

 Copyright (C) Neil Robertson 1999, 2000
 ***************************************************************************/

#include "globals.h"


/*** Set up what to do when a signal is received ***/
int com_signal(int com, int pc)
{
PROC_T proc;
OBJECT_T obj;
SIGHANDLER_T sh,new;
int ret,pc2,i;

if (pword[pc].end - pc < 2) return ERR_SYNTAX;

/* Get signal to handle */
pc2=pc+1;
if ((ret=push_rstack_result(&pc2)) != OK) return ret;
if (isnull(rstack_ptr->value)) return ERR_INVALID_ARGUMENT;
if (IS_PAST_END) return ERR_SYNTAX;

for(i=0;i < NUM_SIGNALS;++i) 
	if (!strcasecmp(siglist[i].name,rstack_ptr->value)) break;
if (i==NUM_SIGNALS) return ERR_UNKNOWN_SIGNAL;

/* Get procedure to call. If empty string then reset to default behaviour */
if ((ret=push_rstack_result(&pc2)) != OK) return ret;
if (NOT_PAST_END) return ERR_SYNTAX;

if (!isnull(rstack_ptr->value)) {
	if ((ret=get_proc(rstack_ptr->value,&proc,&obj)) != OK) return ret;

	/* Can't call any private procs since we have no idea what object 
	   context we'll be in when the signal handler is called */
	if (obj && proc->private) return ERR_PRIVATE_PROC;

	/* Can't have any procedure parameters */
	if (pword[proc->start].end - proc->start > 1) return ERR_HANDLER_ARGS;
	}
else {
	/* Delete from handlers list */
	FOR_ALL_SIGHANDLERS(sh)
		if (sh->signum==siglist[i].signum) break;
	if (sh==NULL) return OK;

	if (sh->prev) sh->prev->next=sh->next;
	if (sh->next) sh->next->prev=sh->prev;
	if (sh==sighandler_end) sighandler_end=sh->prev;
	if (sh==sighandler_start) sighandler_start=sh->next;

	set_sig_default(sh->signum);

	free(sh);
	return OK;
	}

/* Add data to sighandler list. See if we've got an entry for the given
   signal already first. */
FOR_ALL_SIGHANDLERS(sh)
	if (sh->signum==siglist[i].signum) break;

if (sh) {
	/* Return last procedure that was set for this signal */
	strcpy(text,sh->handler->name);
	sh->handler=proc;  
	return push_rstack(text);
	}

/* Add new entry to list */
if ((new=(SIGHANDLER_T)malloc(sizeof(struct sighandler))) == NULL)
	return ERR_MALLOC;

if (sighandler_end) {
	sighandler_end->next=new;
	new->prev=sighandler_end;
	}
else {
	sighandler_start=new;
	new->prev=NULL;
	}
sighandler_end=new;
new->next=NULL;
new->signum=siglist[i].signum;
new->handler=proc;
new->siglist_index=i;
signal(new->signum,signal_handler);

return push_rstack(NULL);
}



/*** Set up an alarm timer ***/
int com_alarm(int com, int pc)
{
struct itimerval tim;
int ret,pc2;
int secs,usecs;

if (pword[pc].end == pc) return ERR_SYNTAX;

pc2=pc+1;
if ((ret=push_rstack_result(&pc2)) != OK) return ret;
if (number_type(rstack_ptr->value) == NOT_NUMBER ||
    (secs=atoi(rstack_ptr->value)) < 0)
	return ERR_INVALID_ARGUMENT;

if (NOT_PAST_END) return ERR_SYNTAX;
usecs=get_usecs(rstack_ptr->value);

tim.it_value.tv_sec=secs;
tim.it_value.tv_usec=usecs;
tim.it_interval.tv_sec=((com == CALARM) ? secs : 0 );
tim.it_interval.tv_usec=((com == CALARM) ? usecs : 0);
setitimer(ITIMER_REAL,&tim,NULL);

return OK;
}



/*** Send a signal to another process ***/
int com_sendsig(int com, int pc)
{
int ret,pc2,i;
pid_t pid;

if (pword[pc].end - pc < 3) return ERR_SYNTAX;

/* Get signal */
pc2=pc+1;
if ((ret=push_rstack_result(&pc2)) != OK) return ret;
if (IS_PAST_END) return ERR_SYNTAX;
if (isnull(rstack_ptr->value)) return ERR_INVALID_ARGUMENT;

for(i=0;i < NUM_SIGNALS;++i) 
	if (!strcasecmp(siglist[i].name,rstack_ptr->value)) break;
if (i==NUM_SIGNALS) return ERR_UNKNOWN_SIGNAL;

/* Noise word */
if (strcasecmp(pword[pc2].word,"to")) return ERR_SYNTAX;
if (++pc2 > pword[pc].end) return ERR_SYNTAX;

/* Get process pid */
if ((ret=push_rstack_result(&pc2)) != OK) return ret;
if (NOT_PAST_END) return ERR_SYNTAX;
if (number_type(rstack_ptr->value) != INTEGER) return ERR_INVALID_ARGUMENT;
pid=(pid_t)atoi(rstack_ptr->value);

/* Send the signal */
return (!kill(pid,siglist[i].signum) ? OK : ERR_FAILED_TO_SEND_SIGNAL);
}
