/***************************************************************************
 FILE: com_stacks.c
 LVU : 1.0.0

 DESC:
 This implements user level stacks. They have nothing to do with the code
 level stacks implemented in stacks.c

 Copyright (C) Neil Robertson 1999, 2000
 ***************************************************************************/

#include "globals.h"


/*** Push a value onto the top of the stack ***/
int com_push(int com, int pc)
{
USTACK_T new,stk;
char *stack;
int ret,pc2;

if (pword[pc].end - pc < 2) return ERR_SYNTAX;

/* Get stack name */
pc2=pc+1;
if ((ret=push_rstack_result(&pc2)) != OK) return ret;
if (IS_PAST_END) return ERR_SYNTAX;
if (isnull(rstack_ptr->value)) return ERR_INVALID_ARGUMENT;
stack=rstack_ptr->value;

/* Get data to push */
if ((ret=push_rstack_result(&pc2)) != OK) return ret;
if (NOT_PAST_END) return ERR_SYNTAX;

/* See if the stack already exists. If not , create a new one */
FOR_USER_STACK(stk) if (!strcmp(stk->value,stack)) break;

if (stk==NULL) {
	if ((new=(USTACK_T)malloc(sizeof(struct ustack)))==NULL)
		return ERR_MALLOC;
	if (ustack_start) ustack_end->next=new;
	else ustack_start=new;
	ustack_end=new;
	new->value=strdup(stack);
	new->next=NULL;
	stk=new;

	new->prev=NULL;  /* Not used on 1st level */
	}

/* Create entry in stack */
if ((new=(USTACK_T)malloc(sizeof(struct ustack)))==NULL)
	return ERR_MALLOC;
new->start=NULL; /* Not used on 2nd level */
new->end=NULL;

if (com==PUSH) {
	if (stk->start) {
		stk->end->next=new;
		new->prev=stk->end;
		}
	else {
		stk->start=new;
		new->prev=NULL;
		}
	new->next=NULL;
	stk->end=new;
	}
else {
	/* BPUSH */
	if (stk->start) {
		stk->start->prev=new;
		new->next=stk->start;
		}
	else {
		stk->end=new;
		new->next=NULL;
		}
	new->prev=NULL;
	stk->start=new;
	}

new->value=NULL;
set_string(&new->value,rstack_ptr->value);

return OK;
}



/*** Pop a value off the top. If we empty the given stack don't delete
     it as it will probably be used again. ***/
int com_pop(int com, int pc)
{
USTACK_T stk;
int ret,pc2;

if (pword[pc].end == pc) return ERR_SYNTAX;

pc2=pc+1;
if ((ret=push_rstack_result(&pc2)) != OK) return ret;
if (NOT_PAST_END) return ERR_SYNTAX;
if (isnull(rstack_ptr->value)) return ERR_INVALID_ARGUMENT;

/* Find the stack */
FOR_USER_STACK(stk)
	if (!strcmp(rstack_ptr->value,stk->value)) break;
if (stk==NULL) return ERR_UNDEFINED_STACK;

/* See if the stack is empty , if so return the empty marker */
if (stk->start==NULL) return push_rstack(empty_stack_marker);

if (com==POP) {
	push_rstack(stk->end->value);
	FREE(stk->end->value);
	if (stk->end->prev) {
		stk->end->prev->next=NULL;
		stk->end=stk->end->prev;
		}
	else {
		stk->end=NULL;
		stk->start=NULL;
		}
	}	
else {
	/* BPOP */
	push_rstack(stk->start->value);
	FREE(stk->start->value);
	if (stk->start->next) {
		stk->start->next->prev=NULL;
		stk->start=stk->start->next;
		}
	else {
		stk->end=NULL;
		stk->start=NULL;
		}
	}
return OK;
}



/*** Clear out user stack(s). This doesn't actually delete the stack
     though.  ***/
int com_wipestack(int com, int pc)
{
USTACK_T stk,stkent,next;
int ret,pc2,cnt;

if (pword[pc].end == pc) return ERR_SYNTAX;

cnt=0;
pc2=pc+1;
while(NOT_PAST_END) {
	/* Get the stack */
	if ((ret=push_rstack_result(&pc2)) != OK) return ret;
	if (isnull(rstack_ptr->value)) return ERR_INVALID_ARGUMENT;

	FOR_USER_STACK(stk)
		if (!strcmp(rstack_ptr->value,stk->value)) break;
	if (stk==NULL) return ERR_UNDEFINED_STACK;

	/* Clear out its entries */
	for(stkent=stk->start;stkent;stkent=next) {
		next=stkent->next;
		FREE(stkent->value);
		free(stkent);
		++cnt;
		}
	stk->start=stk->end=NULL;
	}
sprintf(text,"%d",cnt);
return push_rstack(text);
}
