/***************************************************************************
 FILE: com_strings.c
 LVU : 1.0.0

 DESC:
 This contains the code for the string commands. This does not include the
 list commands which have their own module.

 Copyright (C) Neil Robertson 1999
 ***************************************************************************/


#include "globals.h"


/*** Add strings together ***/
int com_addstr(int com, int pc)
{
int ret,pc2,cnt;
char *result;

if (pword[pc].end - pc < 2) return ERR_SYNTAX;

cnt=0;
result=NULL;

/* Loop through args */
for(pc2=pc+1;NOT_PAST_END;) {
	if (number_type(pword[pc2].word) != NOT_NUMBER) 
		return ERR_INVALID_ARGUMENT;

	if ((ret=push_rstack_result(&pc2))!=OK) return ret;	
	if ((ret=append_string(&result,rstack_ptr->value))!=OK) return ret;
	++cnt;
	}

/* Must have had at least 2 arguments */
if (cnt < 2) {
	FREE(result);  return ERR_SYNTAX;
	}

push_rstack(result);
FREE(result);
return OK;
}



/*** String commands that take 1 argument ***/
int com_strfuncs1(int com, int pc)
{
int ret,pc2,last,cnt;
char *s,*s2,*result,c2[2];

if (pword[pc].end == pc) return ERR_SYNTAX;

/* Get string */
pc2=pc+1;
if ((ret=push_rstack_result(&pc2))!=OK) return ret;
if (NOT_PAST_END) return ERR_SYNTAX;

switch(com) {
	case STRTYPE:
	switch(number_type(rstack_ptr->value)) {
		case NOT_NUMBER: return push_rstack("STRING");
		case INTEGER   : return push_rstack("INTEGER");
		case FLOAT     : return push_rstack("FLOAT");
		}
	return ERR_INTERNAL;


	case STRLEN:
	if (isnull(rstack_ptr->value)) return push_rstack("0");
	sprintf(text,"%d",strlen(rstack_ptr->value));
	return push_rstack(text);


	case UNIQSTR:
	result=NULL;
	if (isnull(rstack_ptr->value)) return push_rstack(NULL);

	/* Always have the 1st char in the result */
	c2[0]=*rstack_ptr->value;
	c2[1]='\0';
	append_string(&result,c2);

	/* Loop through string */
	for(s=rstack_ptr->value+1;*s;++s) {
		for(s2=result;*s2;++s2) if (*s2 == *s) break;
		if (!*s2) {
			c2[0]=*s;  c2[1]='\0';  append_string(&result,c2);
			}
		}
	push_rstack(result);
	FREE(result);
	return OK;


	case NLRM:
	if (isnull(rstack_ptr->value)) return OK;

	/* Terminate string at 1st \n found */
	for(s=rstack_ptr->value+strlen(rstack_ptr->value)-1;
	    s >= rstack_ptr->value; --s) {
		if (*s=='\n') {  *s='\0';  break;  }
		}
	return OK;


	case CHOP:
	case CHOMP:
	if (isnull(rstack_ptr->value)) return OK;
	last=strlen(rstack_ptr->value)-1;
	if (com==CHOP || rstack_ptr->value[last]=='\n')
		rstack_ptr->value[strlen(rstack_ptr->value)-1]='\0';
	return OK;


	case REVSTR:
	if (isnull(rstack_ptr->value)) return OK;
	cnt=0;
	result=NULL;
	s=rstack_ptr->value+strlen(rstack_ptr->value)-1;
	while(s >= rstack_ptr->value) append_char(&result,*(s--),cnt++); 
	append_char(&result,'\0',cnt);
	push_rstack(result);
	free(result);
	return OK;
	}
return ERR_INTERNAL;
}



/*** String commands that take 2 arguments ***/
int com_strfuncs2(int com, int pc)
{
int end,ret,len;
char *str,*str2,*result;
char *s,*s2,*r,c,c2[2];

if ((end=pword[pc].end) - pc < 2) return ERR_SYNTAX;

/* Get string */
++pc;
if ((ret=push_rstack_result(&pc))!=OK) return ret;
if (pc > end) return ERR_SYNTAX;
if (isnull(rstack_ptr->value)) return OK; /* Leave null as result */
str=rstack_ptr->value;

/* Get length or string2 */
if ((ret=push_rstack_result(&pc))!=OK) return ret;
if (pc <= end) return ERR_SYNTAX;
switch(com) {
	case LEFTSTR:
	case RIGHTSTR:
	if (number_type(rstack_ptr->value) != INTEGER ||
	    (len=atoi(rstack_ptr->value)) < 1) return ERR_INVALID_ARGUMENT;
	break;

	case SAMESTR:
	if (isnull(rstack_ptr->value)) return OK;
	break;

	case DIFFSTR:
	if (isnull(rstack_ptr->value)) return push_rstack(str);
	}

/* Get the answer */
s=str;
str2=rstack_ptr->value;
switch(com) {
	case LEFTSTR:
	while(*s && s-str!=len) ++s;
	c=*s;  *s='\0';
	push_rstack(str);
	*s=c;
	return OK;

	case RIGHTSTR: 
	s=s2=str+strlen(str);
	while(s2-s!=len && s!=str) --s;
	return push_rstack(s);	

	case MATCHSTR:
	case MATCHSTR_SHORTCUT:
	if (regexp_match(str,str2)) return push_rstack("1");
	return push_rstack("0");

	case SAMESTR:
	c2[1]='\0';
	result=NULL;
	for(s=str;*s;++s) {
		for(s2=str2;*s2;++s2) {
			if (*s2==*s) {
				c2[0]=*s;
				if (result) {
					/* Make sure we don't repeat a
					   character in the output */
					for(r=result;*r;++r) 
						if (*r==*s) break;
					if (!*r) append_string(&result,c2);
					}
				else append_string(&result,c2);
				break;
				}
			}
		}
	push_rstack(result);
	FREE(result);
	return OK;

	case DIFFSTR:
	c2[1]='\0';
	result=NULL;
	for(s=str;*s;++s) {
		for(s2=str2;*s2;++s2) if (*s2==*s) break;
		if (!*s2) {
			c2[0]=*s;
			if (result) {
				/* Don't repeat chararacters */
				for(r=result;*r;++r) 
					if (*r==*s) break;
				if (!*r) append_string(&result,c2);
				}
			else append_string(&result,c2);
			}
		}
	push_rstack(result);
	FREE(result);
	return OK;
	}
return ERR_INTERNAL;
}



/*** upperstr & lowerstr commands ***/
int com_ulstr(int com, int pc)
{
int end,ret,pos,len,cnt;
char *str,*s;

if ((end=pword[pc].end) - pc < 1) return ERR_SYNTAX;

/* Get string */
++pc;
if ((ret=push_rstack_result(&pc))!=OK) return ret;
if (isnull(rstack_ptr->value)) return OK; /* Leave null as result */
str=rstack_ptr->value;

/* Get start position in string */
pos=1;
len=-1;
if (pc <= end) {
	if ((ret=push_rstack_result(&pc))!=OK) return ret;
	if (number_type(rstack_ptr->value) != INTEGER ||
	    (pos=atoi(rstack_ptr->value)) < 1) return ERR_INVALID_ARGUMENT;
	if (pos > strlen(str)) return push_rstack(str);

	/* Get length */
	if (pc <= end) {
		if ((ret=push_rstack_result(&pc))!=OK) return ret;
		if (pc <= end) return ERR_SYNTAX;
		if (number_type(rstack_ptr->value) != INTEGER ||
		    (len=atoi(rstack_ptr->value)) < 1) return ERR_INVALID_ARGUMENT;
		}
	}

/* Go through string */
for(s=str+(pos-1),cnt=0;*s && (cnt < len || len==-1);++s,++cnt) {
	if (com==UPPERSTR) *s=toupper(*s);
	else *s=tolower(*s);
	}
return push_rstack(str);
}



/*** Midstr command ***/
int com_midstr(int com, int pc)
{
int end,ret,pos,len;
char *str,c;

if ((end=pword[pc].end) - pc < 2) return ERR_SYNTAX;

/* Get string */
++pc;
if ((ret=push_rstack_result(&pc))!=OK) return ret;
if (pc > end) return ERR_SYNTAX;
if (isnull(rstack_ptr->value)) return OK; /* Leave null as result */
str=rstack_ptr->value;

/* Get position in string */
if ((ret=push_rstack_result(&pc))!=OK) return ret;
if (number_type(rstack_ptr->value) != INTEGER ||
    (pos=atoi(rstack_ptr->value)) < 1) return ERR_INVALID_ARGUMENT;
if (pos > strlen(str)) return push_rstack(NULL);

/* Get optional length. Defaults to length of string - pos */
if (pc <= end) {
	if ((ret=push_rstack_result(&pc))!=OK) return ret;
	if (pc <= end) return ERR_SYNTAX;
	if (number_type(rstack_ptr->value) != INTEGER ||
	    (len=atoi(rstack_ptr->value)) < 1) return ERR_INVALID_ARGUMENT;
	if (pos+len >= strlen(str)) push_rstack(str+(pos-1));
	}
else len=strlen(str)-pos+1;

c=str[pos+len-1]; 
str[pos+len-1]='\0';
push_rstack(str+(pos-1));
str[pos+len-1]=c;
return OK;
}



/*** Format an output string. Deal with % operators \ escape codes and
     colour codes. ***/
int com_format(int com, int pc)
{
int ret,end,pos,dot,star;
char *s,*s2,*out,ftype,c;
char *format,*arg1,col[3];
int i,arg1i,arg2,arg3;
int len,had_colour;
double arg1f;

if ((end=pword[pc].end) - pc < 1) return ERR_SYNTAX;

/* Get format string */
++pc;
if ((ret=push_rstack_result(&pc))!=OK) return ret;
if (isnull(rstack_ptr->value)) return OK;

/* Loop through it and add in the following args */
pos=0;
out=NULL;
had_colour=0;
s=rstack_ptr->value;
while(*s) {
	if (pos == SCRATCH_SIZE-1) {
		text[pos]='\0';
		append_string(&out,text);
		pos=0;
		}

	switch(*s) {
		default:  
		/* Send a reset because some terminals continue
		   the settings onto the next line */
		if (*s=='\n' && had_colour) {
			memcpy(text+pos,COL_RESET,4);
			had_colour=0;
			pos+=4;
			}
		text[pos]=*s;  
		break;


		case '%':
		/* Clear out text to use with sprintf */
		if (pos) {
			text[pos]='\0';
			append_string(&out,text);
			pos=0;
			}

		s2=s+1;
		if (*s2=='%') {  text[0]='%';  break; }

		/* If we haven't got any command arguments left flag an 
		   error */
		if (pc > end) return ERR_MISSING_FORMAT_ARG;

		/* Find end of '%' code */
		dot=0;
		star=0;
		if (*s2=='-' || *s2=='+') ++s2;
		while(*s2 && (*s2=='.' || *s2=='*' || isdigit(*s2))) {
			if (*s2=='.') dot++;
			else
			if (*s2=='*') {
				if (star && !dot) 
					return ERR_INVALID_FORMAT_CODE;
				star++;
				}
			++s2;
			}
		if (!(*s2) || dot>1 || star>2 || (star > 1 && !dot)) 
			return ERR_INVALID_FORMAT_CODE;

		/* Save the format code */
		ftype=*s2;
		++s2;  c=*s2;  *s2='\0';	
		format=strdup(s);
		*s2=c;  s=s2;

		if ((ret=push_rstack_result(&pc))!=OK) return ret;
		arg1=rstack_ptr->value;

		/* If we've got a star-dot or dot-star format (eg: %*.*s) we 
		   need an extra 1 or 2 command args */
		for(i=0;i < star;++i) {
			if (pc > end) return ERR_MISSING_FORMAT_ARG;
			if ((ret=push_rstack_result(&pc))!=OK) return ret;
			if (number_type(rstack_ptr->value) != INTEGER)
				return ERR_INVALID_ARGUMENT;
			if (!i) arg2=atoi(rstack_ptr->value);
			else arg3=atoi(rstack_ptr->value);
			}

		/* Get the result */
		switch(ftype) {
			case 'c':
			case 'd':
			case 'u':
			case 'o':
			case 'x': 
			case 'X': 
			if (number_type(arg1) != INTEGER)
				return ERR_INVALID_ARGUMENT;
			arg1i=atoi(arg1);
			switch(star) {
				case 0: sprintf(text,format,arg1i);  break;
				case 1: sprintf(text,format,arg2,arg1i); break;
				case 2: sprintf(text,format,arg2,arg3,arg1i);
				}
			break;

			case 'f':
			if (number_type(arg1) == NOT_NUMBER)
				return ERR_INVALID_ARGUMENT;
			arg1f=atof(arg1);
			switch(star) {
				case 0: sprintf(text,format,arg1f);  break;
				case 1: sprintf(text,format,arg2,arg1f); break;
				case 2: sprintf(text,format,arg2,arg3,arg1f);
				}
			break;

			default:
			if (arg1==NULL) arg1=""; /* Just use empty string */
			switch(star) {
				case 0: sprintf(text,format,arg1);  break;
				case 1: sprintf(text,format,arg2,arg1); break;
				case 2: sprintf(text,format,arg2,arg3,arg1);
				}
			}
		append_string(&out,text);
		continue;



		/* Deal with escape codes including octal numbers */
		case '\\':
		switch(*(s+1)) {
			case '\\': text[pos]='\\';  break;
			case '~' : text[pos]='~';  break;

			case 'n': 
			/* Send a reset */
			if (had_colour) {
				memcpy(text+pos,COL_RESET,4);
				had_colour=0;
				pos+=4;
				}
			text[pos]='\n';  
			break;

			case 'r': text[pos]='\r';  break;

			case 'b': text[pos]='\b';  break;
			case 't': text[pos]='\t';  break; /* Tab */
			case 'f': text[pos]='\f';  break; /* Form feed */
			case 'a': text[pos]='\07'; break; /* Bell */
			case '"': text[pos]='"';   break;

			default: 
			/* See if its an octal code */
			text[pos]=strtol(s+1,&s2,8);
			if (s2==s+1 || s2-s > 4) return ERR_UNKNOWN_ESC_CODE;
			s=s2-2;
			}
		++s;
		break;


		/* Parse colour codes */
		case '~':
		if (!*(s+1) || !*(s+2)) { /* At end of string? */
			text[pos]='~';  break; 
			}
		col[0]=*(s+1);
		col[1]=*(s+2);
		col[2]='\0';
		for(i=0;i < NUM_COLS;++i) 
			if (!strcmp(col,colcom[i])) break;
		if (i==NUM_COLS) {  text[pos]='~';  break;  }

		len=strlen(colcode[i]);
		memcpy(text+pos,colcode[i],len);
		had_colour=1;
		pos+=(len-1);
		s+=2;
		}
	++pos;
	++s;
	}

/* Put reset on end */
if (had_colour) {
	memcpy(text+pos,COL_RESET,4);  pos+=4;
	}
if (pos) {
	text[pos]='\0';
	append_string(&out,text);
	}
push_rstack(out);
FREE(out);
return OK;
}



/*** Overwrite or insert into a string ***/
int com_oistr(int com, int pc)
{
int ret,pc2,pos,cnt,i;
char *str1,*str2,*result;

if (pword[pc].end - pc < 3) return ERR_SYNTAX;

/* Get string to overwrite */
pc2=pc+1;
if ((ret=push_rstack_result(&pc2))!=OK) return ret;
if (IS_PAST_END) return ERR_SYNTAX;
str1=rstack_ptr->value;

/* Get string to overwrite with */
if ((ret=push_rstack_result(&pc2))!=OK) return ret;
if (IS_PAST_END) return ERR_SYNTAX;
str2=rstack_ptr->value;

/* Get position to put it at */
if ((ret=push_rstack_result(&pc2))!=OK) return ret;
if (NOT_PAST_END) return ERR_SYNTAX;
if (number_type(rstack_ptr->value) != INTEGER ||
    (pos=atoi(rstack_ptr->value)) < 1) return ERR_INVALID_ARGUMENT;

/* If first string is too small then pad it out */
result=NULL;
set_string(&result,str1);
if (isnull(str1) || strlen(str1) < pos-1) {
	/* Can't use sprintf here cos inserted position may be too big for
	   array passed to it */
	if (isnull(str1)) cnt=pos-1;
	else cnt=pos-strlen(str1)-1;

	/* Pad result up to position to put str2 */
	for(i=0;i < cnt;++i) append_string(&result," ");
	append_string(&result,str2);
	goto DONE;
	}

/* If string to add is null just return original */
if (isnull(str2)) {
	free(result);  return push_rstack(str1);
	}

if (com==OVERSTR) {
	/* Overstr */
	if (isnull(str1) || strlen(str1) < pos) {
		append_string(&result,str2);
		goto DONE;
		}
	result[pos-1]='\0';
	append_string(&result,str2);
	if (!isnull(str2) && strlen(str1) >= pos+strlen(str2)) 
		append_string(&result,str1+pos+strlen(str2)-1); 
	}
else {
	/* Inststr */
	result[pos-1]='\0';
	append_string(&result,str2);
	append_string(&result,str1+pos-1);
	}

DONE:
push_rstack(result);
free(result);
return OK;
}



/*** Replace string command ***/
int com_replstr(int com, int pc)
{
int ret,pc2,pos,len1,len2;
char *str1,*str2,*str3;
char *s,*s2,c,d[2],*result;

if (pword[pc].end - pc < 3) return ERR_SYNTAX;

/* Get main string */
pc2=pc+1;
if ((ret=push_rstack_result(&pc2)) != OK) return ret;
if (IS_PAST_END) return ERR_SYNTAX;
str1=rstack_ptr->value;

/* Get old string to replace */
if ((ret=push_rstack_result(&pc2)) != OK) return ret;
if (IS_PAST_END) return ERR_SYNTAX;
str2=rstack_ptr->value;

/* Get new string */
if ((ret=push_rstack_result(&pc2)) != OK) return ret;
str3=rstack_ptr->value;

/* Get optional start position */
if (NOT_PAST_END) {
	if ((ret=push_rstack_result(&pc2)) != OK) return ret;
	if (NOT_PAST_END) return ERR_SYNTAX;
	if (number_type(rstack_ptr->value) != INTEGER ||
	    (pos=atoi(rstack_ptr->value)) < 1) return ERR_INVALID_ARGUMENT;
	}
else pos=1;

/* Check strings */
if (isnull(str2)) return ERR_INVALID_ARGUMENT;
if (isnull(str1)) return push_rstack(NULL);
len1=strlen(str1);
len2=strlen(str2);
if (len1 < pos || len2+(pos-1) > len1) return push_rstack(str1);

/* Do the replacing */
d[1]='\0';
if (pos > 1) {
	c=str1[pos-1];
	str1[pos-1]='\0';
	result=strdup(str1);
	str1[pos-1]=c;
	}
else result=NULL;

for(s=str1+(pos-1);*s;++s) {
	if (s - str1 + len2 > len1) {
		append_string(&result,s);  break;
		}
	s2=s+len2;
	c=*s2;
	*s2='\0';
	if (!strcmp(s,str2)) {
		append_string(&result,str3);  s=s2-1;
		}
	else {
		d[0]=*s;  append_string(&result,d);
		}
	*s2=c;
	}
push_rstack(result);
FREE(result);
return OK;
}



/*** String commands that take 2 strings and an optional numnber arg at
     the end ***/
int com_strfuncs3(int com, int pc)
{
int ret,pc2,pos;
char c,*s,*s2,*str1,*str2,*result;

if (pword[pc].end - pc < 2) return ERR_SYNTAX;

/* Get 1st string */
pc2=pc+1;
if ((ret=push_rstack_result(&pc2))!=OK) return ret;
if (IS_PAST_END) return ERR_SYNTAX;
str1=rstack_ptr->value;

/* Get 2nd string/regexp pattern */
if ((ret=push_rstack_result(&pc2))!=OK) return ret;
str2=rstack_ptr->value;

/* Get optional start position */
if (NOT_PAST_END) {
	if ((ret=push_rstack_result(&pc2))!=OK) return ret;
	if (NOT_PAST_END) return ERR_SYNTAX;
	if (number_type(rstack_ptr->value) != INTEGER ||
	    (pos=atoi(rstack_ptr->value)) < 1) return ERR_INVALID_ARGUMENT;

	if (pos > strlen(str1)) {
		if (com==INSTR) return push_rstack("0");
		return push_rstack(str1);
		}
	}
else pos=1;
	

switch(com) {
	case INSTR:
	if (isnull(str1)) {
		if (isnull(str2) && pos==1) return push_rstack("1");
		return push_rstack("0");
		}
	if (isnull(str2)) {
		if (isnull(str1) && pos==1) return push_rstack("1");
		return push_rstack("0");
		}

	/* If its not a regular expression instr then do simple search */
	if (!is_regexp(str2)) {
		if (s=strstr(str1+pos-1,str2)) {
			sprintf(text,"%d",(int)(s-str1+1));  
			return push_rstack(text);
			}
		return push_rstack("0");
		}

	/* Wildcard search */
	for(s=str1+(pos-1);*s;++s) {
		for(s2=s+strlen(s);s2 > s;--s2) {
			c=*s2;  *s2='\0';
			if (regexp_match(s,str2)) {
				sprintf(text,"%d",(int)(s-str1+1));
				*s2=c;
				return push_rstack(text);
				}
			*s2=c;
			}
		}
	return push_rstack("0");


	case SUBSTR:
	case SUBSTR_SHORTCUT:
	if (isnull(str2)) return push_rstack(str1);

	/* If its not a regular expression do a simple subtract */
	if (!is_regexp(str2)) {
		if ((s=strstr(str1+(pos-1),str2))==NULL) 
			return push_rstack(str1);

		/* Remove first occurance of string2 in string */
		result=NULL;
		c=*s;  *s='\0';
		set_string(&result,str1);
		*s=c;
		append_string(&result,s+strlen(str2));
		push_rstack(result);

		free(result);
		return OK;
		}

	/* Do a match starting from the start of string1 with varying lengths
	   starting at the length of the string and reducing */
	for(s=str1+(pos-1);*s;++s) {
		for(s2=s+strlen(s);s2 > s;--s2) {
			c=*s2;  *s2='\0';
			if (regexp_match(s,str2)) {
				*s2=c;
				result=NULL;
				c=*s;  *s='\0';
				set_string(&result,str1);
				append_string(&result,s2);
				*s=c;
				push_rstack(result);
				FREE(result);
				return OK;
				}
			*s2=c;
			}
		}
	return push_rstack(str1);
	}
return ERR_INTERNAL;
}

