/***************************************************************************
 FILE: com_tcp.c
 LVU : 1.0.0

 DESC:
 Code for the TCP/IP commands.

 Copyright (C) Neil Robertson 1999, 2000
 ***************************************************************************/

#include "globals.h"


/*** Create an outgoing TCP connection ***/
int com_connect(int com, int pc)
{
STREAM_T st;
struct sockaddr_in con_addr;
struct hostent *hp;
int ret,pc2,port,sock;
char *addr,*s;

if (pword[pc].end - pc < 2) return ERR_SYNTAX;

/* Get the IP address string */
pc2=pc+1;
if ((ret=push_rstack_result(&pc2))!=OK) return ret;
if (isnull(rstack_ptr->value)) return ERR_INVALID_ARGUMENT;
addr=rstack_ptr->value;

/* Get optional port number */
if (NOT_PAST_END) {
	if ((ret=push_rstack_result(&pc2))!=OK) return ret;
	if (NOT_PAST_END) return ERR_SYNTAX;
	if (number_type(rstack_ptr->value) != INTEGER ||
	    (port=atoi(rstack_ptr->value)) < 1 || port > 65535)
		return ERR_INVALID_ARGUMENT;
	}
else port=23; /* Standard default port */

/* See if numeric address given */
s=addr;
while(*s && (*s=='.' || isdigit(*s))) ++s;

/* Set up the address structure & create socket */
if (*s) {
	/* Name address given */
	if (!(hp=gethostbyname(addr))) return ERR_UNKNOWN_HOST;
	memcpy((char *)&con_addr.sin_addr,hp->h_addr,hp->h_length);
	}
else if ((con_addr.sin_addr.s_addr=inet_addr(addr))==-1) 
	return ERR_UNKNOWN_HOST;

con_addr.sin_family=AF_INET;
con_addr.sin_port=htons(port);

if ((sock=socket(AF_INET,SOCK_STREAM,0))==-1) return ERR_SOCKET;

/* Attempt the connect */
if (connect(sock,(struct sockaddr *)&con_addr,sizeof(con_addr))==-1)
	return ERR_CONNECT;

/* Create the stream */
if ((ret=create_stream("SOCK_CLIENT",
                       ST_SOCKET,STM_READWRITE,0,sock,0,&st))!=OK) {
	close(sock);  return ret;
	}

/* Copy conn_addr into net_addr to gethost... etc functions will work */
memcpy(&st->net_addr,&con_addr,sizeof(con_addr));

return push_rstack(st->name);
}



/*** Create a TCP listening stream ***/
int com_listen(int com, int pc)
{
STREAM_T st;
struct sockaddr_in bind_addr;
struct in_addr iaddr;
struct hostent *hp;
u_int addr;
int port,sock,on;
int ret,pc2,qsize;
char *s;

if (pword[pc].end == pc) return ERR_SYNTAX;

port=0;
qsize=20; /* seems like a reasonable default value to me */
addr=INADDR_ANY;
pc2=pc+1;

/* Get port, optional address and optional queue size  */
while(NOT_PAST_END) {
	if (pword[pc].end - pc2 < 1) return ERR_SYNTAX;

	/* Port */
	if (!strcasecmp(pword[pc2].word,"port")) {
		pc2++;
		if ((ret=push_rstack_result(&pc2))!=OK) return ret;
		if (number_type(rstack_ptr->value) != INTEGER ||
		    (port=atoi(rstack_ptr->value)) < 1 || 
		    port > 65535)  return ERR_INVALID_ARGUMENT;
		}
	else
	/* Queue size */
	if (!strcasecmp(pword[pc2].word,"qsize")) {
		pc2++;
		if ((ret=push_rstack_result(&pc2))!=OK) return ret;
		if (number_type(rstack_ptr->value) != INTEGER ||
		    (qsize=atoi(rstack_ptr->value)) < 1)
			return ERR_INVALID_ARGUMENT;
		}
	else
	/* Interface address */
	if (!strcasecmp(pword[pc2].word,"addr")) {
		pc2++;
		if ((ret=push_rstack_result(&pc2))!=OK) return ret;
		if (isnull(rstack_ptr->value)) return ERR_INVALID_ARGUMENT;

		/* See if numeric address given */
		s=rstack_ptr->value;
		while(*s && (*s=='.' || isdigit(*s))) ++s;

		if (*s) {
			/* Name address given */
			if (!(hp=gethostbyname(rstack_ptr->value))) 
				return ERR_UNKNOWN_HOST;
			memcpy((char *)&iaddr,hp->h_addr,hp->h_length);
			addr=iaddr.s_addr;
			}
		else addr=inet_addr(rstack_ptr->value);
		}	
	else return ERR_SYNTAX;		
	}
if (!port) return ERR_SYNTAX;

/* Create a listening socket */
if ((sock=socket(AF_INET,SOCK_STREAM,0)) == -1) return ERR_SOCKET;

/* Allow port to be re-used when on a TIME_WAIT */
on=1;
setsockopt(sock,SOL_SOCKET,SO_REUSEADDR,(char *)&on,sizeof(on));

/* Bind socket to the port */
bind_addr.sin_family=AF_INET;
bind_addr.sin_addr.s_addr=addr;
bind_addr.sin_port=htons(port);
#ifdef IRIX
/* Shouldn't need to do this but if we don't bind won't work with any
   specific addresses (as opposed to INADDR_ANY) */
memset(bind_addr.sin_zero,0,sizeof(bind_addr.sin_zero)); 
#endif
if (bind(sock,(struct sockaddr *)&bind_addr,sizeof(struct sockaddr_in))==-1)
	return ERR_BIND;

/* Start the socket listening for connections */
if (listen(sock,qsize)==-1) return ERR_LISTEN;

/* Create the stream */
if ((ret=create_stream("SOCK_LISTEN",ST_LSOCKET,STM_READ,0,sock,0,&st))!=OK) {
	close(sock);  return ret;
	}
return push_rstack(st->name);
}



/*** Accept an incoming connection or return the IP number or name of the 
     host connected to the streams socket or the remote port number that
     they're connecting from. ***/
int com_acc_gethost(int com, int pc)
{
STREAM_T st;
struct hostent *host;
int ret,pc2,fd,size,port;
char ip[100];

if (pword[pc].end == pc) return ERR_SYNTAX;

/* Get stream */
pc2=pc+1;
if ((ret=push_rstack_result(&pc2))!=OK) return ret;
if (NOT_PAST_END) return ERR_SYNTAX;
if (isnull(rstack_ptr->value)) return ERR_UNDEFINED_STREAM;

if ((st=get_stream(rstack_ptr->value)) == NULL)
	return ERR_UNDEFINED_STREAM;

/* Make sure its a socket for gethost commands */
if (com!=ACCEPT && st->type != ST_SOCKET) return ERR_INVALID_STREAM_TYPE;

switch(com) {
	case ACCEPT:
	/* Make sure its a listening socket */
	if (st->type != ST_LSOCKET) return ERR_INVALID_STREAM_TYPE;
	fd=st->fd;

	/* Create the stream before accept so accept() can copy in net_addr */
	if ((ret=create_stream("SOCK_SERVER",
	                       ST_SOCKET,STM_READWRITE,0,-1,0,&st)) != OK) 
		return ret;

	/* Accept connection */
	size=sizeof(struct sockaddr_in);
	if ((st->fd=accept(fd,(struct sockaddr *)&st->net_addr,&size)) == -1)
		return ERR_ACCEPT;
	return push_rstack(st->name);


	case GETHOSTNAME:
	host=gethostbyaddr((char *)&(st->net_addr.sin_addr.s_addr),4,AF_INET);
	if (host) return push_rstack(host->h_name);
	/* Can't get the name so fall through to return the ip number */


	case GETHOSTNUM:
	strcpy(ip,(char *)inet_ntoa(st->net_addr.sin_addr));
	return push_rstack(ip);


	case GETHOSTPORT:
	sprintf(text,"%d",ntohs(st->net_addr.sin_port));
	return push_rstack(text);


	case GETHOSTINFO:
	/* Return ip name, number and port in a list */
	host=gethostbyaddr((char *)&(st->net_addr.sin_addr.s_addr),4,AF_INET);
	strcpy(ip,(char *)inet_ntoa(st->net_addr.sin_addr));
	port=ntohs(st->net_addr.sin_port);

	if (host) sprintf(text,"%s %s %d",host->h_name,ip,port);
	else sprintf(text,"%s %s %d",ip,ip,port);
	return push_rstack(text);
	}
return ERR_INTERNAL;
}



/*** Resolve an ip name or number ***/
int com_getip(int com, int pc)
{
struct hostent *host;
struct in_addr iaddr;
int ret,pc2;
u_int addr;

/* Get ip number */
pc2=pc+1;
if ((ret=push_rstack_result(&pc2))!=OK) return ret;
if (NOT_PAST_END) return ERR_SYNTAX;
if (isnull(rstack_ptr->value)) return ERR_INVALID_ARGUMENT;

if (com==GETIPNAME) {
	/* Get hostname */
	addr=inet_addr(rstack_ptr->value);
	host=gethostbyaddr((char *)&addr,4,AF_INET);
	if (host) return push_rstack(host->h_name);
	return push_rstack("");
	}

/* GETIPNUM */
if ((host=gethostbyname(rstack_ptr->value))==NULL)
	return push_rstack("");
memcpy((char *)&iaddr,host->h_addr,host->h_length);
return push_rstack((char *)inet_ntoa(iaddr));
}
