/***************************************************************************
 FILE: com_terminal.c
 LVU : 1.0.0

 DESC:
 Commands for vt/ansi terminals.

 Copyright (C) Neil Robertson 1999, 2000
 ***************************************************************************/

#include "globals.h"


/*** Clear the screen. Alternative to print "~CL" ***/
int com_cls(int com, int pc)
{
STREAM_T st;
int ret,pc2;

/* Use current output stream */
if (pword[pc].end == pc) {
	if (out_stream==NULL) return ERR_OUTPUT_STREAM_CLOSED;
	write_stream(out_stream,"\033[H\033[J",6);
	return OK;
	}

/* Use supplied stream */
pc2=pc+1;
if ((ret=push_rstack_result(&pc2)) != OK) return ret;
if (NOT_PAST_END) return ERR_SYNTAX;
if (isnull(rstack_ptr->value)) return ERR_INVALID_ARGUMENT;

if ((st=get_stream(rstack_ptr->value)) == NULL)
	return ERR_UNDEFINED_STREAM;

if (st->open_mode==STM_READ) return ERR_STREAM_INPUT_ONLY;

write_stream(st,"\033[H\033[J",6);
return OK;
}



/*** Locate the terminal cursor at the supplied positions. Allow float
     arguments but just convert to ints. ***/
int com_locate(int com, int pc)
{
STREAM_T st;
int ret,pc2,x,y;
char code[20];

if (pword[pc].end - pc < 2) return ERR_SYNTAX;

/* Get X co-ord. */
pc2=pc+1;
if ((ret=push_rstack_result(&pc2)) != OK) return ret;
if (IS_PAST_END) return ERR_SYNTAX;
if (number_type(rstack_ptr->value) == NOT_NUMBER)
	return ERR_INVALID_ARGUMENT;
if ((x=(int)rint(atof(rstack_ptr->value))) < 0) x=0;

/* Get Y co-ord */
if ((ret=push_rstack_result(&pc2)) != OK) return ret;
if (number_type(rstack_ptr->value) == NOT_NUMBER)
	return ERR_INVALID_ARGUMENT;
if ((y=(int)rint(atoi(rstack_ptr->value))) < 0) y=0;

/* Create control code */
sprintf(code,"\033[%d;%dH",y,x);	

/* Write to current output stream if no optional stream given */
if (IS_PAST_END) {
	if (out_stream==NULL) return ERR_OUTPUT_STREAM_CLOSED;
	if (location_visible(out_stream,x,y)) 
		write_stream(out_stream,code,strlen(code));
	return OK;
	}

/* Get optional stream */
if ((ret=push_rstack_result(&pc2)) != OK) return ret;
if (NOT_PAST_END) return ERR_SYNTAX;
if (isnull(rstack_ptr->value)) return ERR_INVALID_ARGUMENT;

if ((st=get_stream(rstack_ptr->value)) == NULL)
	return ERR_UNDEFINED_STREAM;

if (st->open_mode==STM_READ) return ERR_STREAM_INPUT_ONLY;

if (location_visible(st,x,y)) write_stream(st,code,strlen(code));
return OK;
}



/*** Switch out of bounds screen printing on and off ***/
int com_oob(int com, int pc)
{
int pc2;

if (pword[pc].end == pc) {
	oob_top = (com==OOBON);
	oob_bottom = (com==OOBON);
	oob_left = (com==OOBON);
	oob_right = (com==OOBON);
	return OK;
	}
for(pc2=pc+1; pc2 <= pword[pc].end;++pc2) {
	if (!strcasecmp(pword[pc2].word,"top")) oob_top = (com==OOBON);
	else
	if (!strcasecmp(pword[pc2].word,"bottom")) oob_bottom = (com==OOBON);
	else
	if (!strcasecmp(pword[pc2].word,"left")) oob_left = (com==OOBON);
	else
	if (!strcasecmp(pword[pc2].word,"right")) oob_right = (com==OOBON);
	else
	return ERR_INVALID_ARGUMENT;
	}
return OK;
}



/*** Draw a line between two co-ordinates with the given character 
     or string. ***/
int com_line(int com, int pc)
{
STREAM_T st;
float x,y,xinc,yinc,xdiff,ydiff;
int xf,yf,xt,yt;
int ret,pc2,cnt,i;
char *str;

if (pword[pc].end - pc < 5) return ERR_SYNTAX;

/* Get the arguments */
cnt=0;
pc2=pc+1;
while(NOT_PAST_END) {
	if (cnt > 5) return ERR_SYNTAX;

	if ((ret=push_rstack_result(&pc2)) != OK) return ret;
	if (cnt > 1 && cnt < 5 && number_type(rstack_ptr->value) == NOT_NUMBER)
		return ERR_INVALID_ARGUMENT;

	switch(++cnt) {
		case 1:
		if (rstack_ptr->value==NULL) str="";
		else str=rstack_ptr->value;  
		break;

		case 2:
		xf=atoi(rstack_ptr->value);  break;

		case 3:
		yf=atoi(rstack_ptr->value);  break;

		case 4:
		xt=atoi(rstack_ptr->value);  break;

		case 5:
		yt=atoi(rstack_ptr->value);  break;

		case 6:
		/* Get optional stream */
		if (isnull(rstack_ptr->value)) return ERR_INVALID_ARGUMENT;
		if ((st=get_stream(rstack_ptr->value)) == NULL)
			return ERR_UNDEFINED_STREAM;

		if (st->open_mode==STM_READ) return ERR_STREAM_INPUT_ONLY;
		}
	}
if (cnt < 5) return ERR_SYNTAX;
if (cnt==5 && (st=out_stream)==NULL) return ERR_OUTPUT_STREAM_CLOSED;

/* Work out increments */
xdiff=(float)abs(xt-xf);
ydiff=(float)abs(yt-yf);
if (xdiff > ydiff) {
	cnt=xdiff;
	xinc=(xt > xf) ? 1 : -1;
	yinc=(ydiff/xdiff) * ((yt > yf) ? 1 : -1);
	}
else {
	cnt=ydiff;
	yinc=(yt > yf) ? 1 : -1;
	xinc=(xdiff/ydiff) * ((xt > xf) ? 1 : -1);
	}
	
/* Do the drawing. If cols and rows are zero assume we couldn't get terminal
   size and write anyway.  */
for(i=0, x=xf, y=yf;i <= cnt;++i) {
	if (location_visible(st,x,y)) {
		sprintf(text,"\033[%d;%dH%s",(int)rint(y),(int)rint(x),str);
		write_stream(st,text,strlen(text));
		}
	x+=xinc;
	y+=yinc;
	}
return OK;
}



/*** Draw a circle, ellipse or arc ***/
int com_circle(int com, int pc)
{
STREAM_T st;
double x,y,x2,y2,x_radius,y_radius;
double ang,angadd,angfrom,angto;
int ret,pc2,cnt;
char *str;

if (pword[pc].end - pc < 4+(com==ELLIPSE)) return ERR_SYNTAX;

cnt=0;
pc2=pc+1;
st=NULL;
angfrom=0;
angto=6.3;

/* Get the arguments */
while(NOT_PAST_END) {
	if (cnt > 8) return ERR_SYNTAX;
	if (cnt!=7 && (ret=push_rstack_result(&pc2)) != OK) return ret;

	switch(++cnt) {
		case 1:
		if (rstack_ptr->value==NULL) str="";
		else str=rstack_ptr->value;  
		break;

		case 2:
		if (number_type(rstack_ptr->value) == NOT_NUMBER)
			return ERR_INVALID_ARGUMENT;
		x=atof(rstack_ptr->value);
		break;

		case 3:
		if (number_type(rstack_ptr->value) == NOT_NUMBER)
			return ERR_INVALID_ARGUMENT;
		y=atof(rstack_ptr->value);
		break;

		case 4:
		if (number_type(rstack_ptr->value) == NOT_NUMBER)
			return ERR_INVALID_ARGUMENT;
		x_radius=atof(rstack_ptr->value);
		break;

		case 5:
		if (com!=CIRCLE) {
			if (number_type(rstack_ptr->value) == NOT_NUMBER)
				return ERR_INVALID_ARGUMENT;
			y_radius=atof(rstack_ptr->value);
			break;
			}

		case 6:
		if (com==ARC) {
			if (number_type(rstack_ptr->value) == NOT_NUMBER)
				return ERR_INVALID_ARGUMENT;
			angfrom=atof(rstack_ptr->value);
			break;
			}
		/* Get optional stream */
		if (isnull(rstack_ptr->value)) return ERR_INVALID_ARGUMENT;
		if ((st=get_stream(rstack_ptr->value)) == NULL)
			return ERR_UNDEFINED_STREAM;

		if (st->open_mode==STM_READ) return ERR_STREAM_INPUT_ONLY;
		break;

		case 7:
		if (com!=ARC) return ERR_SYNTAX;
		if (number_type(rstack_ptr->value) == NOT_NUMBER)
			return ERR_INVALID_ARGUMENT;
		angto=atof(rstack_ptr->value);
		break;

		case 8:
		if (!strcasecmp(pword[pc2].word,"rad")) break;
		else
		if (!strcasecmp(pword[pc2].word,"deg")) {
			angfrom/=DEGS_PER_RADIAN;
			angto/=DEGS_PER_RADIAN;
			}
		else
		return ERR_SYNTAX;
		++pc2;
		break;
		
		case 9:
		/* Get optional stream for ARC */
		if (isnull(rstack_ptr->value)) return ERR_INVALID_ARGUMENT;
		if ((st=get_stream(rstack_ptr->value)) == NULL)
			return ERR_UNDEFINED_STREAM;

		if (st->open_mode==STM_READ) return ERR_STREAM_INPUT_ONLY;
		}
	}
if (cnt < 4+(com==ELLIPSE)+3*(com==ARC)) return ERR_SYNTAX;

if (st==NULL && (st=out_stream)==NULL) return ERR_OUTPUT_STREAM_CLOSED;

/* Terminals have low resolution. This tries to speed up the drawing process
   by going around the circle in greater steps if the circle is smaller. Its
   not a very good algorithm but hey... */
if (com==CIRCLE) angadd=1/(x_radius*2);
else angadd=1/(x_radius+y_radius);
angadd=(angadd < 0.02 ? 0.02 : angadd);

/* Draw the circle or ellipse. There is a faster way of doing this without
   using the trig functions but I couldn't get it to work and it was a serious
   hassle to code anyway. The trig is arranged so that zero degrees is at the
   top of the screen. */
for(ang=angfrom;ang < angto;ang+=angadd) {
	x2=rint(x+(sin(ang) * x_radius));
	y2=rint(y+(sin(ang-1.571) * (com==CIRCLE ? x_radius : y_radius)));
	if (location_visible(st,(int)x2,(int)y2)) {
		sprintf(text,"\033[%d;%dH%s",(int)y2,(int)x2,str);
		write_stream(st,text,strlen(text));
		}
	}
return OK;
}



/*** See if the given co-ordinate is visible on the terminal the stream is
     attached to. If we don't have a size for it assume we can print since
     we might be writing to a file to cat it directly to the screen at some
     other time. Return 1 if screen oob is off for the given position. ***/
int location_visible(STREAM_T st, int x, int y)
{
if (st==NULL) return 0;

if (x < 1 && oob_left) return 0;
if (y < 1 && oob_top) return 0;
if (st->cols && x > st->cols && !oob_right) return 0;
if (st->rows && y > st->rows && !oob_bottom) return 0;

return 1;
}



/*** Get the size of the current terminal screen. Have these commands instead
     of system vars so we can ignore SIGWINCH if we choose. ***/
int com_termsize(int com, int pc)
{
#ifdef TIOCGWINSZ
struct winsize ws;
STREAM_T st;
int ret,pc2;

if (pc != pword[pc].end) {
	/* Get optional stream */
	pc2=pc+1;
	if ((ret=push_rstack_result(&pc2)) != OK) return ret;
	if (NOT_PAST_END) return ERR_SYNTAX;
	if (isnull(rstack_ptr->value)) return ERR_INVALID_ARGUMENT;
	if ((st=get_stream(rstack_ptr->value)) == NULL)
		return ERR_UNDEFINED_STREAM;
	}
else 
if ((st=out_stream)==NULL) return ERR_OUTPUT_STREAM_CLOSED;

if (ioctl(st->fd,TIOCGWINSZ,&ws)==-1) return ERR_CANT_GET_TERMSIZE;

out_stream->rows=ws.ws_row;
out_stream->cols=ws.ws_col;

switch(com) {
	case GETROWS: sprintf(text,"%d",ws.ws_row);  break;
	case GETCOLS: sprintf(text,"%d",ws.ws_col);  break;
	case GETSIZE: sprintf(text,"%d %d",ws.ws_col,ws.ws_row);  break;

	default: return ERR_INTERNAL;
	}
return push_rstack(text);
#endif

return ERR_UNAVAILABLE;
}
