/***************************************************************************
 FILE: com_timing.c
 LVU : 1.0.0

 DESC:
 This contains the code for time & date commands.

 Copyright (C) Neil Robertson 1999, 2000
 ***************************************************************************/

#include "globals.h"


/*** sleep command ***/
int com_sleep(int com, int pc)
{
struct timeval tm;
int ret,pc2,secs,usecs;

if (pword[pc].end == pc) return ERR_SYNTAX;

/* Get time */
pc2=pc+1;
if ((ret=push_rstack_result(&pc2))!=OK) return ret;
if (NOT_PAST_END) return ERR_SYNTAX;

if (number_type(rstack_ptr->value) == NOT_NUMBER ||
    (secs=atoi(rstack_ptr->value)) < 0) return ERR_INVALID_ARGUMENT;

usecs=get_usecs(rstack_ptr->value);

/* I use select here instead of using a usleep() call because select will
   return -1 if its interrupted by a signal , usleep gives no such indication
   and not all systems have it anyway */
tm.tv_sec=secs;
tm.tv_usec=usecs;
if (select(0,NULL,0,0,&tm) == -1) return push_rstack("0");
return push_rstack("1");
}



/*** Return the time and date in the format specified. The format string is
     passed straight to strftime ***/
int com_date(int com, int pc)
{
int ret,pc2;
char *format,out[100];
time_t secs;
struct tm *tm;

if (pword[pc].end == pc) return ERR_SYNTAX;

/* Get format string */
pc2=pc+1;
if ((ret=push_rstack_result(&pc2)) != OK) return ret;
if (isnull(rstack_ptr->value)) return ERR_INVALID_ARGUMENT;
format=rstack_ptr->value;

/* Get optional seconds */
if (NOT_PAST_END) {
	if ((ret=push_rstack_result(&pc2)) != OK) return ret;
	if (NOT_PAST_END) return ERR_SYNTAX;
	if (number_type(rstack_ptr->value) != INTEGER) 
		return ERR_INVALID_ARGUMENT;
	secs=atoi(rstack_ptr->value);
	}
else secs=time(0);

/* Get result */
tm=localtime(&secs);
strftime(out,99,format,tm);
return push_rstack(out);
}

