/***************************************************************************
 FILE: com_vars.c
 LVU : 1.0.0

 DESC:
 This contains code for the variable commands SET, VAR & WIPE.

 Copyright (C) Neil Robertson 1999
 ***************************************************************************/


#include "globals.h"


/*** Parse the variable line for a procedures local vars ***/
int com_var(int com, int pc)
{
if (pword[pc].end - pc < 1) return ERR_SYNTAX;
return parse_variable_line(pc,NULL,pstack_ptr);
}



/*** Set a variable. If '=' is the 2nd word then we don't have an explicit
     SET command. eg: a=2 rather than set a 2 ***/
int com_set(int com, int pc)
{
int ret,pc2,add;

if (pword[pc].end - pc < 2) return ERR_SYNTAX;

pc2=pc+2;
if ((ret=push_rstack_result(&pc2))!=OK) return ret;
if (NOT_PAST_END) return ERR_SYNTAX;
add=(pword[pc+1].opcode!=EQUALS);

return set_var_value(pword[pc+add].word,pstack_ptr,rstack_ptr->value,NO);
}



/*** Set a 1D array from a selection of values ***/
int com_aset(int com, int pc)
{
VAR_T main_var,var,curr_elem,new;
int add,ret,pc2,tmp,cnt;
char *sub,elem_name[10];
char *val[2];

if (pword[pc].end - pc < 2) return ERR_SYNTAX;

add=(pword[pc+1].opcode != STAR_EQUALS && pword[pc+1].opcode != PCT_EQUALS);
pc2=pc+add;

/* Get main variable */
if ((ret=get_var_main(pword[pc2].word,pstack_ptr,&var,&sub)) != OK)
	return ret;
main_var=var;

/* Get sub var. If subscript doesn't exist then we must create it */
if (sub && 
    (ret=getset_var_element(var,pstack_ptr,&var,NULL,sub,NO,&tmp)) != OK) {
	if (ret != ERR_UNDEFINED_SUBSCRIPT) return ret;

	/* Create new subscript */
	if ((ret=getset_var_element(main_var,pstack_ptr,&var,NULL,sub,YES,&tmp)) !=OK) 
		return ret;
	}
if (!var->dim) return ERR_VAR_IS_NOT_ARRAY;

/* Find last element of array list */
for(curr_elem=var->array,cnt=1;curr_elem;curr_elem=curr_elem->next,++cnt)
	if (curr_elem->next==NULL) {  ++cnt;  break;  }

/* Set up array elements (and subscripts if AFSET) as we go through command 
   args */
for(pc2=pc2+1+!add;NOT_PAST_END;) {
	/* Get key if afset or value if aset */
	if ((ret=push_rstack_result(&pc2)) != OK) return ret;
	val[0]=rstack_ptr->value;

	if (com==AFSET) {
		if (IS_PAST_END) return ERR_MISSING_ARRAY_VALUE;

		/* Check for optional => operator */
		if (pword[pc2].opcode==ASSOCIATIVE_SET) {
			++pc2;
			if (IS_PAST_END) return ERR_MISSING_ARRAY_VALUE;
			}

		/* Get value */
		if ((ret=push_rstack_result(&pc2)) != OK) return ret;
		val[1]=rstack_ptr->value;

		if ((new=create_variable(val[0],NULL,NULL,NO,NO))==NULL)
			return ERR_MALLOC;
		}
	else {
		/* ASET */
		sprintf(elem_name,"%d",cnt);
		if ((new=create_variable(elem_name,NULL,NULL,NO,NO))==NULL)
			return ERR_MALLOC;
		}
	new->dim=var->dim-1;
	new->value=strdup(val[com==AFSET]);
	new->next=NULL;

	if (curr_elem) {
		curr_elem->next=new;
		new->prev=curr_elem;
		}
	curr_elem=new;
	if (var->array==NULL) var->array=new;
	++cnt;
	}
return OK;
}



/*** Wipe all values and elements from a variable/array or completely
     erase the variable (ie delete it) ***/
int com_wevar(int com, int pc)
{
VAR_T mainvar,var,var2;
char *sub;
int ret,pc2,dummy;

if (pword[pc].end - pc < 1) return ERR_SYNTAX;

/* Loop through variables */
for(pc2=pc+1;NOT_PAST_END;++pc2) {
	if (pword[pc2].word[0]=='"' || pword[pc2].word[0]=='{')
		return ERR_SYNTAX;
		
	/* Get the var */
	if ((ret=get_var_main(pword[pc2].word,pstack_ptr,&mainvar,&sub))!=OK)
		return ret;
	var=mainvar;
	if (sub && 
	    (ret=getset_var_element(var,pstack_ptr,&var,NULL,sub,NO,&dummy))!=OK)
		return ret;

	if (com==ERASE) {
		/* Check var isn't being referenced by a refvar */
		FOR_ALL_VARS(var2) {
			if (var2->refvar == var)
				return ERR_CANT_WIPE_VAR_BEING_REFERENCED;
			}

		/* Set main var->array if appropriate */
		if (sub && var==mainvar->array) mainvar->array=var->next;

		/* Destroy variable */
		remove_variable(var);
		destroy_variable(var);
		continue;
		}

	/* WIPE everything except the main variable */
	FREE(var->value);
	var->value=NULL;
	if ((ret=erase_vararray(var->array))!=OK) return ret;
	var->array=NULL;
	}
return OK;
}



/*** Recurse through the array elements. Referencing error will only occur
     if someone tries to wipe a global that is being refeneced by a 
     procedure parameter. ***/
int erase_vararray(VAR_T var)
{
VAR_T var2,var3,next;
LSTACK_T lst;

for(var2=var;var2;var2=next) {
	FOR_ALL_VARS(var3)
		if (var3->refvar == var2) 
			return ERR_CANT_WIPE_VAR_BEING_REFERENCED;

	/* Check for forent loop referencing it */
	FOR_LOOP_STACK(lst) {
		if (lst->forent==var2) lst->forent=var2->next;
		if (lst->arr_index==var2) lst->arr_index=NULL;
		if (lst->arr_value==var2) lst->arr_index=NULL;
		}

	next=var2->next;
	free(var2->name);
	FREE(var2->value);
	if (var2->array && erase_vararray(var2->array)!=OK) 
		return ERR_CANT_WIPE_VAR_BEING_REFERENCED;
	free(var2);
	}
return OK;
}



/*** Return the number of entries in an array ***/
int com_arrsize(int com, int pc)
{
VAR_T var,ent;
char *sub;
int ret,tmp,cnt;

if (pword[pc].end - pc != 1) return ERR_SYNTAX;

/* Get variable */
if ((ret=get_var_main(pword[pc+1].word,pstack_ptr,&var,&sub)) != OK)
	return ret;
if (sub && (ret=getset_var_element(var,pstack_ptr,&var,NULL,sub,NO,&tmp))!=OK) 
	return ret;
if (!var->dim) return ERR_VAR_IS_NOT_ARRAY;

/* Count entries */
for(ent=var->array,cnt=0;ent;ent=ent->next,++cnt);
sprintf(text,"%d",cnt);
return push_rstack(text);
}
