/***************************************************************************
 FILE: loops.c
 LVU : 1.0.0

 DESC:
 This contains the low level code for creating, searching and controlling
 loops, switch-switched constructs and goto/labels.

 Copyright (C) Neil Robertson 1999, 2000
 ***************************************************************************/

#include "globals.h"


/*** Create a linked list entry for a label ***/
int create_label(int p)
{
LABEL_T l,new;
char *name;

if (pword[p].end - p != 1) return ERR_SYNTAX;

if ((new=(LABEL_T)malloc(sizeof(struct labels)))==NULL) return ERR_MALLOC;

/* If name is in quotes treat */
if ((name=pword[p+1].word)[0]=='"') {
	name++;
	name[strlen(name)-1]='\0';
	}

/* See if label already exists */
FOR_ALL_LABELS(l) 
	if (!strcmp(l->name,name) && l->proc==curr_proc) 
		return ERR_DUPLICATE_DECLARATION;

/* Create entry */
if ((new->name=strdup(name))==NULL) return ERR_MALLOC;
if (name!=pword[p+1].word) name[strlen(name)]='"';

if (first_label) last_label->next=new;
else first_label=new;
last_label=new;
new->next=NULL;

new->proc=curr_proc;
new->pos=p;
return OK;
}



/*** Create a linked list entry for a loop ***/
int create_loop(int p)
{
LOOP_T new;

if ((new=(LOOP_T)malloc(sizeof(struct loops)))==NULL) return ERR_MALLOC;
if (first_loop) {
	last_loop->next=new;
	new->prev=last_loop;
	}
else {
	first_loop=new;
	new->prev=NULL;
	}
last_loop=new;
new->next=NULL;

new->type=pword[p].com;
new->start=p;
new->end=0;

new->proc=curr_proc;

return OK;
}



/*** Set the end position of a loop ***/
int terminate_loop(int pc)
{
LOOP_T l;
int otype;

switch(pword[pc].com) {
	case LOOPING : otype=LOOP;    break;
	case NEXT    : otype=FOR;      break;
	case NEXTENT : otype=FORENT;   break;
	case NEXTELEM: otype=FORELEM;  break;
	case WEND    : otype=WHILE;    break;
	case WHILST  : otype=DO;       break;
	case SWITCHED: otype=SWITCH; 
	}
for(l=last_loop;l;l=l->prev) {
	if (l->type==otype && l->proc==curr_proc && !l->end) {
		l->end=pc;  return OK;
		}
	}
return ERR_UNEXPECTED_COMMAND;
}



/*** Get the loop that starts at the given position ***/
LOOP_T get_loop(int pc)
{
LOOP_T l;

FOR_ALL_LOOPS(l) if (l->start==pc) return l;
return NULL;
}
