/***************************************************************************
 FILE: main.c
 LVU : 1.0.0

 DESC:
 This is where the main() function, the initialisation code and the 
 interpreters main loop are kept.

 Copyright (C) Neil Robertson 1999, 2000
 ***************************************************************************/

#define MAIN_FILE
#include "globals.h"
#undef MAIN_FILE

char build[50];

/*** Lets start doing something ... ***/
main(int argc, char **argv)
{
init_interpreter();
get_command_line(argc,argv);
load_and_first_parse();
second_parse();
mainloop(NO,0);
system_exit(0);
}



/*** Initialise the interpreter including all the Loor system variables ***/
void init_interpreter()
{
VAR_T new;
STREAM_T st;
struct utsname un;
struct passwd *pwd;
char *err,num[10];
int i,e;

/* Create the build string */
build[0]='\0';

#ifdef LINUX
strcpy(build,"LINUX");
#endif

#ifdef IRIX
strcpy(build,"IRIX");
#endif

#ifdef SOLARIS
strcpy(build,"SOLARIS");
#endif

#ifdef NO_CRYPT
if (build[0]) strcat(build," ");
strcat(build,"NO_CRYPT");
#endif

if (!build[0]) strcpy(build,"STANDARD");


/* Set to zero so files are created with absolute given perms unless
   the umask command is used */
umask(0);

/* Init some vars */
pword=NULL;
alloced_words=0;
num_words=0;
include_level=0;
private_member=NO;
manual_throw=NO;
element_seps=NULL;
oob_top=1;
oob_bottom=1;
oob_left=1;
oob_right=1;

first_class=NULL;
last_class=NULL;

first_obj=NULL;
last_obj=NULL;
curr_obj=NULL;

first_var=NULL;
last_var=NULL;
curr_var=NULL;

first_proc=NULL;
last_proc=NULL;
curr_proc=NULL;

first_try=NULL;
last_try=NULL;
curr_try=NULL;
first_excp=NULL;
last_excp=NULL;
curr_excp=NULL;

first_with=NULL;
last_with=NULL;
curr_with=NULL;

first_loop=NULL;
last_loop=NULL;

first_label=NULL;
last_label=NULL;

first_stream=NULL;
last_stream=NULL;
stream_cnt=1;

pstack_start=NULL;
pstack_ptr=NULL;

lstack_start=NULL;
lstack_ptr=NULL;

rstack_start=NULL;
rstack_ptr=NULL;

wstack_start=NULL;
wstack_ptr=NULL;

ustack_start=NULL;
ustack_end=NULL;

sighandler_start=NULL;
sighandler_end=NULL;
sigq_start=0;
sigq_next=0;
for(i=0;i < SIGQ_SIZE;++i) sigq[i]=NULL;

empty_stack_marker=NULL;


/* Create loor system variables. Can't be bothered to check the return values, 
   if theres not enough memory to create these then somethings seriously wrong 
   with your system and everything will die anyway. >:) */
new=create_variable("$ES",NULL,NULL,NO,NO);
add_variable(new);

new=create_variable("$EOF",NULL,NULL,NO,YES);
new->value=strdup("0");
add_variable(new);

new=create_variable("$version",NULL,NULL,NO,YES);
new->value=strdup(VERSION);
add_variable(new);

new=create_variable("$build",NULL,NULL,NO,YES);
new->value=strdup(build);
add_variable(new);

new=create_variable("$build_date",NULL,NULL,NO,YES);
new->value=strdup(BUILD_DATE);
add_variable(new);

new=create_variable("$argc",NULL,NULL,NO,YES);
new->value=strdup("1");
add_variable(new);
new=create_variable("@$argv",NULL,NULL,NO,YES);
add_variable(new);

new=create_variable("@$error",NULL,NULL,NO,YES);
add_variable(new);
for(e=1;e < NUM_ERRORS;++e) {
	sprintf(text,"$error[\"%d\"]",e);
	set_var_value(text,NULL,error_mesg[e],YES);
	}

new=create_variable("$last_error",NULL,NULL,NO,YES);
new->value=strdup("0");
add_variable(new);

new=create_variable("@$os_error",NULL,NULL,NO,YES);
add_variable(new);
for(e=1;e <= NUM_OS_ERRORS;++e) {
	sprintf(text,"$os_error[\"%d\"]",e);
	if (err=(char *)strerror(e)) set_var_value(text,NULL,err,YES);
	else set_var_value(text,NULL,"unused",YES);
	}

new=create_variable("$last_os_error",NULL,NULL,NO,YES);
new->value=strdup("0");
add_variable(new);

new=create_variable("$colcodes",NULL,NULL,NO,YES);
new->value=strdup(COLCOM_LIST);
add_variable(new);

new=create_variable("$pid",NULL,NULL,NO,YES);
sprintf(text,"%d",getpid());
new->value=strdup(text);
add_variable(new);

new=create_variable("$ppid",NULL,NULL,NO,YES);
sprintf(text,"%d",getppid());
new->value=strdup(text);
add_variable(new);

new=create_variable("$in_stream",NULL,NULL,NO,YES);
new->value=strdup("STDIN");
add_variable(new);

new=create_variable("$out_stream",NULL,NULL,NO,YES);
new->value=strdup("STDOUT");
add_variable(new);

new=create_variable("$telopt",NULL,NULL,NO,YES);
new->value=strdup("0");
add_variable(new);

new=create_variable("$empty",NULL,NULL,NO,NO);
new->value=strdup("");
add_variable(new);

new=create_variable("$signal",NULL,NULL,NO,YES);
add_variable(new);

new=create_variable("$siglist",NULL,NULL,NO,YES);
text[0]='\0';
for(i=0;i < NUM_SIGNALS;++i) sprintf(text,"%s%s ",text,siglist[i].name);	
text[strlen(text)-1]='\0';  /* Remove trailing space */
new->value=strdup(text);
add_variable(new);

new=create_variable("$uname",NULL,NULL,NO,YES);
uname(&un);
sprintf(text,"%s %s %s %s %s",un.sysname,un.nodename,un.release,un.version,un.machine);
new->value=strdup(text);
add_variable(new);

/* Create some arrays */
new=create_variable("@$os_user",NULL,NULL,NO,YES);
add_variable(new);
pwd=getpwuid(getuid());
strcpy(text,"$os_user[\"name\"]");
set_var_value(text,NULL,pwd ? pwd->pw_name : NULL,YES);
	
strcpy(text,"$os_user[\"passwd\"]");
set_var_value(text,NULL,pwd ? pwd->pw_passwd : NULL,YES);
	
strcpy(text,"$os_user[\"uid\"]");
if (pwd) sprintf(num,"%d",pwd->pw_uid); else strcpy(num,"0");
set_var_value(text,NULL,num,YES);

strcpy(text,"$os_user[\"gid\"]");
if (pwd) sprintf(num,"%d",pwd->pw_gid); else strcpy(num,"0");
set_var_value(text,NULL,num,YES);

strcpy(text,"$os_user[\"gecos\"]");
set_var_value(text,NULL,pwd ? pwd->pw_gecos : NULL,YES);

strcpy(text,"$os_user[\"home\"]");
set_var_value(text,NULL,pwd ? pwd->pw_dir : NULL,YES);

strcpy(text,"$os_user[\"shell\"]");
set_var_value(text,NULL,pwd ? pwd->pw_shell : NULL,YES);

strcpy(text,"$os_user[\"age\"]");
#ifndef LINUX
set_var_value(text,NULL,pwd ? pwd->pw_age : NULL,YES);
#else 
set_var_value(text,NULL,"unused",YES);
#endif

strcpy(text,"$os_user[\"comment\"]");
#ifndef LINUX
set_var_value(text,NULL,pwd ? pwd->pw_comment : NULL,YES);
#else 
set_var_value(text,NULL,"unused",YES);
#endif


new=create_variable("@$streams",NULL,NULL,NO,YES);
add_variable(new);

new=create_variable("@$objects",NULL,NULL,NO,YES);
add_variable(new);


/* Dummy variables. Their values are actually set by code in get_var_value
   but a dummy entry is needed in case users try to give them a value. 
   We would get "undefined variable" error if they didn't exist but creating
   them causes correct "read only" error to be displayed */
new=create_variable("$time",NULL,NULL,NO,YES);
add_variable(new);

new=create_variable("$utime",NULL,NULL,NO,YES);
add_variable(new);

new=create_variable("$dir",NULL,NULL,NO,YES);
add_variable(new);

/* Object self reference variables */
new=create_variable("$this",NULL,NULL,NO,YES);
add_variable(new);
new=create_variable("$self",NULL,NULL,NO,YES);
add_variable(new);


/* Create built in streams & associated array */
create_stream("STDIN",ST_CONSOLE,STM_READ,0,0,0,&in_stream);
in_stream->stripnl=YES;
create_stream("STDOUT",ST_CONSOLE,STM_WRITE,0,1,0,&out_stream);
create_stream("STDERR",ST_CONSOLE,STM_WRITE,0,2,0,&st);

/* Set up signals */
set_sighandler();
}



/*** Get the command line parameters and create the loor program equivalent
     parameters ***/
void get_command_line(int argc, char **argv)
{
char *usage="Usage: %s [-h] [-v] [-f <filename>] [<command line arguments>]\n";
int i,fnpos;

/* $argv[1] contains the interpreter binary name, $argv[2] contains the
   loor program source file name. Set 2 here so that it appears after 1 
   in an array scan. */
fnpos=0;
set_var_value("$argv[1]",NULL,argv[0],YES);
set_var_value("$argv[2]",NULL,"<STDIN>",YES);

/* Go through arguments and set up $argv and file any filename following 
   a -f option */
for(i=1;i < argc;++i) {
	if (argv[i][0]=='-') {
		switch(argv[i][1]) {
			case 'h':
			puts("\nLOOR - a Language of Object Orientation by mr Robertson");
			puts("Copyright (C) Neil Robertson 1999, 2000\n");
			printf("Version   : %s\n",VERSION);
			printf("Build     : %s\n",build);
			printf("Build date: %s\n\n",BUILD_DATE);
			printf("Command line options:\n");
			printf("    -h           : Print this help.\n");
			printf("    -v           : Display version number , build and build date.\n");
			printf("    -f <filename>: The Loor source code file to run.\n\n");
			printf("The command line arguments will be passed to the Loor program as $argv with\n$argc giving the count.\n\n");
			puts("Please read the disclaimer and copyright notices provided with this software\nto be fully aware of the legal situation applying to it as far as use,\ncopying and liability are concerned.\n");
			puts("Bug reports and queries to: neil@ogham.demon.co.uk\n");
			exit(0);

			case 'f':
			if (i == argc-1) {	
				printf(usage,argv[0]);  exit(0);
				}
			fnpos=++i;
			continue;

			case 'v':
			printf("Loor %s %s (%s)\n",VERSION,build,BUILD_DATE);  
			exit(0);
			}
		printf(usage,argv[0]);  
		exit(0);
		}

	/* Set command line args */
	sprintf(text,"$argv[%d]",i+2-2*(fnpos > 0));
	set_var_value(text,NULL,argv[i],YES);
	}

/* Set up input filename and argument count */
if (fnpos) {
	ifile[0].filename=strdup(argv[fnpos]);	
	set_var_value("$argv[2]",NULL,argv[fnpos],YES);
	sprintf(text,"%d",argc - 1);
	}
else {
	ifile[0].filename=strdup("<STDIN>");
	sprintf(text,"%d",argc + 1);
	}
set_var_value("$argc",NULL,text,YES);
}



/*** The main interpreter loop. This gets called recursively by com_call() 
     when executing a call command ***/
int mainloop(int subloop,int pc)
{
PROC_T p;
EXCP_T excp;
VAR_T param;
int line,ret,e,pc2;
char *file;
static int saved_ret,saved_errno,saved_line;
static char *saved_file;

/* Find position of the main procedure if we've not in a recursive call to
   this function. */
if (!subloop) {
	pc=-1;
	FOR_ALL_PROCS(p)
		if (!strcmp(p->name,"main")) {  pc=p->start;  break;  }
	if (pc==-1) error_exit(ERR_MAIN_PROC_MISSING,0,pword[0].filename);

	if (p->vproc) error_exit(ERR_MAIN_IS_VPROC,pword[pc].realline,pword[0].filename);
	push_pstack(p,0,NULL);

	/* Create any formal parameters. Since $argc and $argv are global these
	   will be null unless main until is called recursively. */
	for(pc2=pc+2;NOT_PAST_END;++pc2) {
		if ((param=create_variable(pword[pc2].word,
		                           NULL,pstack_ptr,NO,NO))==NULL)
			return ERR_MALLOC;
		add_variable(param);
		}
	pc=pword[pc].end+1;
	}

while(1) {
	errno=0;
	line=pword[pc].realline;
	file=pword[pc].filename;
	saved_ret=0;
	saved_errno=0;
	saved_line=0;
	saved_file=NULL;

	/* If command returned an error and if we're in a TRY - TRIED block
	   then look for exception handler for the error */
	if ((ret=run_command(&pc)) != OK) {
		if (manual_throw || 
		    (pstack_ptr->catchreturn_pc[1]=in_try_block(pc))) {
			/* See if the error can be handled by an exception 
			   block. Can't do this in run_command() because of 
			   recursion. Block must follow line that caused 
			   error */
			FOR_ALL_EXCEPTIONS(excp) {
				if (excp->proc==pstack_ptr->proc &&
				    excp->start >= pc) {
					for(e=0;e < NUM_ERRORS && excp->trapped[e];++e) {
						if (excp->trapped[e]!=ret) continue;

						/* For catchreturn & catchbreak */
						pstack_ptr->catchreturn_pc[0]=pc;
						pstack_ptr->catchbreak_pc=pword[excp->end].end+1;

						/* Jump to block */
						pc=pword[excp->start].end+1;
						cleardown_lw_stacks(pc);
						sprintf(text,"%d",ret);
						set_var_value("$last_error",NULL,text,YES);
						sprintf(text,"%d",errno);
						set_var_value("$last_os_error",NULL,text,YES);
						saved_ret=0;
						saved_errno=0;
						saved_line=0;
						saved_file=NULL;
						manual_throw=NO;
						goto CONT;
						}
					}
				}
			}
		/* If no exception block to handle the error and we're back in
		   the main procedure just exit with the error */
		if (!subloop) {
			if (!saved_line) {
				if (manual_throw) error_exit(ERR_UNHANDLED_EXCEPTION,line,file);
				error_exit(ret,line,file);
				}

			if (manual_throw) error_exit(ERR_UNHANDLED_EXCEPTION,saved_line,saved_file);
			errno=saved_errno;
			error_exit(saved_ret,saved_line,saved_file);
			}
		
		/* Save info about where and what the error is so we can
		   exit back up the procedure tree in the hope of finding
		   an exception block to handle it. */
		if (!saved_line) {
			saved_ret=ret;
			saved_errno=errno;
			saved_line=line;
			saved_file=file;
			}
		exit_procedure();
		return ret;
		}

	CONT:
	/* Check for any signals */
	if ((ret=process_sigqueue(pc)) != OK) error_exit(ret,line,file); 

	if (pc==pstack_ptr->ret_pc) {
		exit_procedure();  return OK;
		}
	}
}



/*** Run a single command ***/
int run_command(int *pc)
{
EXCP_T excp;
int pc2,ret,com;
static int recursion=0;

/* If no command found what the hell has gone wrong? */
if (!pword[*pc].com) return ERR_INTERNAL;

ret=OK;
recursion++;
pc2=*pc;
com=pword[*pc].com;

/* If the command has an associated function then run it */
if (command[com].func) {
	if (command[com].ptr_param) ret=command[com].func(com,pc);
	else {
		ret=command[com].func(com,pc2);
		*pc=pword[*pc].end+1;
		}
	if (pword[pc2].com==RETURN) {
		if (ret!=OK) return ret;
		if (!(*pc=pstack_ptr->ret_pc)) {
			/* If we're returning from "proc main" with a value
			   convert it to a number and use as exit code */
			if (rstack_ptr) system_exit(atoi(rstack_ptr->value));
			system_exit(0);
			}
		goto DONE;
		}
	}
else {
	/* Else either skip to the end of it or ... */
	switch(com) {
		case ENDPROC:
		/* Return from proc or exit if its main */
		push_rstack("");
		if (!(*pc=pstack_ptr->ret_pc)) system_exit(0);
		goto DONE;


		case CATCH:
		/* Skip to end of the exception handler */
		FOR_ALL_EXCEPTIONS(excp) {
			if (excp->start==pc2) {
				*pc=pword[excp->end].end+1;  return OK;
				}
			}
		return ERR_INTERNAL;
		}
	*pc=pword[*pc].end+1;
	}

/* Clear out the result stack if we're at the top of the call tree */
DONE:
if (!--recursion) {
	while(rstack_ptr) pop_rstack();
	}
return ret;
}



/*** See if we're in a TRY - TRIED block ***/
int in_try_block(int pc)
{
TRY_T tr;

for(tr=first_try;tr;tr=tr->next)
	if (pc >= tr->start && pc <= tr->end) return tr->end+1;
return 0;
}
