/***************************************************************************
 FILE: misc.c
 LVU : 1.0.0

 DESC:
 I've stuck code here that I can't be bothered making seperate modules for 
 mainly because its not worth it as they would be so small.

 Copyright (C) Neil Robertson 1999, 2000
 ***************************************************************************/

#include "globals.h"


/*** Terminate the program and the interpreter with an error ***/
void error_exit(int sys_error, int realline, char *filename)
{
fflush(stdout);

if (errno) 
	fprintf(stderr,"ERROR %d: File \"%s\", line %d: %s: %s\n",
		sys_error,
	        filename,
	        realline,
		error_mesg[sys_error],
	        strerror(errno));
else
	fprintf(stderr,"ERROR %d: File \"%s\", line %d: %s\n",
		sys_error,
	        filename,
	        realline,
		error_mesg[sys_error]);

system_exit(errno);
}



/*** Exit the intepreter. Set console to echo just in case it was switched
     off by something ***/
void system_exit(int err)
{
struct termios tio;

tcgetattr(0,&tio);
tio.c_lflag |= ECHO;
tcsetattr(0,TCSANOW,&tio);

exit(err);
}



/*** Check a variable/class/object/procedure name for validity ***/
int check_name(char *name)
{
int alpha;

if (name==NULL || !*name) return 0;

alpha=0;
for(;*name;++name) {
	if (isalpha(*name) || *name=='_' || *name=='$') {  
		alpha=1;  continue;  
		}
	if (isdigit(*name)) continue;
	return 0;
	}
if (!alpha) return 0;
return 1;
}



/*** See if a string is a number, either integer or float. Whitespace at
     the start is NOT dealt with. ***/
int number_type(char *str)
{
int neg;
int dot;
char *s;

if (str==NULL) return NOT_NUMBER;

neg=0;
dot=0;
s=str;
while(*s) {
	if (*s=='-') {
		if (s!=str || neg || !(*(s+1))) return NOT_NUMBER;
		neg=1;
		goto CONT;
		}
	if (*s=='.') {
		if (dot || !(*(s+1))) return NOT_NUMBER;
		dot=1;
		goto CONT;
		}
	if (!isdigit(*s)) return NOT_NUMBER;

	CONT:
	++s;
	}
if (dot) return FLOAT;
return INTEGER;
}



/*** Set a string ***/
int set_string(char **str,char *value)
{
FREE(*str);
if (value==NULL || !value[0]) {
	*str=NULL;  return OK;
	}
if ((*str=strdup(value))==NULL) return ERR_MALLOC;
return OK;
}



/*** Append to a string ***/
int append_string(char **str,char *value)
{
if (*str==NULL) return set_string(str,value);
if (value==NULL || !value[0]) return OK;
if ((*str=(char *)realloc(*str,strlen(*str)+strlen(value)+1))==NULL)
	return ERR_MALLOC;
strcat(*str,value);
return OK;
}



/*** Append a char to a string. Used for inputting binary data ***/
int append_char(char **str, char c, int len)
{
if ((*str=(char *)realloc(*str,len+1))==NULL) return ERR_MALLOC;
(*str)[len]=c;
return OK;
}



/*** Returns YES if string is NULL or of zero length ***/
int isnull(char *str)
{
if (str==NULL || !strlen(str)) return YES;
return NO;
}



/*** Used to find a call command in nested command calls ***/
int nested_call_command(int pc)
{
int i;

for(i=pc;i <= pword[pc].end;++i)
	if (pword[i].com==CALL) return YES;
return NO;
}



/*** See if a character is a list element seperator ***/
int is_element_sep(char c, char *es)
{
if (es==NULL) {
	if ((u_char)c < 33) return YES;
	return NO;
	}
if (strchr(es,c)) return YES;
return NO;
}



/*** Return the value after the decimal point as a value between 0 and
     10^6-1 to be used as a usec value ***/
int get_usecs(char *num)
{
char *s,*dot,c;
int usecs;

for(s=num,dot=NULL;*s;++s) if (*s=='.') dot=s;
if (dot==NULL || dot==s-1) return 0;
c='\0';
if (s-dot > 7) {
	s=dot+7;  c=*s;  *s='\0';
	}
usecs=atoi(dot+1)*(int)pow((double)10,(double)(7-(int)(s-dot)));
if (c) *s=c;
return usecs;
}
