/***************************************************************************
 FILE: parse.c
 LVU : 1.0.0

 DESC:
 This contains the program code load & parse functions.

 Copyright (C) Neil Robertson 1999, 2000
 ***************************************************************************/

#include "globals.h"


/*** Load up the program text and parse it into words and check bracket
     counts. ***/
void load_and_first_parse()
{
FILE *fp;
int c,c2,lastc,progline;
int level,pos,empty_line;
int in_string,in_rem,in_word,cont_slash;
int cur_brackets,rnd_brackets;
int com_start[100];

errno=0;
level=0;
progline=1;
com_start[0]=0;
open_file(0);

while(level != -1) {
	START:
	in_string=0;
	in_word=0;
	in_rem=0;
	lastc='\n';
	pos=0;
	text[0]='\0';
	fp=ifile[level].fp;
	cur_brackets=0;
	rnd_brackets=0;
	cont_slash=0;
	empty_line=1;

	/* 4 is ^D */
	while((c=getc(fp))!=EOF && c!=4) {
		if (num_words >= alloced_words) {
			/* Realloc more space if we've run out. +2 is so we 
			   have some overrun words as add_word can be called
			   twice in one loop iteration */
			alloced_words+=PROG_WORD_ALLOC;
			if ((pword=(struct program *)realloc(pword,(alloced_words+2) * sizeof(struct program)))==NULL)
				error_exit(ERR_MALLOC,ifile[level].realline,ifile[level].filename);
			}

		if (c=='\n') {
			if (cont_slash) {
				ifile[level].realline++;
				cont_slash=0;
				goto CONT;
				}
			if (in_string) 
				error_exit(ERR_UNTERMINATED_STRING,ifile[level].realline,ifile[level].filename);
			if (cur_brackets || rnd_brackets) 
				error_exit(ERR_BRACKETS,ifile[level].realline,ifile[level].filename);

			if (in_rem) in_rem=0;
			if (in_word) add_word(level,progline,&pos,&in_word);

			ifile[level].realline++;
			if (!empty_line) {
				com_start[0]=num_words;  
				progline++;
				}
			empty_line=1;
			goto CONT;
			}

		if (in_rem) goto CONT;
		if (c=='\\' && lastc!='\\' && !in_string) {  
			cont_slash=1;  goto CONT;  
			}

		if (!in_string) {
			switch(c) {
				case '#': 
				if (lastc!='[') {  in_rem=1;  goto CONT;  }
				break;

				case ' ':
				case TAB:
				if (in_word) {
					add_word(level,progline,&pos,&in_word);  
					goto CONT;
					}
				in_word=0;
				goto CONT;

				case '{': 
				cur_brackets++;  
				com_start[cur_brackets]=num_words+1;

				case '}':
				case '(': 
				case ')':
				case '+':
				case '*':
				case '/':
				case '^':
				case '%':
				case '!':
				case '&':
				case '|':
				case '~':
				if (c=='&' && !in_string && lastc=='\n') break;

				empty_line=0;
				if (in_word) 
					add_word(level,progline,&pos,&in_word);
				text[0]=(u_char)c;
				pos=1;
				add_word(level,progline,&pos,&in_word);
				switch(c) { 
					case '}': cur_brackets--;  break;
					case '(': rnd_brackets++;  break;	
					case ')': rnd_brackets--;
					}
				cont_slash=0;
				goto CONT;


				case '-':
				empty_line=0;
				cont_slash=0;
				if (in_word) {
					/* Look ahead to see if next char
					   is '>' */
					c2=getc(fp);
					ungetc(c2,fp);
					if (c2=='>') break;

					add_word(level,progline,&pos,&in_word);
					}

				/* Decide whether its the minus operator or
				   its a negative value */
				c2=getc(fp);
				ungetc(c2,fp);
				if ((u_char)c2 < 33) break;

				/* No whitespace following '-' so decide
				   whether we have a minus operator or a
				   negated value */

				/* If its the 1st or 2nd word on a line its
				   negative */
				if (num_words==com_start[cur_brackets] || 
				    num_words==com_start[cur_brackets]+1) break;

				/* If not in an eqn  command then it will
				   always be a negative value */
				if (pword[com_start[cur_brackets]].word &&
				    strcasecmp(pword[com_start[cur_brackets]].word,"EQN"))
						break;

				/* If maths op or opening brackets preceeding 
				   then its neg val. */ 
				if (strchr("({<>!=+-*-^%&|~",pword[num_words-1].word[0]))
					break;

				/* Store as an operator */
				text[0]=(u_char)c;
				pos=1;
						
				add_word(level,progline,&pos,&in_word);
				switch(c) { 
					case '}': cur_brackets--;  break;
					case '(': rnd_brackets++;  break;	
					case ')': rnd_brackets--;
					}
				goto CONT;
				
				
				case '<':
				if (lastc=='<') {
					pos=2;
					strcpy(text,"<<");
					num_words--;
					}
				else {
					if (in_word) 
						add_word(level,progline,&pos,&in_word);
					text[0]=(u_char)c;
					pos=1;
					}
				add_word(level,progline,&pos,&in_word);
				cont_slash=0;
				goto CONT;


				case '>':
				cont_slash=0;
				if (lastc=='-') break; /* -> var op */

				switch(lastc) {
					case '>':
					pos=2;
					strcpy(text,">>");
					num_words--;
					break;

					case '=':
					pos=2;
					strcpy(text,"=>"); /* Assoc. set */
					num_words--;
					break;

					default:
					if (in_word) add_word(level,progline,&pos,&in_word);
					text[0]=(u_char)c;
					pos=1;
					}
				add_word(level,progline,&pos,&in_word);
				goto CONT;


				case '=':
				if (strchr("<>!%*",lastc)) {
					pos=2;
					sprintf(text,"%c=",lastc);
					num_words--;
					}
				else {
					if (in_word) 
						add_word(level,progline,&pos,&in_word);
					text[0]=(u_char)c;
					pos=1;
					}
				add_word(level,progline,&pos,&in_word);
				cont_slash=0;
				goto CONT;


				case ';':
				if (empty_line) 
					error_exit(ERR_SYNTAX,ifile[level].realline,ifile[level].filename);
				if (in_word) add_word(level,progline,&pos,&in_word);
				progline++;
				com_start[0]=num_words;
				empty_line=1;
				cont_slash=0;
				goto CONT;
				}
			}
		empty_line=0;

		switch(c) {
			case '"': 
			if (in_string && lastc=='\\') pos--;
			else in_string=!in_string;
			if (in_string) in_word=1;
			text[pos++]=(u_char)c;
			break;

			case '&':
			if (!in_string && lastc=='\n') {
				ifile[level].realline++;
				level++;
				get_include_filename(fp,level);
				open_file(level);
				goto START;
				}
			/* Fall through */

			default:
			in_word=1;
			text[pos++]=(u_char)c;
			}
		cont_slash=0;

		CONT:
		lastc=c;
		}
	if (ifile[level].fp != stdin) fclose(ifile[level].fp);
	level--;
	}
pword[num_words].word=NULL;
pword[num_words].filename=NULL;
pword[num_words].opcode=0;
pword[num_words].com=0;
pword[num_words].realline=0;
pword[num_words].progline=0;
pword[num_words].end=0;
pword[num_words].processed=NO;
pword[num_words].result_set=NO;
pword[num_words].next_pc=NO;
pword[num_words].result=-1;
}




/*** Add a program word into the pword structure array ***/
void add_word(
	int level,
	int progline,
	int *pos,
	int *in_word)
{
int i;

text[*pos]='\0';

if ((pword[num_words].word=strdup(text))==NULL ||
    (pword[num_words].filename=strdup(ifile[level].filename))==NULL)
	error_exit(ERR_MALLOC,ifile[level].realline,ifile[level].filename);

pword[num_words].realline=ifile[level].realline;
pword[num_words].progline=progline;
pword[num_words].end=0;
pword[num_words].com=0;
pword[num_words].processed=NO;
pword[num_words].result=-1;

/* Find the opcode if any */
pword[num_words].opcode=0;
for(i=1; i <= NUM_OPERATORS ;++i) {
	if (!strcmp(text,op_string[i])) {
		pword[num_words].opcode=i;  break;
		}
	}
num_words++;
*pos=0;
*in_word=0;
}



/*** Get the name of the file to be included ***/
void get_include_filename(FILE *fp,int level)
{
int c,pos=0;

while((c=getc(fp))!=EOF && c!='\n' && c!=' ' && c!=TAB) {
	text[pos++]=(char)c;  text[pos]='\0';
	}
ifile[level].filename=strdup(text);
}



/*** Open main file or an include file. <STDIN> means we're reading the
     program from standard input. Ok, if for some wierd reason you want to
     name a file <STDIN> and read in from it you'll have problems. 
     Well tough! :) ***/
void open_file(int level)
{
if (!level && !strcmp(ifile[0].filename,"<STDIN>")) ifile[0].fp=stdin;
else
if (!(ifile[level].fp=fopen(ifile[level].filename,"r"))) 
	error_exit(ERR_CANT_OPEN_FILE, \
	           ifile[level-(level > 0)].realline, \
	           ifile[level-(level > 0)].filename);
ifile[level].realline=1;
}



/*** This sets the endpos in the pword struct , deals with aliases and
     creates global variables ***/
void second_parse()
{
LOOP_T l;
int i,j,line,pline,ret,if_cnt;
int brackets,brackets2,com;
char *w,*file;

if_cnt=0;
brackets=0;
pline=0;
curr_class=NULL;
curr_proc=NULL;

for(i=0;i < num_words;++i) {
	if (pline!=pword[i].progline && curr_class==NULL) private_member=NO;

	w=pword[i].word;
	line=pword[i].realline;
	pline=pword[i].progline;
	file=pword[i].filename;

	if (w[0]=='{') brackets++;  
	else
	if (w[0]=='}') brackets--;

	/* get endpos */
	brackets2=brackets;
	for(j=i;j < num_words;++j) {
		w=pword[j].word;
		if (w[0]=='{') brackets2++;
		else
		if (w[0]=='}') brackets2--;
		if ((brackets && brackets2==brackets-1) || 
		    pword[j].progline!=pword[i].progline) break;
		}
	pword[i].end=j-1;

	/* Deal with classes, procs and vars */
	switch(pword[i].com=get_command_number(i)) {
		case NOT_COMMAND: 
		if (!i && curr_proc==NULL) error_exit(ERR_SYNTAX,line,file);

		/* If we have a preceeding '{' or we're at the start of a new 
		   line or we're following a trap command then we're either 
		   calling a procedure without using the optional CALL command 
		   or if an '=' follows its a SET operation */
		if (i && 
		    (pword[i-1].word[0]=='{' ||
		     !strcasecmp(pword[i-1].word,"TRAP") || 
		     pline > pword[i-1].progline)) {

			/* Check for command shortcuts */
			switch(pword[i+1].opcode) {
				case EQUALS: 
				pword[i].com=SET;  
				if (curr_proc) break;
				goto JUMP_SET;

				case PCT_EQUALS:
				pword[i].com=ASET;  
				if (curr_proc) break;
				goto JUMP_ASET;

				case STAR_EQUALS: 
				pword[i].com=AFSET;  
				if (curr_proc) break;
				goto JUMP_ASET;

				default: 
				if (curr_proc==NULL) 
					error_exit(ERR_SYNTAX,line,file);
				pword[i].com=CALL;
				}
			}
		break;

		case SET:
		JUMP_SET:
		if (curr_proc) break; /* Set those when we run the proc */
		if (curr_class) goto ERROR1;
		if ((ret=com_set(SET,i)) != OK) goto ERROR2;
		break;


		case ASET:
		case AFSET:
		JUMP_ASET:
		if (curr_proc) break; /* Set those when we run the proc */
		if (curr_class) goto ERROR1;
		if ((ret=com_aset(pword[i].com,i)) != OK) goto ERROR2;
		break;


		case VAR: 
		if (curr_proc) break; /* Create those when we run the proc */
		if (curr_class) {
			if ((ret=parse_class_variable_line(i,curr_class,NO))!=OK)
				goto ERROR2;
			}
		else 
		if ((ret=parse_variable_line(i,NULL,NULL))!=OK) 
			goto ERROR2;
		break;


		case ALIAS: set_alias(i);  break;


		case CLASS:
		if (curr_class) goto ERROR1;
		if ((ret=parse_class_line(i))!=OK) goto ERROR2;
		private_member=NO;
		break;


		case ENDCLASS:
		if (i != pword[i].end) error_exit(ERR_SYNTAX,line,file);
		if (curr_class==NULL || curr_proc) goto ERROR1;
		curr_class=NULL;
		break;


		case PUBLIC:
		case PRIVATE:
		if (curr_proc) goto ERROR1;
		/* Can be outside a CLASS - ENDCLASS block only if preceeding
		   a procedure or variable that is defined to be a class member:
		   eg: "private: proc fred::fred_proc"
		 */
		if (curr_class==NULL) {
			if (pword[i].end - i) {
				if ((com=get_command_number(i+1)) != PROC &&
				    com!=VPROC_COM &&
				    com!=VAR) goto ERROR1;
				pword[i].end=i;
				}
			else goto ERROR1;
			}
		private_member=(pword[i].com==PRIVATE ? YES : NO);
		break;


		case PROC:
		case VPROC_COM:
		if (curr_proc) goto ERROR1;
		if ((ret=parse_proc_line(i))!=OK) goto ERROR2;
		if_cnt=0;
		break;


		case ENDPROC:
		if (i != pword[i].end) error_exit(ERR_SYNTAX,line,file);
		if (curr_proc==NULL) goto ERROR1;

		if (if_cnt) error_exit(ERR_MISSING_FI,line,file);
		if (curr_excp) error_exit(ERR_MISSING_CAUGHT,line,file);
		if (curr_try)  error_exit(ERR_MISSING_TRIED,line,file);
		if (curr_with && curr_with->proc==curr_proc) 
			error_exit(ERR_MISSING_WITHOUT,line,file);

		curr_proc->end=i;
		curr_proc=NULL;
		break;


		case TRY:
		if (i != pword[i].end) error_exit(ERR_SYNTAX,line,file);
		if (curr_proc==NULL) goto ERROR1;
		if (curr_try) error_exit(ERR_NESTED_TRY_BLOCKS,line,file);
		if ((ret=create_try_block(i)) != OK) goto ERROR2;
		break;
		

		case TRIED:
		if (i != pword[i].end) error_exit(ERR_SYNTAX,line,file);
		if (curr_proc==NULL || curr_try==NULL) goto ERROR1;
		curr_try->end=i;
		curr_try=NULL;
		break;


		case CATCH:
		case CATCHRUN:
		if (curr_proc==NULL) goto ERROR1;
		if (curr_excp)
			error_exit(ERR_NESTED_EXCEPTION_BLOCKS,line,file);
		if ((ret=parse_exception_handler(i))!=OK) goto ERROR2;
		break;


		case CAUGHT:
		if (i != pword[i].end) error_exit(ERR_SYNTAX,line,file);
		if (curr_proc==NULL || curr_excp==NULL) goto ERROR1;
		curr_excp->end=i;
		curr_excp=NULL;
		break;


		case CATCHRETURN: 
		case CATCHBREAK:
		if (curr_proc==NULL || curr_excp==NULL) goto ERROR1;
		break;


		case WITH:
		if (curr_proc==NULL) goto ERROR1;
		if ((ret=create_with_block(i)) != OK) goto ERROR2;
		break;


		case WITHOUT:
		if (i != pword[i].end) error_exit(ERR_SYNTAX,line,file);
		if (curr_proc==NULL || curr_with==NULL) goto ERROR1;
		curr_with->end=i;
		curr_with=curr_with->prev;  /* Because withs can be nested */
		break;


		default:
		if (curr_proc==NULL) goto ERROR1;
		}

	/* Deal with loops, labels & ifs */
	switch(pword[i].com) {
		case LABEL:
		if ((ret=create_label(i))!=OK) goto ERROR2;
		break;

		case LOOP:
		case FOR:
		case FORENT:
		case FORELEM:
		case WHILE:
		case DO:
		case SWITCH:
		if ((ret=create_loop(i))!=OK)  goto ERROR2;
		break;

		case LOOPING:
		case NEXT:
		case NEXTENT:
		case NEXTELEM:
		case WEND:
		case WHILST: 
		case SWITCHED:
		if ((ret=terminate_loop(i))!=OK)  goto ERROR2;
		break;

		case IF: if_cnt++;  break;
		case FI: 
		if (!if_cnt) goto ERROR1;
		if_cnt--;  
		}	
	}
if (curr_proc)  error_exit(ERR_MISSING_ENDPROC,line,file);
if (curr_class) error_exit(ERR_MISSING_ENDCLASS,line,file);
if (curr_with && curr_with->proc==curr_proc)
	error_exit(ERR_MISSING_WITHOUT,line,file);

FOR_ALL_LOOPS(l) {
	if (!l->end) 
		error_exit(ERR_UNTERMINATED_LOOP,pword[l->start].realline,pword[l->start].filename);
	}
return;

ERROR1:
error_exit(ERR_UNEXPECTED_COMMAND,line,file);

ERROR2:
error_exit(ret,line,file);
}



/*** Get the number of the command at position w in pword ***/
int get_command_number(int w)
{
int i;

for(i=1;i <= NUM_COMS;++i) 
	if (!strcasecmp(command[i].name,pword[w].word)) return i;
return 0;
}



/*** Go through the program code and swap any alias words for their
     intended replacement values. Start search just after occurance of
     the alias command. ***/
void set_alias(int pos)
{
char *old,*new;
int i;

if (pword[pos].end - pos != 2) 
	error_exit(ERR_SYNTAX,pword[pos].progline,pword[pos].filename);

if (((old=pword[pos+1].word)[0] == '{') ||
    ((new=pword[pos+2].word)[0] == '{')) 
	fprintf(stderr,"WARNING: potential alias error on line %d in file %s...\n",pword[pos].realline,pword[pos].filename);

for(i=pword[pos].end+1;i < num_words;++i) {
	if (!strcmp(pword[i].word,old)) {
		free(pword[i].word);
		pword[i].word=strdup(new);
		}
	}
}


