/***************************************************************************
 FILE: procs.c
 LVU : 1.0.0

 DESC:
 Some low level procedure setup code which includes setting up exception
 handler and try-tried blocks.
 
 Copyright (C) Neil Robertson 1999, 2000
 ***************************************************************************/

#include "globals.h"


/*** Create a procedure entry. ***/
int parse_proc_line(int pc)
{
PROC_T pr;
CLASS_T cl;
char *name,*nameend,*s;
int got_class,i;

if (pword[pc].end == pc) return ERR_SYNTAX;

cl=curr_class;
s=name=pword[pc+1].word;

/* See if we're defining a specific class for this procedure to belong
   to: eg  proc fredclass::tom ... ***/
nameend=name+strlen(name);
got_class=0;

for(;s < nameend; ++s) {
	if (*s==':') {
		if (*(s+1) != ':') return ERR_INVALID_PROCNAME;
		if (curr_class) return ERR_UNEXPECTED_CLASS_SPEC;
		*s='\0';	
		FOR_ALL_CLASSES(cl) 
			if (!strcmp(cl->name,name)) break;
		*s=':';
		if (cl==NULL) return ERR_UNDEFINED_CLASS;
		name=s+2;
		got_class=1;
		break;
		}
	}
/* Can't have private: or public: preceeding a procedure that does not
   have a class definition */
if (pc &&
    !got_class && 
    curr_class==NULL &&
    (pword[pc-1].com==PRIVATE || pword[pc-1].com==PUBLIC)) 
	return ERR_UNEXPECTED_COMMAND;

/* Check procname */
if (!check_name(name)) return ERR_INVALID_PROCNAME;
for(i=1;i <= NUM_COMS;++i) 
	if (!strcmp(command[i].name,name)) return ERR_INVALID_PROCNAME;

FOR_ALL_PROCS(pr)
	if (pr->class==cl && !strcmp(pr->name,name))
		return ERR_DUPLICATE_DECLARATION;

/* Create list entry */
if ((curr_proc=(PROC_T)malloc(sizeof(struct procedures)))==NULL)
	return ERR_MALLOC;
if ((curr_proc->name=strdup(name))==NULL)
	return ERR_MALLOC;

/* If constructor, destructor or duplicator function set appropriate value 
   in the class */
if (cl) {
	if (!strcmp(name,"constructor")) {
		if (pword[pc].com == VPROC_COM) return ERR_CDD_IS_VPROC;
		if (pword[pc].end - pc > 1) return ERR_CDD_ARGS;
		cl->constructor=curr_proc;
		}
	else
	if (!strcmp(name,"destructor")) {
		if (pword[pc].com == VPROC_COM) return ERR_CDD_IS_VPROC;
		if (pword[pc].end - pc > 1) return ERR_CDD_ARGS;
		cl->destructor=curr_proc;
		}
	else
	if (!strcmp(name,"duplicator")) {
		if (pword[pc].com == VPROC_COM) return ERR_CDD_IS_VPROC;
		if (pword[pc].end - pc > 1) return ERR_CDD_ARGS;
		cl->duplicator=curr_proc;
		}
	}

if (first_proc) last_proc->next=curr_proc;
else first_proc=curr_proc;
curr_proc->next=NULL;
last_proc=curr_proc;

/* Set up class stuff */
if (cl) curr_proc->class=cl; else curr_proc->class=curr_class;
if (curr_proc->class) curr_proc->private=private_member;
else curr_proc->private=NO;

curr_proc->vproc = (pword[pc].com==VPROC_COM);
curr_proc->start=pc;
curr_proc->end=0;

return OK;
}



/*** Parse the CATCH-CAUGHT exception handlers inside a procedure.
     err_trapped is defined in globals.h ***/
int parse_exception_handler(int pc)
{
int i,j,cnt,k;

if (pword[pc].end == pc) return ERR_SYNTAX;
if ((curr_excp=create_exception_block())==NULL) return ERR_MALLOC;

/* Match exceptions to be handled with valid ones in above list */
cnt=0;
for(i=pc+1;i <= pword[pc].end;++i) {
	for(j=0;j < EXCP_TRAP_COUNT;++j) {
		/* Got valid error to trap */
		if (!strcasecmp(pword[i].word,err_trapped[j].name)) {
			/* See if its already in stored list */
			for(k=0;k < cnt;++k) {
				if (curr_excp->trapped[k]==err_trapped[j].err)
					return ERR_DUPLICATE_DECLARATION;
				}

			/* Add to list */
			curr_excp->trapped[cnt]=err_trapped[j].err;
			++cnt;
			break;
			}
		}
	if (j==EXCP_TRAP_COUNT) return ERR_INVALID_EXCEPTION;
	}
curr_excp->start=pc;
curr_excp->proc=curr_proc;
curr_excp->end=0;  /* Set in second_parse() */
return OK;
}



/*** Create an exception entry in the linked list ***/
EXCP_T create_exception_block()
{
EXCP_T new;
int i;

if ((new=(EXCP_T)malloc(sizeof(struct exceptions)))==NULL) return NULL;
if (first_excp) last_excp->next=new;
else first_excp=new;
last_excp=new;
new->next=NULL;

for(i=0;i < NUM_ERRORS;++i) new->trapped[i]=0;
return new;
}



/*** Create a try entry ***/
int create_try_block(int pc)
{
if ((curr_try=(TRY_T)malloc(sizeof(struct try))) == NULL) return ERR_MALLOC;
if (first_try) last_try->next=curr_try;
else first_try=curr_try;
last_try=curr_try;
curr_try->next=NULL;
curr_try->start=pc;
return OK;
}



/*** Create a new WITH-WITHOUT block ***/
int create_with_block(int pc)
{
WITH_T new;

if ((new=(WITH_T)malloc(sizeof(struct with))) == NULL)
	return ERR_MALLOC;

if (first_with) last_with->next=new;
else first_with=new;
new->next=NULL;
new->prev=last_with;
last_with=curr_with=new;

new->proc=curr_proc;
new->start=pc;

return OK;
}



/*** Exit out of a procedure reseting and deleting appropriate stuff.
     It does NOT reset the program counter however, this is done in the
     mainloop() function.  ***/
void exit_procedure()
{
VAR_T v,next;

/* Delete variables */
for(v=first_var;v;v=next) {
	next=v->next;
	if (v->context == pstack_ptr) {
		remove_variable(v);  destroy_variable(v);  
		}
	}

/* Clear out loop and "with" stacks for this context ***/
while(lstack_ptr && lstack_ptr->context == pstack_ptr) pop_lstack();
while(wstack_ptr && wstack_ptr->context == pstack_ptr) pop_wstack();

/* Pop proc stack */
pop_pstack();
}



/*** This will parse a procedure call line, eg: obj->tom::hello line ***/
int get_proc(char *str,PROC_T *found_proc, OBJECT_T *found_obj)
{
CLASS_T cl;
PROC_T proc;
OBJECT_T obj;
char *s,*pname;
int ret;

proc=NULL;
obj=NULL;
pname=str;

/* Start at end since we could have multiple variable calls. 
   eg: call tom->dick->harry->proc1 */
for(s=str+strlen(str)-1;s >= str;--s) {
	if (*s=='.') {
		pname=s+1;
		if (!*pname) return ERR_SYNTAX;
		*s='\0';

		/* If no object name given then assume we're in a WITH block */
		if (s==str) {
			if (wstack_ptr==NULL) return ERR_NOT_IN_WITH_BLOCK;
			obj=wstack_ptr->obj;
			}
		else {
			/* Make sure object exists */
			FOR_ALL_OBJECTS(obj)
				if (!strcmp(obj->name,str)) break;
			if (obj==NULL) return ERR_UNDEFINED_OBJECT;
			}
		*s='.';
		break;
		}
	if (*s=='>') {
		pname=s+1;
		if (!*pname) return ERR_SYNTAX;
		--s;
		if (s <= str || *s != '-') return ERR_INVALID_PROCNAME;

		/* Get the name of the object stored in the variable */
		*s='\0';
		if ((ret=get_var_value(str,pstack_ptr))!=OK) {
			*s='-';  return ret;
			}
		*s='-';
		if (rstack_ptr->value==NULL) return ERR_UNDEFINED_OBJECT;
		FOR_ALL_OBJECTS(obj)
			if (!strcmp(obj->name,rstack_ptr->value)) break;
		if (obj==NULL) return ERR_UNDEFINED_OBJECT;
		break;
		}
	}

/* In case we've gone below */
if (s < str) s=str;

/* Find the procedure. If its an object procedure then the search will
   proceed up the class inheritance tree depending on the order the classes
   were listed in the inheritance list unless the class has been explicitly
   specified.
   eg: class tom inherits harry dick  - if procedure hello_world is called
   and tom does not have one but classes harry and dick do then the one in
   harry will be called since it is listed first. */
if (obj) {
	/* Is object procedure. See if class specified. 
	   eg: call obj->tom::hello_world */
	for(s=pname;*s;++s) {
		if (*s==':') {
			if (*(s+1)!=':') return ERR_INVALID_PROCNAME;
			*s='\0';

			FOR_ALL_CLASSES(cl)
				if (!strcmp(cl->name,pname)) break;	
			*s=':';
			if (cl==NULL) return ERR_UNDEFINED_CLASS;

			/* Check object is derived from the class */
			if (!is_derived_class(obj->class,cl))
				return ERR_OBJ_NOT_DERIVED_FROM_CLASS;

			/* Remove class:: bit and find actual proc */
			pname=s+2;
			if (!*pname) return ERR_SYNTAX;
			FOR_ALL_PROCS(proc)
				if (proc->class==cl &&
				    !strcmp(proc->name,pname)) break;
			if (proc==NULL) return ERR_UNDEFINED_PROC;
			break;
			}
		}
	if (proc==NULL) proc=get_class_proc(obj->class,pname);
	}
else {
	/* No class given , do straight search */
	FOR_ALL_PROCS(proc)
		if (proc->class==NULL && !strcmp(proc->name,pname)) break;
	}
if (proc==NULL) return ERR_UNDEFINED_PROC;

*found_proc=proc;
*found_obj=obj;

return OK;
}

