/***************************************************************************
 FILE: signals.c
 LVU : 1.0.0

 DESC:
 Some control functions for signal handling.

 Copyright (C) Neil Robertson 1999, 2000
 ***************************************************************************/

#include "globals.h"


/*** Set up the signals. Set most of them to ignore ***/
void set_sighandler()
{
int i;

for(i=0;i < NUM_SIGNALS;++i) set_sig_default(siglist[i].signum);
}



/*** Set the default behaviour for the signals ***/
void set_sig_default(int signum)
{
switch(signum) {
	case SIGINT:
	case SIGQUIT:
	case SIGTERM:
	signal(signum,signal_handler);
	break;

	case SIGCLD:
	signal(SIGCLD,SIG_DFL); /* Otherwise wait(), waitpid() won't work */
	break;
 
	default:
	signal(signum,SIG_IGN);
	}
}



/*** Main signal handler. Upon receipt of a signal this will queue it if 
     we have a valid handler for it. ***/
void signal_handler(int signum)
{
SIGHANDLER_T sh;

/* See if we've got a handler for the signal */
FOR_ALL_SIGHANDLERS(sh) {
	if (sh->signum==signum) {
		push_sigqueue(sh);  break;
		}
	}

/* Carry out certain defaults if no handler is available */
if (sh==NULL) {
	switch(signum) {
		case SIGINT:
		case SIGQUIT:
		puts("Interrupted.");  
		system_exit(1);

		case SIGTERM:
		puts("Terminated.");
		system_exit(1);
		}
	}

/* Reset */
signal(signum,signal_handler);
}



/*** Process all the pending signals on the queue ***/
int process_sigqueue(int pc)
{
int ret,i;
static int being_processed=0;

if (being_processed) return OK;

being_processed=1;
i=sigq_start;
while(sigq[i]) {
	set_var_value("$signal",NULL,siglist[sigq[i]->siglist_index].name,YES);

	/* Handler can be NULL in which case we just ignore the signal */
	if (sigq[i]->handler) {
		push_pstack(sigq[i]->handler,pc,pstack_ptr->obj);
		if ((ret=mainloop(YES,sigq[i]->handler->start)) != OK) 
			return ret;
		}
	pop_sigqueue();

	if (i==sigq_next) break;
	i=(i+1)%SIGQ_SIZE;
	}
being_processed=0;

return OK;
}



/*** Push a pending signal onto the queue. ***/
void push_sigqueue(SIGHANDLER_T sh)
{
/* If no room is left theres not much we can do */
if (sigq_next==sigq_start && sigq[sigq_next]!=NULL) return;

sigq[sigq_next]=sh;
sigq_next=(sigq_next+1)%SIGQ_SIZE;
}



/*** Pop the first signal from queue ***/
void pop_sigqueue()
{
sigq[sigq_start]=NULL;
sigq_start=(sigq_start+1)%SIGQ_SIZE;
}
