/***************************************************************************
 FILE: stacks.c
 LVU : 1.0.0

 DESC:
 This contains the main push & pop functions for the result, procedure
 loop & with stacks. Also has store_in_pstack() for storing temporary maths
 and boolean results.

 Copyright (C) Neil Robertson 1999, 2000
 ***************************************************************************/

#include "globals.h"

/*** Get the value from a string , a variable or a command ***/
int push_rstack_result(int *pc)
{
char *w;
int ret,com;

w=pword[*pc].word;
switch(*w) {
	case '"':
	w=pword[*pc].word+1;
	w[strlen(w)-1]='\0';
	push_rstack(w);
	w[strlen(w)]='"';
	++*pc;
	return OK;


	case '{':
	/* Can't have commands executed if we're outside a procedure eg: in
	   a set command */
	if (pstack_ptr==NULL) return ERR_SYNTAX;

	++*pc;

	/* See if command can be nested */
	com=pword[*pc].com;
	if (com >= LABEL && com <= SWCONTINUE)
		return ERR_COMMAND_CANT_BE_NESTED;

	switch(com) {
		case TRY:
		case TRIED:
		case CATCH:
		case CATCHRUN:
		case CAUGHT:
		case CATCHRETURN:
		case CATCHRETURNEND:
		case CATCHBREAK:
		case IF:
		case ELSE:
		case FI:
		case THROW:
		case EXIT:
		case WITH:
		case WITHOUT:
		case CLS:
		case LOCATE:
		return ERR_COMMAND_CANT_BE_NESTED;
		}
		
	/* Run the nested command */
	ret=run_command(pc);
	if (pword[*pc].word[0]=='}') ++*pc;
	return ret;


	default:
	++*pc;
	if (number_type(w)!=NOT_NUMBER) return push_rstack(w);
	if (*w!='-') return get_var_value(w,pstack_ptr);

	/* Deal with a negated var value */
	++w;
	if ((ret=get_var_value(w,pstack_ptr))!=OK) return ret;
	if (number_type(rstack_ptr->value)==NOT_NUMBER) 
		return ERR_CANT_NEG_NON_NUM_VAR;

	/* If value already is negative remove the minus sign */
	if (rstack_ptr->value && rstack_ptr->value[0]=='-') {
		strcpy(rstack_ptr->value,rstack_ptr->value+1);
		return OK;
		}

	/* Else add in the sign */
	w=strdup("-");
	append_string(&w,rstack_ptr->value);
	return push_rstack(w);
	}
}



/*** Push a value onto the result stack ***/
int push_rstack(char *value)
{
RSTACK_T new;

if ((new=(RSTACK_T)malloc(sizeof(struct rstack)))==NULL)
	return ERR_MALLOC;

if (rstack_start) new->prev=rstack_ptr;
else {
	rstack_start=new;
	new->prev=NULL;
	}

rstack_ptr=new;
new->value=NULL;
new->context=pstack_ptr;

return set_string(&new->value,value);
}



/*** Pop the result stack ***/
void pop_rstack()
{
RSTACK_T prev;

if (rstack_ptr==NULL) return;
FREE(rstack_ptr->value);
prev=rstack_ptr->prev;
free(rstack_ptr);
if (prev) rstack_ptr=prev;  
else rstack_start=rstack_ptr=NULL;
}



/*** Cleardown the result stack for current procedure. Used at end of loops 
     etc to stop stack building up to an obscene size ***/
void cleardown_rstack()
{
RSTACK_T rs;

while(rstack_ptr) {
	if (rstack_ptr->context==pstack_ptr) pop_rstack();
	else {
		/* Might have return values from other procs sandwiched 
		   inbetween */
		for(rs=rstack_ptr;rs;rs=rs->prev) 
			if (rs->context==pstack_ptr) break;
		if (rs==NULL) return;
		pop_rstack();
		}
	}
}



/*** Push a prcoedure and return position onto the proc stack ***/
int push_pstack(PROC_T proc,int ret_pc, OBJECT_T obj)
{
PSTACK_T new;

if ((new=(PSTACK_T)malloc(sizeof(struct pstack)))==NULL)
	return ERR_MALLOC;

if (pstack_start) new->prev=pstack_ptr;
else {
	pstack_start=new;
	new->prev=NULL;
	}

pstack_ptr=new;
new->proc=proc;
new->obj=obj;
new->ret_pc=ret_pc;
new->catchreturn_pc[0]=0;
new->catchreturn_pc[1]=0;
new->catchbreak_pc=0;
new->pstore=NULL;
return OK;
}



/*** Pop the proc stack ***/
void pop_pstack()
{
PSTACK_T prev;

if (pstack_ptr==NULL) return;
prev=pstack_ptr->prev;
free(pstack_ptr);
if (prev) pstack_ptr=prev;  
else pstack_start=pstack_ptr=NULL;
}



/*** Copy various results for com_maths() and evaluate_single_boolean() for 
     this proc context ***/
int store_in_pstack(int from)
{
struct pstack_pstore **new;
int i,len;

/* Whole lot stored at once , not in bits so don't need to check
   to see if anything is there already */
len=pword[from].end - from + 1;
new=(struct pstack_pstore **)malloc(len*sizeof(struct pstack_pstore *));
if (new==NULL) return ERR_MALLOC;

for(i=0;i < len;++i) {
	if ((new[i]=(struct pstack_pstore *)malloc(sizeof(struct pstack_pstore)))==NULL)
		return ERR_MALLOC;

	new[i]->processed=pword[i+from].processed;
	new[i]->result_set=pword[i+from].result_set;
	new[i]->next_pc=pword[i+from].next_pc;
	new[i]->result=pword[i+from].result;
	}

pstack_ptr->pstore=new;
return OK;
}



/*** Copy results back into context ***/
void copy_from_pstack(int from)
{
int i;

for(i=from;i <= pword[from].end;++i) {
	pword[i].processed=pstack_ptr->pstore[i-from]->processed;
	pword[i].result_set=pstack_ptr->pstore[i-from]->result_set;
	pword[i].next_pc=pstack_ptr->pstore[i-from]->next_pc;
	pword[i].result=pstack_ptr->pstore[i-from]->result;

	free(pstack_ptr->pstore[i-from]);
	}
free(pstack_ptr->pstore);
}



/*** Push loop entry onto the loop stack ***/
int push_lstack(LOOP_T loop)
{
LSTACK_T new;

if ((new=(LSTACK_T)malloc(sizeof(struct lstack)))==NULL)
	return ERR_MALLOC;

if (lstack_start) new->prev=lstack_ptr;
else {
	lstack_start=new;
	new->prev=NULL;
	}
lstack_ptr=new;
new->loop=loop;
new->to_val=0;
new->step_val=0;
new->arr_index=NULL;
new->arr_value=NULL;
new->forent=NULL;
new->forelem_list=NULL;
new->forelem_es=NULL;
new->context=pstack_ptr;
return OK;
}



/*** Pop the loop stack ***/
void pop_lstack()
{
LSTACK_T prev;

FREE(lstack_ptr->forelem_list);
FREE(lstack_ptr->forelem_es);

if (lstack_ptr==NULL) return;
prev=lstack_ptr->prev;
free(lstack_ptr);
if (prev) lstack_ptr=prev;  
else lstack_start=lstack_ptr=NULL;
}



/*** Push WITH block onto WITH stack ***/
int push_wstack(WITH_T with, OBJECT_T obj)
{
WSTACK_T new;

if ((new=(WSTACK_T)malloc(sizeof(struct wstack)))==NULL)
	return ERR_MALLOC;

if (wstack_start) new->prev=wstack_ptr;
else {
	wstack_start=new;
	new->prev=NULL;
	}
wstack_ptr=new;

new->with=with;
new->obj=obj;
new->context=pstack_ptr;
return OK;
}



/*** Pop the WITH stack ***/
void pop_wstack()
{
WSTACK_T prev;

if (wstack_ptr==NULL) return;
prev=wstack_ptr->prev;
free(wstack_ptr);
if (prev) wstack_ptr=prev;
else wstack_start=wstack_ptr=NULL;
}



/*** Delete any loops or with blocks off their stacks if we are no longer
     inside then. ***/
void cleardown_lw_stacks(int pc)
{
LSTACK_T ls,lprev;
WSTACK_T ws,wprev;

for(ls=lstack_ptr;ls;ls=lprev) {
	/* Break at 1st loop of another procedure */
	if (ls->context != pstack_ptr) break;
	lprev=ls->prev;
	if (pc < ls->loop->start || pc > ls->loop->end) pop_lstack();
	}

for(ws=wstack_ptr;ws;ws=wprev) {
	if (ws->context != pstack_ptr) return;
	wprev=ws->prev;
	if (pc < ws->with->start || pc > ws->with->end) pop_wstack();
	}
}
